/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.building;

import java.io.Serializable;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerFactory;

import org.ow2.bonita.facade.def.element.AttachmentDefinition;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.FilterDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.IncomingEventDefinition;
import org.ow2.bonita.facade.def.element.MultiInstantiationDefinition;
import org.ow2.bonita.facade.def.element.OutgoingEventDefinition;
import org.ow2.bonita.facade.def.element.RoleMapperDefinition;
import org.ow2.bonita.facade.def.element.SubflowParameterDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.XmlConstants;
import org.w3c.dom.Node;

/**
 * @author Anthony Birembaut
 *
 */
public class XmlDefExporter {

  /**
   * the product version
   */
  private static final String PRODUCT_VERSION = "5.2";

  /**
   * Logger
   */
  private static Logger LOGGER = Logger.getLogger(XmlDefExporter.class.getName());

  /**
   * Document builder factory
   */
  private DocumentBuilderFactory documentBuilderFactory;

  /**
   * Transformer factory
   */
  private TransformerFactory transformerFactory;

  /**
   * Instance attribute
   */
  private static XmlDefExporter INSTANCE = null;

  /**
   * @return the XmlDefExporter instance
   */
  public static XmlDefExporter getInstance() {
    if (INSTANCE == null) {
      INSTANCE = new XmlDefExporter();
    }
    return INSTANCE;
  }

  /**
   * Private constructor to prevent instantiation
   */
  private XmlDefExporter() {
    documentBuilderFactory = DocumentBuilderFactory.newInstance();
    transformerFactory= TransformerFactory.newInstance();
  }

  /**
   * Create the document and root node
   * @return the XML process definition
   * @throws Exception 
   * @throws ParserConfigurationException 
   */
  public byte[] createProcessDefinition(ProcessDefinition processDefinition) {

    URL xsdURL = getClass().getResource("/" + XmlConstants.XML_PROCESS_DEF_1_1_SCHEMA);
    XmlBuilder xmlDefBuilder = new XmlBuilder(documentBuilderFactory, transformerFactory, xsdURL);
    try {
      xmlDefBuilder.createDocument();
      Map<String, Serializable> rootElementAttributes = new HashMap<String, Serializable>();
      rootElementAttributes.put(XmlDef.PRODUCT_VERSION, PRODUCT_VERSION);
      Node rootNode = xmlDefBuilder.createRootNode(XmlDef.PROCESS_DEFINITION, rootElementAttributes);

      createProcess(xmlDefBuilder, rootNode, processDefinition);

      return xmlDefBuilder.done();
    } catch (Exception e) {
      String errorMessage = "Unable to build the process definition XML";
      if (LOGGER.isLoggable(Level.SEVERE)) {
        LOGGER.log(Level.SEVERE, errorMessage, e);
      }
      throw new BonitaRuntimeException(errorMessage, e);
    }
  }

  private Node createProcess(XmlBuilder xmlDefBuilder, Node parentNode, ProcessDefinition processDefinition) throws Exception {

    Map<String, Serializable> processElementAttributes = new HashMap<String, Serializable>();
    processElementAttributes.put(XmlDef.NAME, processDefinition.getName());
    processElementAttributes.put(XmlDef.VERSION, processDefinition.getVersion());
    Node processNode = xmlDefBuilder.createNode(parentNode, XmlDef.PROCESS, processElementAttributes);

    xmlDefBuilder.createNode(processNode, XmlDef.LABEL, processDefinition.getLabel());
    xmlDefBuilder.createNode(processNode, XmlDef.DESCRIPTION, processDefinition.getDescription());

    createParticipants(xmlDefBuilder, processNode, processDefinition.getParticipants());
    createDataFields(xmlDefBuilder, processNode, processDefinition.getDataFields());
    createAttachments(xmlDefBuilder, processNode, processDefinition.getAttachments());
    createActivities(xmlDefBuilder, processNode, processDefinition.getActivities());
    createTransitions(xmlDefBuilder, processNode, processDefinition.getTransitions());
    createConnectors(xmlDefBuilder, processNode, processDefinition.getConnectors());

    return processNode;
  }

  private void createActivities(XmlBuilder xmlDefBuilder, Node parentNode, Set<ActivityDefinition> activities) throws Exception {

    Node activitiesNode = xmlDefBuilder.createNode(parentNode, XmlDef.ACTIVITIES);
    for (ActivityDefinition activityDefinition : activities) {
      Map<String, Serializable> activityElementAttributes = new HashMap<String, Serializable>();
      activityElementAttributes.put(XmlDef.NAME, activityDefinition.getName());
      Node activityNode = xmlDefBuilder.createNode(activitiesNode, XmlDef.ACTIVITY, activityElementAttributes);

      xmlDefBuilder.createNode(activityNode, XmlDef.LABEL, activityDefinition.getLabel());
      xmlDefBuilder.createNode(activityNode, XmlDef.DESCRIPTION, activityDefinition.getDescription());
      xmlDefBuilder.createNode(activityNode, XmlDef.DYNAMIC_LABEL, activityDefinition.getDynamicLabel());
      xmlDefBuilder.createNode(activityNode, XmlDef.DYNAMIC_DESCRIPTION, activityDefinition.getDynamicDescription());
      xmlDefBuilder.createNode(activityNode, XmlDef.PRIORITY, activityDefinition.getPriority());
      xmlDefBuilder.createNode(activityNode, XmlDef.EXECUTING_TIME, activityDefinition.getExecutingTime());
      xmlDefBuilder.createNode(activityNode, XmlDef.SPLIT_TYPE, activityDefinition.getSplitType());
      xmlDefBuilder.createNode(activityNode, XmlDef.JOIN_TYPE, activityDefinition.getJoinType());
      xmlDefBuilder.createNode(activityNode, XmlDef.SUBFLOW_PROCESS_NAME, activityDefinition.getSubflowProcessName());
      xmlDefBuilder.createNode(activityNode, XmlDef.SUBFLOW_PROCESS_VERSION, activityDefinition.getSubflowProcessVersion());
      xmlDefBuilder.createNode(activityNode, XmlDef.TIMER_CONDITION, activityDefinition.getTimerCondition());
      xmlDefBuilder.createNode(activityNode, XmlDef.TYPE, activityDefinition.getType());
      xmlDefBuilder.createNode(activityNode, XmlDef.ASYNCHRONOUS, activityDefinition.isAsynchronous());

      Node performersNode = xmlDefBuilder.createNode(activityNode, XmlDef.PERFORMERS);
      for (String performer : activityDefinition.getPerformers()) {
        xmlDefBuilder.createNode(performersNode, XmlDef.PERFORMER, performer);
      }

      if (activityDefinition.isInALoop()) {
        xmlDefBuilder.createNode(activityNode, XmlDef.LOOP_CONDITION, activityDefinition.getLoopCondition());
        xmlDefBuilder.createNode(activityNode, XmlDef.BEFORE_EXECUTION, activityDefinition.evaluateLoopConditionBeforeExecution());
        xmlDefBuilder.createNode(activityNode, XmlDef.LOOP_MAXIMUM, activityDefinition.getLoopMaximum());
      }
      
      createConnectors(xmlDefBuilder, activityNode, activityDefinition.getConnectors());
      createDataFields(xmlDefBuilder, activityNode, activityDefinition.getDataFields());
      createFilter(xmlDefBuilder, activityNode, activityDefinition.getFilter());
      createDeadlines(xmlDefBuilder, activityNode, activityDefinition.getDeadlines());
      createIncomingEvent(xmlDefBuilder, activityNode, activityDefinition.getIncomingEvent());
      createOutgoingEvents(xmlDefBuilder, activityNode, activityDefinition.getOutgoingEvents());
      createMultiInstantiationDefinition(xmlDefBuilder, activityNode, activityDefinition.getMultiInstantiationDefinition());
      createMultipleActivitiesInstantiator(xmlDefBuilder, activityNode, activityDefinition.getMultipleInstancesInstantiator());
      createMultipleActivitiesJoinChecker(xmlDefBuilder, activityNode, activityDefinition.getMultipleInstancesJoinChecker());

      createSubflowParameters(xmlDefBuilder, activityNode, activityDefinition.getSubflowInParameters(), XmlDef.SUBFLOW_IN_PARAMETERS, XmlDef.SUBFLOW_IN_PARAMETER);
      createSubflowParameters(xmlDefBuilder, activityNode, activityDefinition.getSubflowOutParameters(), XmlDef.SUBFLOW_OUT_PARAMETERS, XmlDef.SUBFLOW_OUT_PARAMETER);
    }
  }

  private void createSubflowParameters(XmlBuilder xmlDefBuilder, Node parentNode,
      Set<SubflowParameterDefinition> subflowParameters, String xmlGroupName, String xmlElementName) throws Exception {

    Node subflowParametersNode = xmlDefBuilder.createNode(parentNode, xmlGroupName);
    for (SubflowParameterDefinition subflowParameter : subflowParameters) {
      Node subflowParameterNode = xmlDefBuilder.createNode(subflowParametersNode, xmlElementName);

      xmlDefBuilder.createNode(subflowParameterNode, XmlDef.SOURCE, subflowParameter.getSource());
      xmlDefBuilder.createNode(subflowParameterNode, XmlDef.DESTINATION, subflowParameter.getDestination());
    }
  }

  private void createMultiInstantiationDefinition(XmlBuilder xmlDefBuilder, Node parentNode, MultiInstantiationDefinition multiInstantiationDefinition) throws Exception {

    if (multiInstantiationDefinition != null) {
      Node multiInstantiationNode = xmlDefBuilder.createNode(parentNode, XmlDef.MULTI_INSTANTIATION);

      if (multiInstantiationDefinition != null) {
        xmlDefBuilder.createNode(multiInstantiationNode, XmlDef.CLASSNAME, multiInstantiationDefinition.getClassName());
        xmlDefBuilder.createNode(multiInstantiationNode, XmlDef.DESCRIPTION, multiInstantiationDefinition.getDescription());
        xmlDefBuilder.createNode(multiInstantiationNode, XmlDef.VARIABLE_NAME, multiInstantiationDefinition.getVariableName());

        Node parametersNode = xmlDefBuilder.createNode(multiInstantiationNode, XmlDef.PARAMETERS);
        Map<String, Object[]> multiInstantiationParameters = multiInstantiationDefinition.getParameters();
        for (String key : multiInstantiationParameters.keySet()) {
          Map<String, Serializable> multiInstantiationParameterAttributes = new HashMap<String, Serializable>();
          multiInstantiationParameterAttributes.put(XmlDef.NAME, key);
          xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(multiInstantiationParameters.get(key)), multiInstantiationParameterAttributes);
        }
      }
    }
  }

  private void createOutgoingEvents(XmlBuilder xmlDefBuilder, Node parentNode, Set<OutgoingEventDefinition> outgoingEvents) throws Exception {

    Node outgoingEventsNode = xmlDefBuilder.createNode(parentNode, XmlDef.OUTGOING_EVENTS);
    for (OutgoingEventDefinition outgoingEvent : outgoingEvents) {
      Map<String, Serializable> outgoingEventElementAttributes = new HashMap<String, Serializable>();
      outgoingEventElementAttributes.put(XmlDef.NAME, outgoingEvent.getName());
      Node outgoingEventNode = xmlDefBuilder.createNode(outgoingEventsNode, XmlDef.OUTGOING_EVENT, outgoingEventElementAttributes);

      xmlDefBuilder.createNode(outgoingEventNode, XmlDef.TIME_TO_LIVE, outgoingEvent.getTimeToLive());
      xmlDefBuilder.createNode(outgoingEventNode, XmlDef.TO_ACTIVITY, outgoingEvent.getToActivityName());
      xmlDefBuilder.createNode(outgoingEventNode, XmlDef.TO_PROCESS, outgoingEvent.getToProcessName());

      Node parametersNode = xmlDefBuilder.createNode(outgoingEventNode, XmlDef.PARAMETERS);
      Map<String, Object> outgoingEventParameters = outgoingEvent.getParameters();
      for (String key : outgoingEventParameters.keySet()) {
        Map<String, Serializable> outgoingEventParameterAttributes = new HashMap<String, Serializable>();
        outgoingEventParameterAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize((Serializable)outgoingEventParameters.get(key)), outgoingEventParameterAttributes);
      }
    }
  }

  private void createIncomingEvent(XmlBuilder xmlDefBuilder, Node parentNode, IncomingEventDefinition incomingEvent) throws Exception {

    if (incomingEvent != null) {
      Map<String, Serializable> incomingEventElementAttributes = new HashMap<String, Serializable>();
      incomingEventElementAttributes.put(XmlDef.NAME, incomingEvent.getName());
      Node incomingEventNode = xmlDefBuilder.createNode(parentNode, XmlDef.INCOMING_EVENT, incomingEventElementAttributes);

      xmlDefBuilder.createNode(incomingEventNode, XmlDef.EXPRESSION, incomingEvent.getExpression());
    }
  }

  private void createDeadlines(XmlBuilder xmlDefBuilder, Node parentNode, Set<DeadlineDefinition> deadlines) throws Exception {

    Node deadlinesNode = xmlDefBuilder.createNode(parentNode, XmlDef.DEADLINES);
    for (DeadlineDefinition deadline : deadlines) {
      Node deadlineNode = xmlDefBuilder.createNode(deadlinesNode, XmlDef.DEADLINE);

      xmlDefBuilder.createNode(deadlineNode, XmlDef.CLASSNAME, deadline.getClassName());
      xmlDefBuilder.createNode(deadlineNode, XmlDef.CONDITION, deadline.getCondition());
      xmlDefBuilder.createNode(deadlineNode, XmlDef.DESCRIPTION, deadline.getDescription());

      Node parametersNode = xmlDefBuilder.createNode(deadlineNode, XmlDef.PARAMETERS);
      Map<String, Object[]> deadlineParameters = deadline.getParameters();
      for (String key : deadlineParameters.keySet()) {
        Map<String, Serializable> deadlineParameterAttributes = new HashMap<String, Serializable>();
        deadlineParameterAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(deadlineParameters.get(key)), deadlineParameterAttributes);
      }
    }
  }

  private void createMultipleActivitiesInstantiator(XmlBuilder xmlDefBuilder, Node parentNode, MultiInstantiationDefinition instantiator) throws Exception {
    if (instantiator != null) {
      Node instantiatorNode = xmlDefBuilder.createNode(parentNode, XmlDef.MULTIPLE_ACT_INSTANTIATOR);
      xmlDefBuilder.createNode(instantiatorNode, XmlDef.CLASSNAME, instantiator.getClassName());
      xmlDefBuilder.createNode(instantiatorNode, XmlDef.DESCRIPTION, instantiator.getDescription());

      Node parametersNode = xmlDefBuilder.createNode(instantiatorNode, XmlDef.PARAMETERS);
      Map<String, Object[]> instantiatorParameters = instantiator.getParameters();
      for (String key : instantiatorParameters.keySet()) {
        Map<String, Serializable> instantiatorParameterAttributes = new HashMap<String, Serializable>();
        instantiatorParameterAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(instantiatorParameters.get(key)), instantiatorParameterAttributes);
      }
    }
  }

  private void createMultipleActivitiesJoinChecker(XmlBuilder xmlDefBuilder, Node parentNode, MultiInstantiationDefinition joinChecker) throws Exception {
    if (joinChecker != null) {
      Node joinCheckerNode = xmlDefBuilder.createNode(parentNode, XmlDef.MULTIPLE_ACT_JOINCHECKER);
      xmlDefBuilder.createNode(joinCheckerNode, XmlDef.CLASSNAME, joinChecker.getClassName());
      xmlDefBuilder.createNode(joinCheckerNode, XmlDef.DESCRIPTION, joinChecker.getDescription());

      Node parametersNode = xmlDefBuilder.createNode(joinCheckerNode, XmlDef.PARAMETERS);
      Map<String, Object[]> joinCheckerParameters = joinChecker.getParameters();
      for (String key : joinCheckerParameters.keySet()) {
        Map<String, Serializable> joinCheckerParameterAttributes = new HashMap<String, Serializable>();
        joinCheckerParameterAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(joinCheckerParameters.get(key)), joinCheckerParameterAttributes);
      }
    }
  }

  private void createFilter(XmlBuilder xmlDefBuilder, Node parentNode, FilterDefinition filter) throws Exception {
    if (filter != null) {
      Node filterNode = xmlDefBuilder.createNode(parentNode, XmlDef.FILTER);
      xmlDefBuilder.createNode(filterNode, XmlDef.CLASSNAME, filter.getClassName());
      xmlDefBuilder.createNode(filterNode, XmlDef.DESCRIPTION, filter.getDescription());

      Node parametersNode = xmlDefBuilder.createNode(filterNode, XmlDef.PARAMETERS);
      Map<String, Object[]> filterParameters = filter.getParameters();
      for (String key : filterParameters.keySet()) {
        Map<String, Serializable> filterParameterAttributes = new HashMap<String, Serializable>();
        filterParameterAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(filterParameters.get(key)), filterParameterAttributes);
      }
    }
  }

  private void createTransitions(XmlBuilder xmlDefBuilder, Node parentNode, Set<TransitionDefinition> transitions) throws Exception {

    Node transitionsNode = xmlDefBuilder.createNode(parentNode, XmlDef.TRANSITIONS);
    for (TransitionDefinition transitionDefinition : transitions) {
      Map<String, Serializable> transitionElementAttributes = new HashMap<String, Serializable>();
      transitionElementAttributes.put(XmlDef.NAME, transitionDefinition.getName());
      Node transitionNode = xmlDefBuilder.createNode(transitionsNode, XmlDef.TRANSITION, transitionElementAttributes);

      xmlDefBuilder.createNode(transitionNode, XmlDef.LABEL, transitionDefinition.getLabel());
      xmlDefBuilder.createNode(transitionNode, XmlDef.CONDITION, transitionDefinition.getCondition());
      xmlDefBuilder.createNode(transitionNode, XmlDef.DESCRIPTION, transitionDefinition.getDescription());
      xmlDefBuilder.createNode(transitionNode, XmlDef.FROM, transitionDefinition.getFrom());
      xmlDefBuilder.createNode(transitionNode, XmlDef.TO, transitionDefinition.getTo());
      xmlDefBuilder.createNode(transitionNode, XmlDef.ISDEFAULT, transitionDefinition.isDefault());
    }
  }

  private void createDataFields(XmlBuilder xmlDefBuilder, Node parentNode, Set<DataFieldDefinition> dataFields) throws Exception {

    Node dataFieldsNode = xmlDefBuilder.createNode(parentNode, XmlDef.DATA_FIELDS);
    for (DataFieldDefinition dataFieldDefinition : dataFields) {
      Map<String, Serializable> dataFieldElementAttributes = new HashMap<String, Serializable>();
      dataFieldElementAttributes.put(XmlDef.NAME, dataFieldDefinition.getName());
      Node dataFieldNode = xmlDefBuilder.createNode(dataFieldsNode, XmlDef.DATA_FIELD, dataFieldElementAttributes);

      xmlDefBuilder.createNode(dataFieldNode, XmlDef.LABEL, dataFieldDefinition.getLabel());
      xmlDefBuilder.createNode(dataFieldNode, XmlDef.VALUE, Misc.serialize(dataFieldDefinition.getInitialValue()));
      xmlDefBuilder.createNode(dataFieldNode, XmlDef.DESCRIPTION, dataFieldDefinition.getDescription());
      xmlDefBuilder.createNode(dataFieldNode, XmlDef.DATATYPE_CLASSNAME, dataFieldDefinition.getDataTypeClassName());
      xmlDefBuilder.createNode(dataFieldNode, XmlDef.SCRIPTING_VALUE, dataFieldDefinition.getScriptingValue());

      Node enumerationValuesNode = xmlDefBuilder.createNode(dataFieldNode, XmlDef.ENUMERATION_VALUES);
      for (String enumerationValue : dataFieldDefinition.getEnumerationValues()) {
        xmlDefBuilder.createNode(enumerationValuesNode, XmlDef.ENUMERATION_VALUE, enumerationValue);
      }
    }
  }

  private void createAttachments(XmlBuilder xmlDefBuilder, Node parentNode, Map<String, AttachmentDefinition> attachments) throws Exception {

    Node attachmentsNode = xmlDefBuilder.createNode(parentNode, XmlDef.ATTACHMENTS);
    for (AttachmentDefinition attachmentDefinition : attachments.values()) {
      Map<String, Serializable> attachmentElementAttributes = new HashMap<String, Serializable>();
      attachmentElementAttributes.put(XmlDef.NAME, attachmentDefinition.getName());
      Node attachmentNode = xmlDefBuilder.createNode(attachmentsNode, XmlDef.ATTACHMENT, attachmentElementAttributes);

      xmlDefBuilder.createNode(attachmentNode, XmlDef.LABEL, attachmentDefinition.getLabel());
      xmlDefBuilder.createNode(attachmentNode, XmlDef.FILE_PATH, attachmentDefinition.getFilePath());
      xmlDefBuilder.createNode(attachmentNode, XmlDef.DESCRIPTION, attachmentDefinition.getDescription());
      xmlDefBuilder.createNode(attachmentNode, XmlDef.FILE_NAME, attachmentDefinition.getFileName());
    }
  }

  private void createParticipants(XmlBuilder xmlDefBuilder, Node parentNode, Set<ParticipantDefinition> participants) throws Exception {

    Node participantsNode = xmlDefBuilder.createNode(parentNode, XmlDef.PARTICIPANTS);
    for (ParticipantDefinition participantDefinition : participants) {
      Map<String, Serializable> participantElementAttributes = new HashMap<String, Serializable>();
      participantElementAttributes.put(XmlDef.NAME, participantDefinition.getName());
      Node participantNode = xmlDefBuilder.createNode(participantsNode, XmlDef.PARTICIPANT, participantElementAttributes);

      xmlDefBuilder.createNode(participantNode, XmlDef.LABEL, participantDefinition.getLabel());
      xmlDefBuilder.createNode(participantNode, XmlDef.DESCRIPTION, participantDefinition.getDescription());

      createRoleMapper(xmlDefBuilder, participantNode, participantDefinition.getRoleMapper());
    }
  }

  private void createRoleMapper(XmlBuilder xmlDefBuilder, Node parentNode, RoleMapperDefinition roleMapperDefinition) throws Exception {

    if (roleMapperDefinition != null) {
      Node roleMapperNode = xmlDefBuilder.createNode(parentNode, XmlDef.ROLE_MAPPER);

      if (roleMapperDefinition != null){
        xmlDefBuilder.createNode(roleMapperNode, XmlDef.CLASSNAME, roleMapperDefinition.getClassName());
        xmlDefBuilder.createNode(roleMapperNode, XmlDef.DESCRIPTION, roleMapperDefinition.getDescription());

        Node parametersNode = xmlDefBuilder.createNode(roleMapperNode, XmlDef.PARAMETERS);
        Map<String, Object[]> roleMapperParameters = roleMapperDefinition.getParameters();
        for (String key : roleMapperParameters.keySet()) {
          Map<String, Serializable> roleMapperParameterAttributes = new HashMap<String, Serializable>();
          roleMapperParameterAttributes.put(XmlDef.NAME, key);
          xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(roleMapperParameters.get(key)), roleMapperParameterAttributes);
        }
      }
    }
  }

  private void createConnectors(XmlBuilder xmlDefBuilder, Node parentNode, List<HookDefinition> hooks) throws Exception {

    Node connectorsNode = xmlDefBuilder.createNode(parentNode, XmlDef.CONNECTORS);
    for (HookDefinition hook : hooks) {
      Node connectorNode = xmlDefBuilder.createNode(connectorsNode, XmlDef.CONNECTOR);

      xmlDefBuilder.createNode(connectorNode, XmlDef.CLASSNAME, hook.getClassName());
      xmlDefBuilder.createNode(connectorNode, XmlDef.DESCRIPTION, hook.getDescription());
      xmlDefBuilder.createNode(connectorNode, XmlDef.EVENT, hook.getEvent());
      xmlDefBuilder.createNode(connectorNode, XmlDef.IS_THROWING_EXCEPTION, hook.isThrowingException());

      Node parametersNode = xmlDefBuilder.createNode(connectorNode, XmlDef.PARAMETERS);
      Map<String, Object[]> hookParameters = hook.getParameters();
      for (String key : hookParameters.keySet()) {
        Map<String, Serializable> hookParametersAttributes = new HashMap<String, Serializable>();
        hookParametersAttributes.put(XmlDef.NAME, key);
        xmlDefBuilder.createNode(parametersNode, XmlDef.PARAMETER, Misc.serialize(hookParameters.get(key)), hookParametersAttributes);
      }
    }
  }
}
