/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.definition.activity;

import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.connector.core.Connector;
import org.ow2.bonita.connector.core.Filter;
import org.ow2.bonita.connector.core.MultiInstantiatorInstantiator;
import org.ow2.bonita.connector.core.MultiInstantiatorJoinChecker;
import org.ow2.bonita.connector.core.MultipleInstancesInstantiator;
import org.ow2.bonita.connector.core.MultipleInstancesJoinChecker;
import org.ow2.bonita.connector.core.PerformerAssignFilter;
import org.ow2.bonita.connector.core.ProcessConnector;
import org.ow2.bonita.connector.core.RoleResolver;
import org.ow2.bonita.connector.core.desc.Getter;
import org.ow2.bonita.connector.core.desc.Setter;
import org.ow2.bonita.definition.Hook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.MultiInstantiatorDescriptor;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.TxHook;
import org.ow2.bonita.facade.QueryDefinitionAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.element.ConnectorDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.MultiInstantiationDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition.Event;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.HookInvocationException;
import org.ow2.bonita.facade.impl.StandardAPIAccessorImpl;
import org.ow2.bonita.facade.impl.StandardQueryAPIAccessorImpl;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.GroovyExpression;
import org.ow2.bonita.util.GroovyUtil;
import org.ow2.bonita.util.Misc;

public final class ConnectorExecutor {

  private static final Logger LOG = Logger.getLogger(ConnectorExecutor.class.getName());

  private ConnectorExecutor() { }

  private static void executeConnector(final TxHook connector, Map<String, Object[]> connectorParameters, ProcessInstanceUUID instanceUUID, ActivityInstance activityInst, Map<String, Object> extraParameters) throws Exception {
    ProcessDefinitionUUID processUUID = instanceUUID.getProcessDefinitionUUID();
    ActivityInstanceUUID activityInstanceUUID = null;
    if (activityInst != null) {
      activityInstanceUUID = activityInst.getUUID();
    }
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    if (connectorParameters != null) {
      inputs = getInputs(connectorParameters);
    }
    if (connector instanceof ProcessConnector) {
      inputs.put("setProcessInstanceUUID", new Object[] {instanceUUID});
      inputs.put("setProcessDefinitionUUID", new Object[] {processUUID});
      if (activityInst != null) {
        inputs.put("setActivityInstanceUUID", new Object[] {activityInstanceUUID});
      }
    }
    ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
    try {
      ClassLoader processClassLoader = ClassDataLoader.getProcessClassLoader(processUUID);
      Thread.currentThread().setContextClassLoader(processClassLoader);
      if (!inputs.isEmpty()) {
        if (activityInst != null) {
          setParameters(inputs, connector, activityInst.getUUID(), null, extraParameters);
        } else {
          setParameters(inputs, connector, null, instanceUUID, extraParameters);
        }
      }
      connector.execute(new StandardAPIAccessorImpl(), activityInst);
      if (connectorParameters != null) {
        Map<String, Object[]> outputs = getOuputs(connectorParameters);
        if (activityInst != null) {
          setProcessOrActivityVariables(outputs, connector, activityInstanceUUID, instanceUUID);
        } else {
          setProcessOrActivityVariables(outputs, connector, null, instanceUUID);
        }
      }
    } finally {
      Thread.currentThread().setContextClassLoader(baseClassLoader);
    }
  }

  private static Map<String, Object> getGetterValues(Connector connector) {
    List<Getter> getters = connector.getGetters();
    Map<String, Object> values = new HashMap<String, Object>();
    if (getters != null) {
      for (Getter getter : getters) {
        try {
          String getterName = Connector.getGetterName(getter.getName());
          Method m = connector.getClass().getMethod(getterName, new Class[0]);
          Object variableValue = m.invoke(connector, new Object[0]);
          values.put(getter.getName(), variableValue);
        } catch (Exception e) {
          throw new BonitaRuntimeException(e.getMessage(), e);
        }
      }
    }
    return values;
  }

  private static void setProcessOrActivityVariables(Map<String, Object[]> outputs,
      Object connector, ActivityInstanceUUID activityInstanceUUID, ProcessInstanceUUID instanceUUID) throws BonitaException {
    if (connector instanceof Connector) {
      Map<String, Object> values = getGetterValues((Connector)connector);
      for (Entry<String, Object[]> output : outputs.entrySet()) {
        String expression = (String) output.getValue()[0];
        String variableName = output.getKey();
        StandardAPIAccessorImpl accessor = new StandardAPIAccessorImpl();
        RuntimeAPI runtime = accessor.getRuntimeAPI();
        Object newValue = null;
        if (activityInstanceUUID != null) {
          Object variableValue = runtime.evaluateGroovyExpression(expression, activityInstanceUUID, values, true);
          newValue = convertIfPossible(variableValue, getDataTypeClassName(variableName, instanceUUID, activityInstanceUUID));
          runtime.setVariable(activityInstanceUUID, variableName, newValue);
        } else {
          Object variableValue = runtime.evaluateGroovyExpression(expression, instanceUUID, values, true);
          newValue = convertIfPossible(variableValue, getDataTypeClassName(variableName, instanceUUID, activityInstanceUUID));
          runtime.setProcessInstanceVariable(instanceUUID, variableName, newValue);
        }
      }
    }
  }

  private static String getDataTypeClassName(String variableName, ProcessInstanceUUID instanceUUID, ActivityInstanceUUID activityUUID) {
    StandardQueryAPIAccessorImpl accessor = new StandardQueryAPIAccessorImpl();
    QueryDefinitionAPI queryDefinitionAPI = accessor.getQueryDefinitionAPI(AccessorUtil.QUERYLIST_JOURNAL_KEY);
    try {
      DataFieldDefinition data = queryDefinitionAPI.getActivityDataField(activityUUID.getActivityDefinitionUUID(), variableName);
      return data.getDataTypeClassName();
    } catch(Exception e) {
      try {
        DataFieldDefinition data = queryDefinitionAPI.getProcessDataField(instanceUUID.getProcessDefinitionUUID(), variableName);
        return data.getDataTypeClassName();
      } catch (Exception a) {
        throw new BonitaRuntimeException(a.getMessage(), a);
      }
    }
  }

  private static Map<String, Object[]> getOuputs(Map<String, Object[]> connectorParameters) {
    Map<String, Object[]> outputs = new HashMap<String, Object[]>();
    for (Entry<String, Object[]> parameter : connectorParameters.entrySet()) {
      String methodName = parameter.getKey();
      if (!Misc.isSetter(methodName)) {
        outputs.put(methodName, parameter.getValue());
      }
    }
    return outputs;
  }

  private static Map<String, Object[]> getInputs(Map<String, Object[]> connectorParameters) {
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    for (Entry<String, Object[]> parameter : connectorParameters.entrySet()) {
      String methodName = parameter.getKey();
      if (Misc.isSetter(methodName)) {
        inputs.put(methodName, parameter.getValue());
      }
    }
    return inputs;
  }

  private static Setter getSetter(List<Setter> inputs, String parameterName) {
    if (inputs != null) {
      for (Setter setter : inputs) {
        if (setter.getSetterName().equals(parameterName)) {
          return setter;
        }
      }
    }
    return null;
  }

  private static Object convertIfPossible(Object variableValue, String dataTypeClassName) {
    try {
      return Misc.convertIfPossible("", variableValue, dataTypeClassName);
    } catch (Exception e) {
      return variableValue;
    }
  }

  private static void setParameters(Map<String, Object[]> parameters, Object connector, ActivityInstanceUUID activityInstanceUUID, ProcessInstanceUUID processInstanceUUID, Map<String, Object> extraParameters) throws BonitaException {
    if (parameters != null) {
      for (Entry<String, Object[]> parameter : parameters.entrySet()) {
        String methodName = parameter.getKey();
        Object[] methodParameters = evaluateParametersWithGroovy(parameter.getValue(), activityInstanceUUID, processInstanceUUID, extraParameters);

        Setter setter = null;
        if (connector instanceof Connector) {
          List<Setter> setters = ((Connector)connector).getSetters();
          if (setters != null) {
            setter = getSetter(setters, methodName);
            if (setter != null) {
              Object[] setterParameters = setter.getParameters();
              for (int i = 0; i < setterParameters.length; i++) {
                Object setterParameter = setterParameters[i];
                Class<?> setterParameterClass = setterParameter.getClass();
                if (setterParameterClass != null && methodParameters[i] != null &&
                    !setterParameterClass.equals(methodParameters[i].getClass())) {
                  Class<?> methodParameterClass = methodParameters[i].getClass();
                  if (methodParameterClass.equals(BigDecimal.class)
                      || methodParameterClass.equals(BigInteger.class)) {
                    methodParameters[i] = convertIfPossible(methodParameters[i].toString(), setterParameterClass.getName());
                  } else if (methodParameterClass.equals(String.class)) {
                    methodParameters[i] = convertIfPossible(methodParameters[i], setterParameterClass.getName());
                  }
                }
              }
            }
          }
        }

        Object[] setterParameters = null;
        if (setter != null) {
          setterParameters = setter.getParameters();
        }
        Class<?>[] parameterClasses = new Class[methodParameters.length];
        for (int i = 0; i < parameterClasses.length; i++) {
          if (methodParameters[i] == null) {
            parameterClasses[i] = setterParameters[i].getClass();
          } else {
            parameterClasses[i] = methodParameters[i].getClass();
          }
        }

        try {
          Method m = Connector.getMethod(connector.getClass(), methodName, parameterClasses);
          if (m == null) {
            StringBuilder builder = new StringBuilder();
            if (parameterClasses != null) {
              builder.append("[");
              for (Class<?> class1 : parameterClasses) {
                builder.append(class1).append(",");
              }
              builder.insert(builder.length() - 1, "").append("]");
            }
            throw new BonitaRuntimeException(
                "Unable to find a method with name: " + methodName
                + " and parameters: " + builder.toString()
                + " in connector: " + connector.getClass());
          }
          m.invoke(connector, methodParameters);
        } catch (Exception e) {
          throw new BonitaRuntimeException(e.getMessage(), e);
        }
      }
    }
  }

  @SuppressWarnings("unchecked")
  private static Object[] evaluateParametersWithGroovy(Object[] parameters, ActivityInstanceUUID activityInstanceUUID, ProcessInstanceUUID processInstanceUUID, Map<String, Object> extraParameters) {
    final Object[] methodParameters = parameters;
    for (int i = 0; i < parameters.length; i++) {
      final Object methodParameter = methodParameters[i];
      if (methodParameter instanceof List<?>) {
        List<Object> temp = (List<Object>) methodParameter;
        for (int j = 0; j < temp.size(); j++) {
          Object tempRow = temp.get(j);
          if (tempRow instanceof List<?>) {
            List<Object> row = (List<Object>) tempRow;
            for (int k = 0; k < row.size(); k++) {
              row.set(k, getEvaluatedExpression(row.get(k), activityInstanceUUID, processInstanceUUID, extraParameters));
            }
            temp.set(j, row);
          } else {
            temp.set(j, getEvaluatedExpression(temp.get(j), activityInstanceUUID, processInstanceUUID, extraParameters));
          }
        }
        methodParameters[i] = temp;
      } else {
        methodParameters[i] = getEvaluatedExpression(methodParameter, activityInstanceUUID, processInstanceUUID, extraParameters);
      }
    }
    return methodParameters;
  }

  private static Object getEvaluatedExpression(Object parameterMethod, ActivityInstanceUUID activityUUID, ProcessInstanceUUID processUUID, Map<String, Object> extraParameters) {
    if (parameterMethod instanceof String) {
      String expression = (String) parameterMethod;
      if (GroovyExpression.isGroovyExpression(expression)) {
        try {
          StandardAPIAccessorImpl accessor = new StandardAPIAccessorImpl();
          RuntimeAPI runtime = accessor.getRuntimeAPI();
          if (activityUUID != null) {
            return runtime.evaluateGroovyExpression(expression, activityUUID, extraParameters, true);
          } else if (processUUID != null) {
            return runtime.evaluateGroovyExpression(expression, processUUID, extraParameters, true);
          } else {
            return GroovyUtil.evaluate(expression, extraParameters);
          }
        } catch (Exception e) {
          throw new BonitaRuntimeException(e.getMessage(), e);
        }
      }
    }
    return parameterMethod;
  }

  public static void executeConnector(final Execution execution, final String activityName, final ConnectorDefinition connector, final Map<String, Object> parameters) {
    if (connector != null) {
      final InternalProcessInstance instance = execution.getInstance();
      final ProcessInstanceUUID instanceUUID = instance.getUUID();
      final ActivityInstanceUUID activityInstanceUUID = execution.getActivityInstanceUUID();
      final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
      final boolean inTransaction = connector.isThrowingException();
      try {
        final Object hookInstance = ClassDataLoader.getInstance(Hook.class, instance.getProcessDefinitionUUID(), connector);
        if (LOG.isLoggable(Level.FINE)) {
          if (activityInstanceUUID != null) {
            LOG.fine("Starting connector (instance=" + instanceUUID + ", process=" + processUUID + ", activityId=" + activityName + ") : " + connector);
          } else {
            LOG.fine("Starting connector (instance=" + instanceUUID + ", process=" + processUUID + ") : "  + connector);
          }
        }
        if (activityInstanceUUID == null) {
          executeConnector((TxHook)hookInstance, connector.getParameters(), instanceUUID, null, parameters);
        } else {
          final ActivityInstance activityInst = EnvTool.getJournalQueriers().getActivityInstance(activityInstanceUUID);
          if (hookInstance instanceof Hook) {
            final Hook hook = (Hook) hookInstance;
            hook.execute(new StandardQueryAPIAccessorImpl(), activityInst);
          } else if (hookInstance instanceof TxHook) {
            final TxHook txHook = (TxHook) hookInstance;
            executeConnector(txHook, connector.getParameters(), instanceUUID, activityInst, parameters);
          } else {
            String message = ExceptionManager.getInstance().getFullMessage("bsi_HEI_2", Hook.class.getName(), TxHook.class.getName());
            throw new BonitaRuntimeException(message);
          }
          if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Finished connector (instance=" + instanceUUID + ", process=" + processUUID + ", activityId=" + activityName + ") : " + connector);
          }
        }
      } catch (final Exception e) {
        String message = e == null ? "null" : e.getMessage();
        if (inTransaction) {
          throw new BonitaWrapperException(new HookInvocationException("bsi_HEI_3", connector.getClassName() + " because " + message, e));
        } else if (LOG.isLoggable(Level.SEVERE)) {
          LOG.severe("Exception caught while executing connector (instance="
              + instanceUUID + ", process=" + processUUID + ", activityId="
              + activityName + ") : " + connector.getClassName() + " - Exception : "
              + message + " " + Misc.getStackTraceFrom(e));
        }
      }
    }
  }

  public static void executeConnectors(final ActivityDefinition activityDef, final Execution execution, final Event event, final Map<String, Object> parameters) {
    List<HookDefinition> hooks = activityDef.getConnectors(); 
    if (hooks != null) {
      for (final HookDefinition hook : hooks) {
        if (hook.getEvent() != null && hook.getEvent().equals(event)) {
          executeConnector(execution, activityDef.getName(), hook, parameters);
        }
      }
    }
  }

  public static void executeConnector(final Execution execution, final String activityName, final ConnectorDefinition connector) {
    executeConnector(execution, activityName, connector, null);
  }

  public static void executeConnectors(final Execution execution, final Event event) {
    final InternalProcessInstance instance = execution.getInstance();
    final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
    ProcessDefinition definition = EnvTool.getJournalQueriers().getProcess(processUUID);
    List<HookDefinition> connectors = definition.getConnectors();
    if (connectors != null) {
      for(HookDefinition connector : connectors) {
        if (connector.getEvent() != null && event.equals(connector.getEvent())) {
          executeConnector(execution, null, connector);
        }
      }
    }
  }

  private static Map<String, Object[]> formatParameters(Map<String, Object[]> parameters) {
    Map<String, Object[]> formattedParameters = new HashMap<String, Object[]>();
    if (parameters != null) {
      for (Entry<String, Object[]> parameter : parameters.entrySet()) {
        String parameterName = parameter.getKey();
        if (!parameterName.startsWith("set")) {
          StringBuilder builder = new StringBuilder("set");
          builder.append(String.valueOf(parameterName.charAt(0)).toUpperCase());
          builder.append(parameterName.substring(1));
          parameterName = builder.toString();
        }
        formattedParameters.put(parameterName, parameter.getValue());
      }
    }
    return formattedParameters;
  }

  public static Map<String, Object> executeConnector(final Connector connector, Map<String, Object[]> parameters) throws Exception {
    setParameters(formatParameters(parameters), connector, null, null, null);
    connector.execute();
    Map<String, Object> results = getGetterValues(connector);
    return results;
  }

  public static Set<String> executeFilter(final Filter filter, Map<String, Object[]> parameters, Set<String> members) throws Exception {
    Misc.checkArgsNotNull(members);
    if (parameters == null) {
      parameters = new HashMap<String, Object[]>();
    }
    parameters.put("setMembers", new Object[] {members});
    setParameters(formatParameters(parameters), filter, null, null, null);
    filter.execute();
    return filter.getCandidates();
  }

  public static Set<String> executeRoleResolver(final RoleResolver resolver, Map<String, Object[]> parameters) throws Exception {
    setParameters(formatParameters(parameters), resolver, null, null, null);
    return resolver.searchMembers(null, null, "test");
  }

  public static void executeConnectors(final ActivityDefinition activityDef, final Execution execution, final Event event) {
    executeConnectors(activityDef, execution, event, null);
  }

  public static MultiInstantiatorDescriptor executeMultiInstantiator(final Execution execution, final String activityId,
      final MultiInstantiator actInstantiator, final Map<String, Object[]> parameters) throws Exception {
    ProcessInstanceUUID instanceUUID = execution.getInstance().getUUID();
    if (parameters != null) {
      Map<String, Object[]> inputs = getInputs(parameters);
      setParameters(inputs, actInstantiator, null, instanceUUID, null);
    }
    return actInstantiator.execute(new StandardQueryAPIAccessorImpl(), instanceUUID, activityId, execution.getIterationId());
  }

  public static Set<String> executeRoleMapper(final RoleMapper roleMapper, final ProcessInstanceUUID instanceUUID,
      final String roleId, Map<String, Object[]> parameters) throws Exception {
    if (parameters != null) {
      Map<String, Object[]> inputs = getInputs(parameters);
      setParameters(inputs, roleMapper, null, instanceUUID, null);
    }
    return roleMapper.searchMembers(new StandardQueryAPIAccessorImpl(), instanceUUID, roleId);
  }

  public static Set<String> executeFilter(final Filter filter, final PerformerAssign performerAssign,
      final ActivityInstance activityInstance, final Set<String> candidates,
      final Map<String, Object[]> parameters) throws Exception {
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    if (filter instanceof PerformerAssignFilter) {
      inputs.put("setPerformerAssign", new Object[] {performerAssign});
      if (parameters != null) {
        setParameters(parameters, performerAssign, activityInstance.getUUID(), null, null);
      }
      inputs.put("setClassName", new Object[] {performerAssign.getClass().getName()});
    } else if (parameters != null) {
      inputs = getInputs(parameters);
    }
    inputs.put("setMembers", new Object[] {candidates});
    inputs.put("setActivityInstanceUUID", new Object[] {activityInstance.getUUID()});
    inputs.put("setProcessInstanceUUID", new Object[] {activityInstance.getProcessInstanceUUID()});
    inputs.put("setProcessDefinitionUUID", new Object[] {activityInstance.getProcessDefinitionUUID()});
    if (!inputs.isEmpty()) {
      setParameters(inputs, filter, activityInstance.getUUID(), null, null);
    }
    filter.execute();
    return filter.getCandidates();
  }

  @SuppressWarnings("unchecked")
  public static List<Map<String, Object>> executeMultipleInstancesInstantiatior(final MultiInstantiationDefinition instantiator, 
      final ProcessInstanceUUID instanceUUID, final String activityName, final String iterationId) throws Exception {
    ProcessDefinitionUUID definitionUUID = instanceUUID.getProcessDefinitionUUID();

    MultipleInstancesInstantiator activityinstantiator =
      ClassDataLoader.getInstance(MultipleInstancesInstantiator.class, definitionUUID, instantiator);

    Map<String, Object[]> parameters = instantiator.getParameters();
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();


    if (parameters != null) {
      inputs = getInputs(parameters);
    }
    inputs.put("setActivityName", new Object[] {activityName});
    inputs.put("setIterationId", new Object[] {iterationId});
    inputs.put("setProcessInstanceUUID", new Object[] {instanceUUID});
    inputs.put("setProcessDefinitionUUID", new Object[] {definitionUUID});
    setParameters(inputs, activityinstantiator, null, instanceUUID, null);

    if (activityinstantiator.getClass().getName().equals(MultiInstantiatorInstantiator.class.getName())) {
      Method method = activityinstantiator.getClass().getMethod("getClassName", (Class[]) null);
      String className = (String) method.invoke(activityinstantiator, (Object[]) null);
      org.ow2.bonita.connector.core.MultiInstantiator multiInstantiator = (org.ow2.bonita.connector.core.MultiInstantiator)
      ClassDataLoader.getInstance(definitionUUID, className);
      method = activityinstantiator.getClass().getMethod("getInstantiatorParameters", (Class[]) null);
      Map<String, Object[]> temp = (Map<String, Object[]>) method.invoke(activityinstantiator, (Object[]) null);
      if (temp != null) {
        temp.put("setActivityId", new Object[] {activityName});
        temp.put("setIterationId", new Object[] {iterationId});
        temp.put("setProcessInstanceUUID", new Object[] {instanceUUID});
        setParameters(temp, multiInstantiator, null, instanceUUID, null);
      }
      method = activityinstantiator.getClass().getMethod("setInstantiator", new Class<?>[] {org.ow2.bonita.connector.core.MultiInstantiator.class});
      method.invoke(activityinstantiator, new Object[] {multiInstantiator});
    }

    activityinstantiator.execute();
    return activityinstantiator.getActivitiesContext();
  }

  @SuppressWarnings("unchecked")
  public static boolean executeMultipleInstancesJoinChecker(final MultiInstantiationDefinition joinChecker,
      final ActivityInstanceUUID activityUUID) throws Exception {
    ProcessInstanceUUID instanceUUID = activityUUID.getProcessInstanceUUID();
    ProcessDefinitionUUID definitionUUID = instanceUUID.getProcessDefinitionUUID();
    MultipleInstancesJoinChecker checker =
      ClassDataLoader.getInstance(MultipleInstancesJoinChecker.class, definitionUUID, joinChecker);
    Map<String, Object[]> parameters = joinChecker.getParameters();
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    if (parameters != null) {
      inputs = getInputs(parameters);
    }
    inputs.put("setActivityInstanceUUID", new Object[] {activityUUID});
    inputs.put("setActivityName", new Object[] {activityUUID.getActivityName()});
    inputs.put("setIterationId", new Object[] {activityUUID.getIterationId()});
    inputs.put("setProcessInstanceUUID", new Object[] {instanceUUID});
    inputs.put("setProcessDefinitionUUID", new Object[] {definitionUUID});
    setParameters(inputs, checker, null, instanceUUID, null);

    if (checker.getClass().getName().equals(MultiInstantiatorJoinChecker.class.getName())) {
      Method method = checker.getClass().getMethod("getClassName", (Class[]) null);
      String className = (String) method.invoke(checker, (Object[]) null);
      org.ow2.bonita.connector.core.MultiInstantiator multiInstantiator = (org.ow2.bonita.connector.core.MultiInstantiator)
      ClassDataLoader.getInstance(definitionUUID, className);
      method = checker.getClass().getMethod("getInstantiatorParameters", (Class[]) null);
      Map<String, Object[]> temp = (Map<String, Object[]>) method.invoke(checker, (Object[]) null);
      if (temp != null) {
        temp.put("setActivityId", new Object[] {activityUUID.getActivityName()});
        temp.put("setIterationId", new Object[] {activityUUID.getIterationId()});
        temp.put("setProcessInstanceUUID", new Object[] {instanceUUID});
        setParameters(temp, multiInstantiator, null, instanceUUID, null);
      }
      method = checker.getClass().getMethod("setInstantiator", new Class<?>[] {org.ow2.bonita.connector.core.MultiInstantiator.class});
      method.invoke(checker, new Object[] {multiInstantiator});
    }

    checker.execute();
    return checker.isJoinable();
  }

}
