/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.def;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.definition.activity.AutomaticActivity;
import org.ow2.bonita.definition.activity.ExternalActivity;
import org.ow2.bonita.definition.activity.ReceiveEvent;
import org.ow2.bonita.definition.activity.SendEvents;
import org.ow2.bonita.definition.activity.SubFlow;
import org.ow2.bonita.definition.activity.Task;
import org.ow2.bonita.definition.activity.Timer;
import org.ow2.bonita.facade.def.element.ConnectorDefinition;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.OutgoingEventDefinition;
import org.ow2.bonita.facade.def.element.impl.ConnectorDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.ConnectorParameters;
import org.ow2.bonita.facade.def.element.impl.OutgoingEventDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.DataFieldDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.type.Variable;
import org.ow2.bonita.util.ProcessUtil;

public class InternalProcessDefinition extends ProcessDefinitionImpl {

  private static final long serialVersionUID = 7746264410679050386L;

  protected String labelOrName;

  //mandatory for hibernate
  protected InternalProcessDefinition() {}

  public InternalProcessDefinition(ProcessDefinition process) {
    super(process);
    this.activities = new HashSet<ActivityDefinition>();
    for (ActivityDefinition activity : process.getActivities()) {
      this.activities.add(new InternalActivityDefinition(activity, createBehaviour(activity)));
    }
    if (process.getLabel() != null) {
      this.labelOrName = process.getLabel();
    } else {
      this.labelOrName = process.getName();
    }
  }
  public static InternalProcessDefinition buildFromClientProcess(ProcessDefinition process) {
    InternalProcessDefinition internalProcess = new InternalProcessDefinition(process);

    // store datafield initial values as variables to use Variable mechanism for persistence
    for (DataFieldDefinition dataField : internalProcess.getDataFields()) {
      if (dataField.getInitialValue() != null) {
        ((DataFieldDefinitionImpl)dataField).setInitialValue(ProcessUtil.createVariable(process.getUUID(), dataField.getName(), dataField.getInitialValue()));
      }
    }
    for (ActivityDefinition activity : internalProcess.getActivities()) {
      for (DataFieldDefinition dataField : activity.getDataFields()) {
        if (dataField.getInitialValue() != null) {
          ((DataFieldDefinitionImpl)dataField).setInitialValue(ProcessUtil.createVariable(process.getUUID(), dataField.getName(), dataField.getInitialValue()));
        }
      }
    }

    for (ParticipantDefinition participant : internalProcess.getParticipants()) {
      setConnectorParameters(process.getUUID(), participant.getRoleMapper());
    }
    for (ConnectorDefinition connector : internalProcess.getConnectors()) {
      setConnectorParameters(process.getUUID(), connector);
    }
    for (ActivityDefinition activity : internalProcess.getActivities()) {
      for (DeadlineDefinition deadline : activity.getDeadlines()) {
        setConnectorParameters(process.getUUID(), deadline);
      }
      for (ConnectorDefinition connector : activity.getConnectors()) {
        setConnectorParameters(process.getUUID(), connector);
      }
      if (activity.getIncomingEvent() != null) {
        for (ConnectorDefinition connector : activity.getIncomingEvent().getConnectors()) {
          setConnectorParameters(process.getUUID(), connector);
        }
      }
      for (OutgoingEventDefinition outgoing : activity.getOutgoingEvents()) {
        OutgoingEventDefinitionImpl outgoingImpl = (OutgoingEventDefinitionImpl) outgoing;
        for (Entry<String, Object> parameter : outgoing.getParameters().entrySet()) {
          String key = parameter.getKey();
          Object value = parameter.getValue();
          outgoingImpl.addParameter(key, ProcessUtil.createVariable(process.getUUID(), key, value));
        }
        outgoingImpl.removeClientParameters();
      }
      setConnectorParameters(process.getUUID(), activity.getMultiInstantiationDefinition());
      setConnectorParameters(process.getUUID(), activity.getFilter());
      
      setConnectorParameters(process.getUUID(), activity.getMultipleInstancesInstantiator());
      setConnectorParameters(process.getUUID(), activity.getMultipleInstancesJoinChecker());
    }

    return internalProcess;
  }

  public String getLabelOrName() {
    return labelOrName;
  }
  @Override
  public InternalActivityDefinition getActivity(String name) {
    return (InternalActivityDefinition) super.getActivity(name);
  }

  private static ExternalActivity createBehaviour(ActivityDefinition activity) {
    String activityName = activity.getName();
    AbstractActivity abstractActivity = null;
    if (activity.isSubflow()) {
      abstractActivity = new SubFlow(activityName);
    } else if (activity.isTimer()) {
      abstractActivity = new Timer(activityName);
    } else if (activity.isAutomatic()) {
      abstractActivity = new AutomaticActivity(activityName);
    } else if (activity.isTask()) {
      abstractActivity = new Task(activityName);
    } else if (activity.isSendEvents()) {
      abstractActivity = new SendEvents(activityName);
    } else if (activity.isReceiveEvent()) {
      abstractActivity = new ReceiveEvent(activityName);
    }

    return abstractActivity;
  }

  private static void setConnectorParameters(final ProcessDefinitionUUID processUUID, final ConnectorDefinition connector) {
    if (connector != null) {
      ConnectorDefinitionImpl connectorImpl = (ConnectorDefinitionImpl) connector;
      for (Map.Entry<String, Object[]> entries : connector.getParameters().entrySet()) {
        String key = entries.getKey();
        Object[] parameters = entries.getValue();
        List<Variable> variables = new ArrayList<Variable>();
        for (Object parameter : parameters) {
          variables.add(ProcessUtil.createVariable(processUUID, key, parameter));
        }
        connectorImpl.addParameter(key, new ConnectorParameters(key, variables));
      }
      connectorImpl.removeClientParameters();
    }
  }

  protected ClassLoader getClassLoader(ProcessDefinitionUUID processUUID) {
    return ClassDataLoader.getProcessClassLoader(processUUID);
  }

  public Map<String, InternalActivityDefinition> getInternalInitialActivities() {
    Map<String, InternalActivityDefinition> result = new HashMap<String, InternalActivityDefinition>();
    for (ActivityDefinition activity : getInitialActivities().values()) {
      result.put(activity.getName(), (InternalActivityDefinition) activity);
    }
    return result;
  }

}
