/**
 * Copyright (C) 2009  BonitaSoft S.A..
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Set;

import org.ow2.bonita.facade.exception.RoleAlreadyExistsException;
import org.ow2.bonita.facade.exception.RoleNotFoundException;
import org.ow2.bonita.facade.exception.UserAlreadyExistsException;
import org.ow2.bonita.facade.exception.UserNotFoundException;
import org.ow2.bonita.facade.identity.Role;
import org.ow2.bonita.facade.identity.User;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.facade.internal.InternalIdentityAPI;
import org.ow2.bonita.services.IdentityService;
import org.ow2.bonita.util.EnvTool;

/**
 * @author Anthony Birembaut
 *
 */
public class IdentityAPIImpl implements InternalIdentityAPI {
  
  protected IdentityAPIImpl() {
  }
  
  public Role addRole(String name) throws RemoteException, RoleAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(name);
    IdentityService identityService = EnvTool.getIdentityService();
    RoleImpl role = identityService.getRole(name);
    if (role != null) {
      throw new RoleAlreadyExistsException("bai_IAPII_1", name);
    }
    role = new RoleImpl(name);
    identityService.addRole(role);
    return role;
  }

  public Role addRole(String name, String label, String description) throws RemoteException, RoleAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(name);
    IdentityService identityService = EnvTool.getIdentityService();
    RoleImpl role = identityService.getRole(name);
    if (role != null) {
      throw new RoleAlreadyExistsException("bai_IAPII_2", name);
    }
    role = new RoleImpl(name);
    role.setLabel(label);
    role.setDescription(description);
    identityService.addRole(role);
    return role;
  }

  public void addRoleToUser(String roleName, String username) throws RemoteException, UserNotFoundException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(roleName, username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_3", username);
    }
    final RoleImpl role = identityService.getRole(roleName);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_4", roleName);
    } else {
      user.addRole(role);
    }
  }

  public void setUserRoles(String username, Set<String> roleNames) throws RemoteException, UserNotFoundException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(username, roleNames);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_8", username);
    }
    Set<String> rolesToRemove = new HashSet<String>();
    for (Role oldRole : user.getRoles()) {
      String oldRoleName = oldRole.getName();
      rolesToRemove.add(oldRoleName);
    }
    for (String roleToRemove : rolesToRemove) {
      removeRoleFromUser(roleToRemove, username);
    }
    for (String roleName : roleNames) {
      addRoleToUser(roleName, username);
    }
  }


  public User addUser(String username, String password) throws RemoteException, UserAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(username, password);
    IdentityService identityService = EnvTool.getIdentityService();
    UserImpl user = identityService.getUser(username);
    if (user != null) {
      throw new UserAlreadyExistsException("bai_IAPII_5", username);
    }
    user = new UserImpl(username, password);
    identityService.addUser(user);
    return user;
  }

  public User addUser(String username, String password, String firstName, String lastName, String email) throws RemoteException, UserAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(username, password);
    IdentityService identityService = EnvTool.getIdentityService();
    UserImpl user = identityService.getUser(username);
    if (user != null) {
      throw new UserAlreadyExistsException("bai_IAPII_6", username);
    }
    user = new UserImpl(username, password);
    user.setEmail(email);
    user.setFirstName(firstName);
    user.setLastName(lastName);
    identityService.addUser(user);
    return user;
  }

  public Set<Role> getRoles() throws RemoteException {
    IdentityService identityService = EnvTool.getIdentityService();
    final Set<RoleImpl> roles = identityService.getRoles();
    final Set<Role> result = new HashSet<Role>();
    if (roles != null) {
      for (final RoleImpl role : roles) {
        result.add(new RoleImpl(role));
      }
    }
    return result;
  }

  public Set<User> getUsers() throws RemoteException {
    IdentityService identityService = EnvTool.getIdentityService();
    final Set<UserImpl> users = identityService.getUsers();
    final Set<User> result = new HashSet<User>();
    if (users != null) {
      for (final UserImpl user : users) {
        result.add(new UserImpl(user));
      }
    }
    return result;
  }

  public Role getRole(String name) throws RemoteException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(name);
    IdentityService identityService = EnvTool.getIdentityService();
    final RoleImpl role = identityService.getRole(name);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_7", name);
    }
    return new RoleImpl(role);
  }

  public User getUser(String username) throws RemoteException, UserNotFoundException {
    FacadeUtil.checkArgsNotNull(username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_8", username);
    }
    return new UserImpl(user);
  }

  public void removeRole(String name) throws RemoteException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(name);
    IdentityService identityService = EnvTool.getIdentityService();
    final RoleImpl role = identityService.getRole(name);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_9", name);
    }
    final Set<UserImpl> users = role.getUsers();
    if (users != null) {
      for (final UserImpl user : users) {
        user.getRoles().remove(role);
      }
    }
    identityService.deleteRole(role);
  }

  public void removeUser(String username) throws RemoteException, UserNotFoundException {
    FacadeUtil.checkArgsNotNull(username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_10", username);
    }
    final Set<RoleImpl> roles = user.getRoles();
    if (roles != null) {
      for (final RoleImpl role : roles) {
        role.getUsers().remove(user);
      }
    }
    identityService.deleteUser(user);
  }

  public void removeRoleFromUser(String roleName, String username) throws RemoteException, UserNotFoundException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(roleName, username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_11", username);
    }
    final RoleImpl role = identityService.getRole(roleName);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_12", roleName);
    }
    user.removeRole(role);
  }

  public Set<Role> getUserRoles(String username) throws RemoteException, UserNotFoundException {
    FacadeUtil.checkArgsNotNull(username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(username);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_13", username);
    }
    final Set<RoleImpl> roles = user.getRoles();
    final Set<Role> result = new HashSet<Role>();
    if (roles != null) {
      for (final RoleImpl role : roles) {
        result.add(new RoleImpl(role));
      }
    }
    return result;
  }

  public Set<User> getUsersInRole(String name) throws RemoteException, RoleNotFoundException {
    FacadeUtil.checkArgsNotNull(name);
    IdentityService identityService = EnvTool.getIdentityService();
    final RoleImpl role = identityService.getRole(name);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_14", name);
    }
    final Set<UserImpl> users = role.getUsers();
    final Set<User> result = new HashSet<User>();
    if (users != null) {
      for (final UserImpl user : users) {
        result.add(new UserImpl(user));
      }
    }
    return result;
  }

  public Role updateRole(String oldName, String name, String label, String description) throws RemoteException,
      RoleNotFoundException, RoleAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(oldName, name);
    IdentityService identityService = EnvTool.getIdentityService();
    final RoleImpl role = identityService.getRole(oldName);
    if (role == null) {
      throw new RoleNotFoundException("bai_IAPII_15", oldName);
    }
    if (name != oldName && identityService.getRole(name) != null) {
      throw new RoleAlreadyExistsException("bai_IAPII_16", name);
    }
    role.setName(name);
    role.setLabel(label);
    role.setDescription(description);
    return role;
  }

  public User updateUser(String oldUsername, String username, String password, String firstName, String lastName,
      String email) throws RemoteException, UserNotFoundException, UserAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(oldUsername, username);
    IdentityService identityService = EnvTool.getIdentityService();
    final UserImpl user = identityService.getUser(oldUsername);
    if (user == null) {
      throw new UserNotFoundException("bai_IAPII_17", oldUsername);
    }
    if (username != oldUsername && identityService.getUser(username) != null) {
      throw new UserAlreadyExistsException("bai_IAPII_18", username);
    }
    user.setUsername(username);
    user.setPassword(password);
    user.setFirstName(firstName);
    user.setLastName(lastName);
    user.setEmail(email);
    return user;
  }
}
