/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte, Anthony Birembaut, Nicolas Chabanoles - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.impl;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.hibernate.HibernateException;
import org.ow2.bonita.deployment.Deployer;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.env.Environment;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.PrivilegeNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.RuleAlreadyExistsException;
import org.ow2.bonita.facade.exception.RuleNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletableProcessException;
import org.ow2.bonita.facade.exception.UserNotFoundException;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.facade.internal.InternalManagementAPI;
import org.ow2.bonita.facade.privilege.PrivilegePolicy;
import org.ow2.bonita.facade.privilege.Rule;
import org.ow2.bonita.facade.privilege.Rule.RuleType;
import org.ow2.bonita.facade.privilege.impl.ProcessReadRuleImpl;
import org.ow2.bonita.facade.privilege.impl.ProcessRuleImpl;
import org.ow2.bonita.facade.privilege.impl.ProcessStartRuleImpl;
import org.ow2.bonita.facade.privilege.impl.RuleImpl;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.uuid.AbstractUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.AuthenticationService;
import org.ow2.bonita.services.IdentityService;
import org.ow2.bonita.services.LargeDataRepository;
import org.ow2.bonita.services.PrivilegeService;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes,
 *         Pierre Vigneras
 */
public class ManagementAPIImpl implements InternalManagementAPI {

  private static final String DEFAULT_USERS_CREATED = "DEFAULT_USERS_CREATED";

  protected ManagementAPIImpl() {}

  public ProcessDefinition deploy(final BusinessArchive businessArchive)
      throws DeploymentException {
    FacadeUtil.checkArgsNotNull(businessArchive);
    try {
      return Deployer.deploy(businessArchive);
    } catch (final Exception e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void deployJar(final String jarName, final byte[] jar)
      throws DeploymentException {
    FacadeUtil.checkArgsNotNull(jarName, jar);
    if (!jarName.endsWith(".jar")) {
      throw new DeploymentException("Invalid jar name: " + jarName
          + ". A jar file name must ends with .jar extension.");
    }
    final LargeDataRepository ldr = EnvTool.getLargeDataRepository();
    if (ldr.getData(byte[].class, Misc.getGlobalClassDataCategories(), jarName) != null) {
      throw new DeploymentException("A jar with name: " + jarName
          + " already exists in repository.");
    }
    ldr.storeData(Misc.getGlobalClassDataCategories(), jarName, jar, true);
    ClassDataLoader.resetCommonClassloader();
  }

  public void removeJar(final String jarName) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(jarName);
    if (!jarName.endsWith(".jar")) {
      throw new DeploymentException("Invalid jar name: " + jarName
          + ". A jar file name must ends with .jar extension.");
    }
    final LargeDataRepository ldr = EnvTool.getLargeDataRepository();
    boolean found = ldr
        .deleteData(Misc.getGlobalClassDataCategories(), jarName);
    if (!found) {
      String message = ExceptionManager.getInstance().getFullMessage(
          "bai_MAPII_6");
      throw new DeploymentException(message, jarName);
    }
    ClassDataLoader.resetCommonClassloader();
  }

  public Set<String> getAvailableJars() throws RemoteException {
    final LargeDataRepository ldr = EnvTool.getLargeDataRepository();
    return ldr.getKeys(Misc.getGlobalClassDataCategories());
  }

  public void delete(Collection<ProcessDefinitionUUID> processUUIDs)
      throws DeploymentException, ProcessNotFoundException,
      UndeletableProcessException, UndeletableInstanceException,
      RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    for (ProcessDefinitionUUID processUUID : processUUIDs) {
      deleteProcess(processUUID);
    }
  }

  public void deleteProcess(final ProcessDefinitionUUID processUUID)
      throws ProcessNotFoundException, UndeletableProcessException,
      UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(processUUID);
    deleteProcess(true, processUUID);
  }

  private void deleteProcess(final boolean deleteAll,
      final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException,
      UndeletableProcessException, UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(processUUID);

    Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    boolean share = false;
    if (access) {
      try {
        Set<ProcessDefinitionUUID> itemsToRemove = new HashSet<ProcessDefinitionUUID>();
        itemsToRemove.add(processUUID);
        removeProcessDefinitionUUIDFromAllRules(itemsToRemove,RuleType.PROCESS_READ,RuleType.PROCESS_START);
      } catch (Exception e) {
        throw new ProcessNotFoundException(processUUID.getProcessName(),
            processUUID);
      }
    }

    if (!share) {
      final Querier journal = EnvTool.getJournalQueriers();
      final Querier history = EnvTool.getHistoryQueriers();

      InternalProcessDefinition processDef = journal.getProcess(processUUID);
      boolean inJournal = processDef != null;
      if (processDef == null) {
        processDef = history.getProcess(processUUID);
      }

      if (processDef == null) {
        throw new ProcessNotFoundException("bai_MAPII_9", processUUID);
      }

      if (deleteAll) {
        final RuntimeAPIImpl runtimeAPI = new RuntimeAPIImpl();
        try {
          runtimeAPI.deleteAllProcessInstances(processUUID);
        } catch (final ProcessNotFoundException e) {
          throw new BonitaRuntimeException(e);
        }
      }

      if (inJournal) {
        final Set<InternalProcessInstance> instances = EnvTool
            .getJournalQueriers().getProcessInstances(processUUID,
                InstanceState.STARTED);
        if (!instances.isEmpty()) {
          final ProcessInstanceUUID instanceUUID = instances.iterator().next()
              .getUUID();
          throw new UndeletableProcessException("bai_MAPII_10", processUUID,
              instanceUUID);
        }
        removeProcessDependencies(deleteAll, processDef);
        final Recorder recorder = EnvTool.getRecorder();
        recorder.remove(processDef);
      }

      if (!inJournal) {
        final Archiver archiver = EnvTool.getArchiver();
        archiver.remove(processDef);
      }
      ClassDataLoader.removeProcessClassLoader(processDef.getUUID());
      LargeDataRepository ldr = EnvTool.getLargeDataRepository();
      ldr.deleteData(Misc.getBusinessArchiveCategories(processUUID));
      ldr.deleteData(Misc.getAttachmentCategories(processUUID));
    }
  }

  private void removeProcessDefinitionUUIDFromAllRules(Set<ProcessDefinitionUUID> itemsToRemove, RuleType... ruleTypes) throws RuleNotFoundException, RemoteException, PrivilegeNotFoundException {
    FacadeUtil.checkArgsNotNull(itemsToRemove, ruleTypes);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    Set<Rule> rules = privilegeService.getRulesByType(ruleTypes);
    for (Rule rule : rules) {
      removeExceptionsFromRule(rule.getName(), itemsToRemove);
    }
  }

  private void removeProcessDependencies(final boolean checkIntegrity,
      final ProcessDefinition processDef) {
    final Querier journal = EnvTool.getJournalQueriers();
    final Set<InternalProcessInstance> instances = EnvTool.getJournalQueriers().getProcessInstances(processDef.getUUID(), InstanceState.STARTED);
    Deployer.removeStartEvents(processDef);
    if (instances != null && !instances.isEmpty()) {
      String message = ExceptionManager.getInstance().getFullMessage("bd_D_9");
      throw new DeploymentRuntimeException(message);
    }
    // check that this process is not used by another package
    // (e.g. external reference, subflow ...)
    if (checkIntegrity) {
      final Set<InternalProcessDefinition> dependentProcesses = journal.getDependentProcesses(processDef.getUUID());
      if (dependentProcesses != null && !dependentProcesses.isEmpty()) {
        // some processes depend on the process being remove. Check if this
        // process has other available versions
        String processId = processDef.getName();
        Set<InternalProcessDefinition> deployedProcessesWithThisProcessId = journal.getProcesses(processId, ProcessState.ENABLED);
        if (deployedProcessesWithThisProcessId.size() == 1) {
          String message = ExceptionManager.getInstance().getFullMessage("bd_D_10", processDef.getUUID(), dependentProcesses);
          throw new DeploymentRuntimeException(message);
        }
      }
    }
    ClassDataLoader.removeProcessClassLoader(processDef.getUUID());
  }

  public void deleteAllProcesses() throws UndeletableInstanceException,
      RemoteException, UndeletableProcessException {
    final Querier querier = EnvTool.getAllQueriers();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    Collection<InternalProcessInstance> parentInstances = new HashSet<InternalProcessInstance>();
    Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null){
        Set<ProcessDefinitionUUID> readyToRemove = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if(!readyToRemove.isEmpty()){
          processes = querier.getProcesses(readyToRemove);
          parentInstances = querier.getProcessInstances(readyToRemove);
        }
      }

    } else {
      parentInstances = querier.getParentInstances();
      processes = querier.getProcesses();
    }
    final RuntimeAPIImpl runtimeAPI = new RuntimeAPIImpl();
    for (InternalProcessInstance internalProcessInstance : parentInstances) {
      try {
        runtimeAPI.deleteProcessInstance(internalProcessInstance.getUUID());
      } catch (InstanceNotFoundException e) {
        throw new BonitaRuntimeException("Unable to delete process instance: "
            + internalProcessInstance.getUUID());
      }
    }

    for (InternalProcessDefinition process : processes) {
      try {
        deleteProcess(false, process.getUUID());
      } catch (ProcessNotFoundException e) {
        throw new BonitaRuntimeException("Unable to delete process: "
            + process.getUUID());
      }
    }
  }

  public String getLoggedUser() throws RemoteException {
    return EnvTool.getUserId();
  }

  public void addMetaData(String key, String value) throws RemoteException {
    EnvTool.getJournal().storeMetaData(key, value);
  }

  public void deleteMetaData(String key) throws RemoteException {
    EnvTool.getJournal().deleteMetaData(key);
  }

  public String getMetaData(String key) throws RemoteException {
    return EnvTool.getJournal().getMetaData(key);
  }

  public void archive(ProcessDefinitionUUID processUUID)
      throws DeploymentException {
    FacadeUtil.checkArgsNotNull(processUUID);
    try {
      Deployer.archiveProcess(processUUID, EnvTool.getUserId());
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void archive(Collection<ProcessDefinitionUUID> processUUIDs)
      throws DeploymentException, RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    for (ProcessDefinitionUUID processUUID : processUUIDs) {
      try {
        Deployer.archiveProcess(processUUID, EnvTool.getUserId());
      } catch (final DeploymentRuntimeException e) {
        throw new DeploymentException(e.getMessage(), e);
      }
    }
  }

  public void disable(ProcessDefinitionUUID processUUID)
      throws DeploymentException {
    FacadeUtil.checkArgsNotNull(processUUID);
    try {
      Deployer.disableProcess(processUUID);
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void disable(Collection<ProcessDefinitionUUID> processUUIDs)
      throws DeploymentException, RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    for (ProcessDefinitionUUID processUUID : processUUIDs) {
      try {
        Deployer.disableProcess(processUUID);
      } catch (final DeploymentRuntimeException e) {
        throw new DeploymentException(e.getMessage(), e, processUUID);
      }
    }
  }

  public void enable(ProcessDefinitionUUID processUUID)
      throws DeploymentException {
    FacadeUtil.checkArgsNotNull(processUUID);
    try {
      Deployer.enableProcess(processUUID);
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void enable(Collection<ProcessDefinitionUUID> processUUIDs)
      throws DeploymentException, RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    for (ProcessDefinitionUUID processUUID : processUUIDs) {
      enable(processUUID);
    }
  }

  public boolean isUserAdmin(String username) throws UserNotFoundException,
      RemoteException {
    AuthenticationService adminService = EnvTool.getAuthenticationService();
    return adminService.isUserAdmin(username);
  }

  public boolean checkUserCredentials(String username, String password)
      throws RemoteException {
    String defaultUsersCreated = EnvTool.getJournal().getMetaData(
        DEFAULT_USERS_CREATED);
    if (defaultUsersCreated == null) {
      IdentityService identityService = EnvTool.getIdentityService();
      RoleImpl userRole = createDefaultRole(identityService, "user", "User",
          "The user Role");
      RoleImpl adminRole = createDefaultRole(identityService, "admin", "Admin",
          "The admin Role");
      Set<RoleImpl> userRoles = new HashSet<RoleImpl>();
      userRoles.add(userRole);
      Set<RoleImpl> adminRoles = new HashSet<RoleImpl>();
      adminRoles.add(adminRole);
      addDefaultUser(identityService, "admin", "", "", "bpm", adminRoles);
      addDefaultUser(identityService, "john", "John", "Doe", "bpm", userRoles);
      addDefaultUser(identityService, "jack", "Jack", "Doe", "bpm", userRoles);
      addDefaultUser(identityService, "james", "James", "Doe", "bpm", userRoles);
      EnvTool.getJournal().storeMetaData(DEFAULT_USERS_CREATED, "true");
    }
    AuthenticationService authenticationService = EnvTool
        .getAuthenticationService();
    return authenticationService.checkUserCredentials(username, password);
  }

  private RoleImpl createDefaultRole(IdentityService identityService,
      String name, String label, String description) throws HibernateException {
    RoleImpl userRole = identityService.getRole(name);
    if (userRole == null) {
      userRole = new RoleImpl(name);
      userRole.setLabel(label);
      userRole.setDescription(description);
    }
    return userRole;
  }

  private void addDefaultUser(IdentityService identityService, String username,
      String firstName, String lastName, String password, Set<RoleImpl> roles)
      throws HibernateException {
    UserImpl existingUser = identityService.getUser(username);
    if (existingUser == null) {
      UserImpl user = new UserImpl(username, password);
      user.setFirstName(firstName);
      user.setLastName(lastName);
      for (RoleImpl role : roles) {
        RoleImpl dbRole = identityService.getRole(role.getName());
        if (dbRole != null) {
          user.addRole(dbRole);
        } else {
          user.addRole(role);
        }
      }
      identityService.addUser(user);
    } else {
      for (RoleImpl role : roles) {
        if (identityService.getRole(role.getName()) == null) {
          identityService.addRole(role);
        }
        existingUser.addRole(identityService.getRole(role.getName()));
      }
    }
  }

  public void grantAccessAuthorisation(String applicationName,
      ProcessDefinitionUUID definitionUUID) throws RemoteException {
    FacadeUtil.checkArgsNotNull(applicationName, definitionUUID);

    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    RuleImpl rule = (RuleImpl) privilegeService.getRule(applicationName);

    if (rule != null) {
      HashSet<String> processeUUIDs = new HashSet<String>();
      processeUUIDs.add(definitionUUID.getValue());
      rule.addItems(processeUUIDs);
    } else {
      HashSet<ProcessDefinitionUUID> processes = new HashSet<ProcessDefinitionUUID>();
      processes.add(definitionUUID);
      try {
        rule = (RuleImpl) createRule(applicationName, applicationName,
            applicationName, processes, RuleType.PROCESS_READ);
        Set<String> rules = new HashSet<String>();
        rules.add(rule.getName());
      } catch (RuleAlreadyExistsException e) {
        e.printStackTrace();
        throw new RemoteException();
      } 
    }
  }

  public void grantAccessAuthorisation(String applicationName,
      Set<ProcessDefinitionUUID> definitionUUIDs) throws RemoteException {
    FacadeUtil.checkArgsNotNull(applicationName, definitionUUIDs);
    for (ProcessDefinitionUUID definitionUUID : definitionUUIDs) {
      grantAccessAuthorisation(applicationName, definitionUUID);
    }
  }
  
  @SuppressWarnings("unchecked")
  public <E extends AbstractUUID> void addExceptionsToRule(String ruleName,
      Set<E> exceptions) throws RuleNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(ruleName, exceptions);
    if (exceptions.size() > 0) {
      PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Rule rule = privilegeService.getRule(ruleName);
      if (rule == null) {
        throw new RuleNotFoundException("bai_IAPII_8", ruleName);
      }
      switch (rule.getType()) {
        case PROCESS_START:
          ((ProcessStartRuleImpl) rule)
              .addProcesses((Set<ProcessDefinitionUUID>) exceptions);
          break;
        case PROCESS_READ:
          ((ProcessReadRuleImpl) rule)
              .addProcesses((Set<ProcessDefinitionUUID>) exceptions);
          break;
        default:
          throw new IllegalArgumentException();
      }
    }
  }

  public void applyRulesToEntity(String entityID, Set<String> rules)
      throws RuleNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(entityID, rules);
    if (rules.size() > 0) {
      PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Set<Rule> storedRules = privilegeService.getRules(rules);
      if(storedRules.size() < rules.size()) {
        HashSet<String> missingRulesName = new HashSet<String>(rules);
        for (Rule rule : storedRules) {
          missingRulesName.remove(rule.getName());
        }
        // The size of the missingRuleNames is strictly positive.
        throw new RuleNotFoundException("bai_IAPII_8", missingRulesName.iterator().next());
      } else {
        HashSet<String> entities = new HashSet<String>();
        entities.add(entityID);
        for (Rule rule : storedRules) {
           ((RuleImpl)rule).addEntitiess(entities);
        }
      }     
    }
  }
  
  public void removeRulesFromEntity(String entityID, Set<String> rules)
        throws RemoteException, PrivilegeNotFoundException, RuleNotFoundException {
  
    FacadeUtil.checkArgsNotNull(entityID, rules);
    if (rules.size() > 0) {
      PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Set<Rule> storedRules = privilegeService.getRules();
      if(storedRules.size() < rules.size()) {
        HashSet<String> missingRulesName = new HashSet<String>(rules);
        for (Rule rule : storedRules) {
          missingRulesName.remove(rule.getName());
        }
        // The size of the missingRuleNames is strictly positive.
        throw new RuleNotFoundException("bai_IAPII_8", missingRulesName.iterator().next());
      } else {
        HashSet<String> entities = new HashSet<String>();
        entities.add(entityID);
        for (Rule rule : storedRules) {
          ((RuleImpl)rule).removeEntities(entities);
        }
      }
    }
    }

  public void applySameRulesToOtherEntities(String entityID, Set<String> entities)
      throws PrivilegeNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(entityID, entities);
    if (entities.size() > 0) {
      PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Set<Rule> storedRules = privilegeService.getAllApplicableRules(entityID);
      
      if (storedRules == null || storedRules.isEmpty()) {
        throw new PrivilegeNotFoundException("bai_IAPII_8", entityID);
      }
      
      Set<Rule> oldRules;
      for (String entity : entities) {
        oldRules = privilegeService.getAllApplicableRules(entity);
        for (Rule rule : oldRules) {
          ((RuleImpl)rule).removeEntities(entities);
        }
      }
      
      for (Rule rule : storedRules) {
        ((RuleImpl)rule).addEntitiess(entities);
      }
    }
  }

  @SuppressWarnings("unchecked")
  public <E extends AbstractUUID> Rule createRule(String name, String label,
      String description, Set<E> exceptions, RuleType type) throws RemoteException,
      RuleAlreadyExistsException {
    FacadeUtil.checkArgsNotNull(name, exceptions, type);

    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    Rule rule = privilegeService.getRule(name);
    if (rule != null) {
      throw new RuleAlreadyExistsException("bai_IAPII_1", name);
    }
    switch (type) {
      case PROCESS_START:
        rule = new ProcessStartRuleImpl(name, label, description,
            (Set<ProcessDefinitionUUID>) exceptions);
        privilegeService.addRule(rule);
        break;
      case PROCESS_READ:
        rule = new ProcessReadRuleImpl(name, label, description,
            (Set<ProcessDefinitionUUID>) exceptions);
        privilegeService.addRule(rule);
        break;
      default:
        throw new IllegalArgumentException();
    }

    return rule;
  }

  public void deleteRule(String name) throws RemoteException,
      RuleNotFoundException {
    FacadeUtil.checkArgsNotNull(name);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    final Rule rule = privilegeService.getRule(name);
    if (rule == null) {
      throw new RuleNotFoundException("bai_IAPII_9", name);
    }

    privilegeService.deleteRule(rule);

  }

  public Rule getRule(String name) throws RuleNotFoundException,
      RemoteException {
    FacadeUtil.checkArgsNotNull(name);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    Rule rule = privilegeService.getRule(name);
    if (rule == null) {
      throw new RuleNotFoundException("bai_IAPII_8", name);
    }
    switch (rule.getType()) {
      case PROCESS_START:
        return (Rule) new ProcessStartRuleImpl((ProcessStartRuleImpl) rule);
      case PROCESS_READ:
        return (Rule) new ProcessReadRuleImpl((ProcessReadRuleImpl) rule);
      default:
        throw new IllegalArgumentException();
    }

  }

  public Set<Rule> getRules() throws RemoteException {
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    Set<Rule> rules = privilegeService.getRules();
    if (rules == null || rules.isEmpty()) {
      return Collections.emptySet();
    }
    
    return buildRulesResultSet(rules);

  }

  public Set<Rule> getRules(Set<String> ruleNames) throws RemoteException,
      RuleNotFoundException {
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    if (ruleNames.size() > 0) {
      Set<Rule> rules = privilegeService.getRules(ruleNames);
      if (ruleNames.size() != rules.size()) {
        // The request tries to get an unknown rule.
        Set<String> storedRuleNames = new HashSet<String>();
        for (Rule rule : rules) {
          storedRuleNames.add(rule.getName());
        }
        for (String ruleName : ruleNames) {
          if (!storedRuleNames.contains(ruleName)) {
            throw new RuleNotFoundException("bai_IAPII_8", ruleName);
          }
        }
      }
      return buildRulesResultSet(rules);
    } else {
      return Collections.emptySet();
    }
  }

  @SuppressWarnings("unchecked")
  public <E extends AbstractUUID> void removeExceptionsFromRule(String ruleName,
      Set<E> exceptions) throws RuleNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(ruleName, exceptions);
    if (exceptions.size() > 0) {
      PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Rule rule = privilegeService.getRule(ruleName);
      if (rule == null) {
        throw new RuleNotFoundException("bai_IAPII_8", ruleName);
      }
      switch (rule.getType()) {
        case PROCESS_START:
          ((ProcessStartRuleImpl) rule)
              .removeProcesses((Set<ProcessDefinitionUUID>) exceptions);
          break;
        case PROCESS_READ:
          ((ProcessReadRuleImpl) rule)
              .removeProcesses((Set<ProcessDefinitionUUID>) exceptions);
          break;
        default:
          throw new IllegalArgumentException("Un-managed rule type: " + rule.getType());
      }
    }
  }

  public void updateRule(String name, Rule newRule) throws RemoteException,
      RuleNotFoundException {
    FacadeUtil.checkArgsNotNull(name, newRule);

    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    Rule rule = privilegeService.getRule(name);
    if (rule == null) {
      throw new RuleNotFoundException("bai_IAPII_8", name);
    }
    switch (rule.getType()) {
      case PROCESS_START:
      case PROCESS_READ:
        ((RuleImpl) rule).setDescription(newRule.getDescription());
        ((RuleImpl) rule).setName(newRule.getName());
        ((RuleImpl) rule).setLabel(newRule.getLabel());
        ((RuleImpl) rule).setExceptions(newRule.getItems());
        break;
      default:
        throw new IllegalArgumentException();
    }

  }

  public Set<Rule> getAllApplicableRules(String entityID) {
    FacadeUtil.checkArgsNotNull(entityID);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Set<Rule> rules = privilegeService.getAllApplicableRules(entityID);
      
      return buildRulesResultSet(rules);
  }

  public Set<Rule> getApplicableRules(String entityID, RuleType ruleType) {
    FacadeUtil.checkArgsNotNull(entityID,ruleType);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
      Set<Rule> rules = privilegeService.getApplicableRules(entityID,ruleType);
      return buildRulesResultSet(rules);
  }

  private Set<Rule> buildRulesResultSet(Set<Rule> rules) {
    Set<Rule> rulesResult = new HashSet<Rule>();
    for (Rule rule : rules) {
      switch (rule.getType()) {
        case PROCESS_START:
          rulesResult.add((Rule) new ProcessStartRuleImpl(
              (ProcessStartRuleImpl) rule));
          break;
        case PROCESS_READ:
          rulesResult.add((Rule) new ProcessReadRuleImpl(
              (ProcessReadRuleImpl) rule));
          break;
        default:
          throw new IllegalArgumentException("Un-managed RuleType: " + rule.getType());
      }
    }
    return rulesResult;
  }
  
  private List<Rule> buildRulesResultList(List<Rule> rules) {
    List<Rule> rulesResult = new ArrayList<Rule>();
    for (Rule rule : rules) {
      switch (rule.getType()) {
        case PROCESS_START:
          rulesResult.add((Rule) new ProcessStartRuleImpl(
              (ProcessStartRuleImpl) rule));
          break;
        case PROCESS_READ:
          rulesResult.add((Rule) new ProcessReadRuleImpl(
              (ProcessReadRuleImpl) rule));
          break;
        default:
          throw new IllegalArgumentException("Un-managed RuleType: " + rule.getType());
      }
    }
    return rulesResult;
  }

  public PrivilegePolicy getDefaultPolicy() {
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    return privilegeService.getDefaultPolicy();
  }

  public void setDefaultPolicy(PrivilegePolicy newPolicy, boolean negateRules) {
    FacadeUtil.checkArgsNotNull(newPolicy);
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    PrivilegePolicy currentPolicy = privilegeService.getDefaultPolicy();
    if(newPolicy != currentPolicy){
      // The policy must change. 
      if(negateRules){
        // We must negate all the rules.
        Set<Rule> processRules = privilegeService.getRulesByType(RuleType.PROCESS_START,RuleType.PROCESS_READ);
        if(processRules!=null && !processRules.isEmpty()){
          Set<InternalProcessDefinition> allProcesses = EnvTool.getAllQueriers().getProcesses();
          Set<ProcessDefinitionUUID> readableProcesses = new HashSet<ProcessDefinitionUUID>();
          Set<ProcessDefinitionUUID> startableProcesses = new HashSet<ProcessDefinitionUUID>();
          for (InternalProcessDefinition processDefinition : allProcesses) {
            if(processDefinition.getState() != ProcessState.ARCHIVED){
              startableProcesses.add(processDefinition.getUUID());
            }
            // All are readable.
            readableProcesses.add(processDefinition.getUUID());
          }
          Set<ProcessDefinitionUUID> ruleProcesses;
          for (Rule rule : processRules) {
            switch (rule.getType()) {
              case PROCESS_READ:
                ruleProcesses = new HashSet<ProcessDefinitionUUID>(readableProcesses);
                ruleProcesses.removeAll(((ProcessRuleImpl)rule).getProcesses());
                ((ProcessRuleImpl)rule).setProcesses(ruleProcesses);
                break;
              case PROCESS_START:
                ruleProcesses = new HashSet<ProcessDefinitionUUID>(startableProcesses);
                ruleProcesses.removeAll(((ProcessRuleImpl)rule).getProcesses());
                ((ProcessRuleImpl)rule).setProcesses(ruleProcesses);
                break;
              default:
                break;
            }
          }
        }
      }
      // Finally update the default policy.
      privilegeService.setDefaultPolicy(newPolicy);
    }
    
  }

  public List<Rule> getRules(RuleType ruleType, int fromIndex, int pageSige)
      throws RemoteException {
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    List<Rule> rules = privilegeService.getRules(ruleType,fromIndex,pageSige);
    if (rules == null || rules.isEmpty()) {
      return Collections.emptyList();
    }
    
    return buildRulesResultList(rules);
  }

  public long getNumberOfRules(RuleType ruleType) throws RemoteException {
    PrivilegeService privilegeService = EnvTool.getPrivilegeService();
    return privilegeService.getNumberOfRules(ruleType);
  }
}
