/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.impl;

import java.io.IOException;
import java.io.InputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ow2.bonita.env.Environment;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.element.AttachmentDefinition;
import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.element.impl.BusinessArchiveImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.DataFieldDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ParticipantDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.exception.ActivityDefNotFoundException;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DataFieldNotFoundException;
import org.ow2.bonita.facade.exception.ParticipantNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.internal.InternalQueryDefinitionAPI;
import org.ow2.bonita.facade.privilege.Rule.RuleType;
import org.ow2.bonita.facade.runtime.InitialAttachment;
import org.ow2.bonita.facade.runtime.impl.InitialAttachmentImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ParticipantDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.light.LightProcessDefinition;
import org.ow2.bonita.light.impl.LightProcessDefinitionImpl;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.services.LargeDataRepository;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Misc;

public class QueryDefinitionAPIImpl implements InternalQueryDefinitionAPI {

	protected QueryDefinitionAPIImpl() {
	}
	
	private Set<ProcessDefinition> getProcessCopy(final Set<ProcessDefinition> src) {
		final Set<ProcessDefinition> result = new HashSet<ProcessDefinition>();
		if (src != null) {
			for (final ProcessDefinition p : src) {
				result.add(new ProcessDefinitionImpl(p));
			}
		}
		return result;
	}

	private Set<ActivityDefinition> getActivityCopy(final Set<ActivityDefinition> src) {
		final Set<ActivityDefinition> result = new HashSet<ActivityDefinition>();
		if (src != null) {
			for (final ActivityDefinition p : src) {
				result.add(new ActivityDefinitionImpl(p));
			}
		}
		return result;
	}

	private Set<ParticipantDefinition> getParticipantCopy(final Set<ParticipantDefinition> src) {
		final Set<ParticipantDefinition> result = new HashSet<ParticipantDefinition>();
		if (src != null) {
			for (final ParticipantDefinition p : src) {
				result.add(new ParticipantDefinitionImpl(p));
			}
		}
		return result;
	}

	public List<ProcessDefinition> getProcesses(int fromIndex, int pageSize, String queryList) throws RemoteException {
	  final List<ProcessDefinition> result = new ArrayList<ProcessDefinition>();
    if (pageSize <= 0) {
      return result;
    }

    List<InternalProcessDefinition> processes = getIndexedProcesses(fromIndex, pageSize, queryList);
    for (final ProcessDefinition record : processes) {
      result.add(new ProcessDefinitionImpl(record));
    }
    return result;
	}

	public List<LightProcessDefinition> getLightProcesses(int fromIndex, int pageSize, String queryList) throws RemoteException {
	  final List<LightProcessDefinition> result = new ArrayList<LightProcessDefinition>();
    if (pageSize <= 0) {
      return result;
    }

    List<InternalProcessDefinition> processes = getIndexedProcesses(fromIndex, pageSize, queryList);
    for (final ProcessDefinition record : processes) {
      result.add(new LightProcessDefinitionImpl(record));
    }
    return result;
	}

	private List<InternalProcessDefinition> getIndexedProcesses(int fromIndex, int pageSize, String queryList) {
	  Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    List<InternalProcessDefinition> processes = new ArrayList<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if(!visibleProcessUUIDs.isEmpty()){
          processes = EnvTool.getAllQueriers(queryList).getProcesses(visibleProcessUUIDs, fromIndex, pageSize);
        }
      }
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses(fromIndex ,pageSize);
    }
    return processes;
	}

	public int getNumberOfProcesses(String queryList) throws RemoteException {
	  Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    if (!access) {
      return EnvTool.getAllQueriers(queryList).getNumberOfProcesses();
    } else {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        return visibleProcessUUIDs.size();
      } else {
        return 0;
      }
    }
	}

	public BusinessArchive getBusinessArchive(ProcessDefinitionUUID processDefinitionUUID, final String queryList) throws ProcessNotFoundException, RemoteException {
		Misc.checkArgsNotNull(processDefinitionUUID);
		final LargeDataRepository ldr = EnvTool.getLargeDataRepository();
		final Map<String, byte[]> resources = ldr.getData(byte[].class, Misc.getBusinessArchiveCategories(processDefinitionUUID));
		final BusinessArchive businessArchive = new BusinessArchiveImpl(processDefinitionUUID, resources);
		return businessArchive;
	}

	private Set<DataFieldDefinition> getDataFieldCopy(final Set<DataFieldDefinition> src) {
		final Set<DataFieldDefinition> result = new HashSet<DataFieldDefinition>();
		if (src != null) {
			for (final DataFieldDefinition p : src) {
				result.add(new DataFieldDefinitionImpl(p));
			}
		}
		return result;
	}

	public ProcessDefinition getLastProcess(final String processId, final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processId);
		Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    ProcessDefinition last = null;
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName==null) {
        throw new ProcessNotFoundException("bai_QDAPII_2", processId);
      }
      Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
      
      if (visibleProcessUUIDs.isEmpty()) {
        throw new ProcessNotFoundException("bai_QDAPII_2", processId);
      }
      
      Set<ProcessDefinitionUUID> definitionUUIDs = new HashSet<ProcessDefinitionUUID>();
      
      for (ProcessDefinitionUUID processUUID : visibleProcessUUIDs) {
        if (processId.equals(processUUID.getProcessName())) {
          definitionUUIDs.add(processUUID);
        }
      }
      last = EnvTool.getAllQueriers(queryList).getLastDeployedProcess(definitionUUIDs, ProcessState.ENABLED);

    } else {
      last = EnvTool.getAllQueriers(queryList).getLastDeployedProcess(processId, ProcessState.ENABLED);
    }
		if (last == null) {
			throw new ProcessNotFoundException("bai_QDAPII_2", processId);
		}		
		return new ProcessDefinitionImpl(last);
	}

	public ProcessDefinition getProcess(final ProcessDefinitionUUID processUUID, 
			final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_5", processUUID);
		}
		return new ProcessDefinitionImpl(process);
	}
	
	public LightProcessDefinition getLightProcess(final ProcessDefinitionUUID processUUID, final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_5", processUUID);
		}
		return new LightProcessDefinitionImpl(process);
	}

	public ProcessDefinition getProcess(final String processId, final String processVersion, final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processId, processVersion);

		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processId, processVersion);
		if (process == null) {
			throw new ProcessNotFoundException(processId, processVersion);
		}
		Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName==null) {
        throw new ProcessNotFoundException(processId, processVersion);
      }
      Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
      if (visibleProcessUUIDs == null || !visibleProcessUUIDs.contains(process.getUUID())) {
        throw new ProcessNotFoundException(processId, processVersion);
      }
    }
    return new ProcessDefinitionImpl(process);
	}

	public Set<ActivityDefinition> getProcessActivities(final ProcessDefinitionUUID processUUID, 
			final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_7", processUUID);
		}
		final Set<ActivityDefinition> activities = process.getActivities();
		return getActivityCopy(activities);
	}

	public ActivityDefinition getProcessActivity(final ProcessDefinitionUUID processUUID, 
			final String activityId, final String queryList)
	throws ProcessNotFoundException, ActivityNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID, activityId);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_8", processUUID);
		}
		final Set<ActivityDefinition> activities = process.getActivities();
		if (activities != null) {
			for (final ActivityDefinition activityDefinition : activities) {
				if (activityDefinition.getName().equals(activityId)) {
					return new ActivityDefinitionImpl(activityDefinition);
				}
			}
		}
		throw new ActivityNotFoundException("bai_QDAPII_9", processUUID, activityId);
	}

	public ParticipantDefinition getProcessParticipant(final ProcessDefinitionUUID processUUID, 
			final String participantId, final String queryList)
	throws ProcessNotFoundException, ParticipantNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID, participantId);
		final Set<ParticipantDefinition> participants = getProcessParticipants(processUUID, queryList);
		if (participants != null) {
			for (final ParticipantDefinition participant : participants) {
				if (participant.getName().equals(participantId)) {
					return new ParticipantDefinitionImpl(participant);
				}
			}
		}
		throw new ParticipantNotFoundException("bai_QDAPII_10", participantId, processUUID);
	}

	public Set<ParticipantDefinition> getProcessParticipants(
			final ProcessDefinitionUUID processUUID, final String queryList) throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_11", processUUID);
		}
		final Set<ParticipantDefinition> participants = process.getParticipants();
		return getParticipantCopy(participants);
	}

	@SuppressWarnings("unchecked")
	public Set<ProcessDefinition> getProcesses(final String queryList) {
	  Environment env = Environment.getCurrent();
	  boolean access = env.isRestrictedApplicationAcces();
	  Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
	  if (access) {
	    String applicationName = env.getApplicationAccessName();
	    if(applicationName!=null) {
	      Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
	      if(visibleProcessUUIDs!=null && !visibleProcessUUIDs.isEmpty()){
	        processes = EnvTool.getAllQueriers(queryList).getProcesses(visibleProcessUUIDs);
	      }
	    }
	  } else {
	    processes = EnvTool.getAllQueriers(queryList).getProcesses();
	  }
	  return getProcessCopy((Set)processes);
	}

	public Set<LightProcessDefinition> getLightProcesses(String queryList) throws RemoteException {
	  Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if (visibleProcessUUIDs != null && !visibleProcessUUIDs.isEmpty()) {
          processes = EnvTool.getAllQueriers(queryList).getProcesses(visibleProcessUUIDs);
        }
      }
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses();
    }

    Set<LightProcessDefinition> result = new HashSet<LightProcessDefinition>();
		for (final ProcessDefinition p : processes) {
			result.add(new LightProcessDefinitionImpl(p));
		}
		return result;
	}

	@SuppressWarnings("unchecked")
	public Set<ProcessDefinition> getProcesses(final String processId, final String queryList) {
		FacadeUtil.checkArgsNotNull(processId);
		Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if (visibleProcessUUIDs != null && !visibleProcessUUIDs.isEmpty()) {
          Set<ProcessDefinitionUUID> definitionUUIDs = new HashSet<ProcessDefinitionUUID>();
          for (ProcessDefinitionUUID processUUID : visibleProcessUUIDs) {
            if (processId.equals(processUUID.getProcessName())) {
              definitionUUIDs.add(processUUID);
            }
          }
          processes = EnvTool.getAllQueriers(queryList).getProcesses(definitionUUIDs);
        }
      }      
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses(processId);
    }
		return getProcessCopy((Set)processes);
	}

	public ParticipantDefinitionUUID getProcessParticipantId(
			final ProcessDefinitionUUID processUUID, final String participantName, final String queryList)
	throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID, participantName);
		final Set<ParticipantDefinition> participants = getProcessParticipants(processUUID,queryList);
		if (participants != null) {
			for (final ParticipantDefinition participant : participants) {
				if (participantName.equals(participant.getName())) {
					return participant.getUUID();
				}
			}
		}
		return null;
	}

	public ActivityDefinitionUUID getProcessActivityId(final ProcessDefinitionUUID processUUID, 
			final String activityName, final String queryList)
	throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processUUID, activityName);
		final Set<ActivityDefinition> activities = getProcessActivities(processUUID, queryList);
		if (activities != null) {
			for (final ActivityDefinition activity : activities) {
				if (activityName.equals(activity.getName())) {
					return activity.getUUID();
				}
			}
		}
		return null;
	}

	@SuppressWarnings("unchecked")
	public Set<ProcessDefinition> getProcesses(final String processId, 
			final ProcessDefinition.ProcessState processState, final String queryList) {
		FacadeUtil.checkArgsNotNull(processId, processState);
    Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if (visibleProcessUUIDs != null && !visibleProcessUUIDs.isEmpty()) {
          Set<ProcessDefinitionUUID> definitionUUIDs = new HashSet<ProcessDefinitionUUID>();
          for (ProcessDefinitionUUID processUUID : visibleProcessUUIDs) {
            if (processId.equals(processUUID.getProcessName())) {
              definitionUUIDs.add(processUUID);
            }
          }
          processes = EnvTool.getAllQueriers(queryList).getProcesses(definitionUUIDs, processState);
        }
      }
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses(processId, processState);
    }
		return getProcessCopy((Set)processes);
	}

	@SuppressWarnings("unchecked")
	public Set<ProcessDefinition> getProcesses(final ProcessDefinition.ProcessState processState, final String queryList) {
		FacadeUtil.checkArgsNotNull(processState);

    Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if (visibleProcessUUIDs != null && !visibleProcessUUIDs.isEmpty()) {
          Set<ProcessDefinitionUUID> definitionUUIDs = new HashSet<ProcessDefinitionUUID>();
          for (ProcessDefinitionUUID processUUID : visibleProcessUUIDs) {
            definitionUUIDs.add(processUUID);
          }
          processes = EnvTool.getAllQueriers(queryList).getProcesses(definitionUUIDs, processState);
        }
      }
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses(processState);
    }

    return getProcessCopy((Set)processes);
	}

	public Set<LightProcessDefinition> getLightProcesses(ProcessState processState, String queryList) throws RemoteException {
		FacadeUtil.checkArgsNotNull(processState);
		Environment env = Environment.getCurrent();
    boolean access = env.isRestrictedApplicationAcces();
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    if (access) {
      String applicationName = env.getApplicationAccessName();
      if(applicationName!=null) {
        Set<ProcessDefinitionUUID> visibleProcessUUIDs = FacadeUtil.getAllowedProcessUUIDsFor(applicationName, RuleType.PROCESS_READ);
        if (visibleProcessUUIDs != null && !visibleProcessUUIDs.isEmpty()) {
          Set<ProcessDefinitionUUID> definitionUUIDs = new HashSet<ProcessDefinitionUUID>();
          for (ProcessDefinitionUUID processUUID : visibleProcessUUIDs) {
            definitionUUIDs.add(processUUID);
          }
          processes = EnvTool.getAllQueriers(queryList).getProcesses(definitionUUIDs, processState);
        }
      }
    } else {
      processes = EnvTool.getAllQueriers(queryList).getProcesses(processState);
    }

    final Set<LightProcessDefinition> result = new HashSet<LightProcessDefinition>();
		for (final ProcessDefinition p : processes) {
			result.add(new LightProcessDefinitionImpl(p));
		}
		return result;
	}
	
	public DataFieldDefinition getProcessDataField(final ProcessDefinitionUUID processDefinitionUUID, 
			final String dataFieldId, final String queryList)
	throws ProcessNotFoundException, DataFieldNotFoundException {
		FacadeUtil.checkArgsNotNull(processDefinitionUUID, dataFieldId);
		final Set<DataFieldDefinition> datafields = getProcessDataFields(processDefinitionUUID, queryList);
		if (datafields != null) {
			for (final DataFieldDefinition datefield : datafields) {
				if (datefield.getName().equals(dataFieldId)) {
					return new DataFieldDefinitionImpl(datefield);
				}
			}
		}
		throw new DataFieldNotFoundException("bai_QDAPII_14", dataFieldId, processDefinitionUUID);
	}

	public Set<DataFieldDefinition> getProcessDataFields(final ProcessDefinitionUUID processDefinitionUUID, final String queryList)
	throws ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(processDefinitionUUID);
		final ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processDefinitionUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_15", processDefinitionUUID);
		}
		final Set<DataFieldDefinition> datafields = process.getDataFields();
		return getDataFieldCopy(datafields);
	}

	public Set<DataFieldDefinition> getActivityDataFields(final ActivityDefinitionUUID activityDefinitionUUID, 
			final String queryList) throws ActivityDefNotFoundException {
		FacadeUtil.checkArgsNotNull(activityDefinitionUUID);
		final ActivityDefinition activity = EnvTool.getAllQueriers(queryList).getActivity(activityDefinitionUUID);
		if (activity == null) {
			throw new ActivityDefNotFoundException("bai_QDAPII_16", activityDefinitionUUID);
		}
		final Set<DataFieldDefinition> datafields = activity.getDataFields();
		return getDataFieldCopy(datafields);
	}

	public DataFieldDefinition getActivityDataField(final ActivityDefinitionUUID activityDefinitionUUID, 
			final String dataFieldId, final String queryList)
	throws ActivityDefNotFoundException, DataFieldNotFoundException {
		FacadeUtil.checkArgsNotNull(activityDefinitionUUID);
		final ActivityDefinition activity = EnvTool.getAllQueriers(queryList).getActivity(activityDefinitionUUID);
		if (activity == null) {
			throw new ActivityDefNotFoundException("bai_QDAPII_17", activityDefinitionUUID);
		}
		final Set<DataFieldDefinition> datafields = activity.getDataFields();
		if (datafields != null) {
			for (final DataFieldDefinition datefield : datafields) {
				if (datefield.getName().equals(dataFieldId)) {
					return new DataFieldDefinitionImpl(datefield);
				}
			}
		}
		throw new DataFieldNotFoundException("bai_QDAPII_18", dataFieldId, activityDefinitionUUID);
	}

	public String getProcessMetaData(ProcessDefinitionUUID uuid, String key, final String queryList)
	throws RemoteException, ProcessNotFoundException {
		FacadeUtil.checkArgsNotNull(uuid, key);
		ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(uuid);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_19", uuid);
		}
		return process.getAMetaData(key);
	}

	public InitialAttachment getProcessAttachment(ProcessDefinitionUUID processUUID, String attachmentName, String queryList)
	throws ProcessNotFoundException, RemoteException {
		FacadeUtil.checkArgsNotNull(processUUID, attachmentName);
		ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_19", processUUID);
		}
		AttachmentDefinition attachment = process.getAttachment(attachmentName);
		if (attachment == null) {
			return null;
		}
		LargeDataRepository ldr = EnvTool.getLargeDataRepository();
		byte[] content = null;
		if (attachment.getFilePath() != null) {
			content = ldr.getData(byte[].class, Misc.getAttachmentCategories(processUUID), attachment.getFilePath());
		}
		return new InitialAttachmentImpl(attachment, content);
	}

	public Set<InitialAttachment> getProcessAttachments(ProcessDefinitionUUID processUUID, String queryList)
	throws ProcessNotFoundException, RemoteException {
		FacadeUtil.checkArgsNotNull(processUUID);
		ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_19", processUUID);
		}
		Map<String, AttachmentDefinition> processAttachments = process.getAttachments();
		Set<InitialAttachment> attachments = new HashSet<InitialAttachment>();
		if (attachments != null) {
			LargeDataRepository ldr = EnvTool.getLargeDataRepository();
			for (AttachmentDefinition processAttachment : processAttachments.values()) {
				byte[] content = null;
				if (processAttachment.getFilePath() != null) {
					content = ldr.getData(byte[].class, Misc.getAttachmentCategories(processUUID), processAttachment.getFilePath());
				}
				attachments.add(new InitialAttachmentImpl(processAttachment, content));
			}
		}
		return attachments;
	}

	public AttachmentDefinition getAttachmentDefinition(ProcessDefinitionUUID processUUID, String attachmentName, String queryList)
	throws ProcessNotFoundException, RemoteException {
		FacadeUtil.checkArgsNotNull(processUUID, attachmentName);
		ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_19", processUUID);
		}
		return process.getAttachment(attachmentName);
	}

	public Set<AttachmentDefinition> getAttachmentDefinitions(ProcessDefinitionUUID processUUID, String queryList)
	throws ProcessNotFoundException, RemoteException {
		FacadeUtil.checkArgsNotNull(processUUID);
		ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(processUUID);
		if (process == null) {
			throw new ProcessNotFoundException("bai_QDAPII_19", processUUID);
		}
		Map<String, AttachmentDefinition> processAttachments = process.getAttachments();
		Set<AttachmentDefinition> attachments = new HashSet<AttachmentDefinition>();
		if (attachments != null) {
			for (Entry<String, AttachmentDefinition> processAttachment : processAttachments.entrySet()) {
				attachments.add(processAttachment.getValue());
			}
		}
		return attachments;
	}

	public byte[] getResource(ProcessDefinitionUUID definitionUUID, String resourcePath, String queryList)
	throws ProcessNotFoundException, RemoteException {
	  FacadeUtil.checkArgsNotNull(definitionUUID);
    ProcessDefinition process = EnvTool.getAllQueriers(queryList).getProcess(definitionUUID);
    if (process == null) {
      throw new ProcessNotFoundException("bai_QDAPII_19", definitionUUID);
    }
	  ClassLoader classLoader = ClassDataLoader.getProcessClassLoader(definitionUUID);
	  InputStream in = classLoader.getResourceAsStream(resourcePath);
	  byte[] resource = null;
	  if (in != null) {
	    try {
	      resource = Misc.getAllContentFrom(in);
	      in.close();
	    } catch (IOException e) {
	    }
	  }
	  return resource;
  }

  public Set<LightProcessDefinition> getLightProcesses(
      Set<ProcessDefinitionUUID> processUUIDs, String queryList)
      throws ProcessNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    
    Set<InternalProcessDefinition> processes = new HashSet<InternalProcessDefinition>();
    processes = EnvTool.getAllQueriers(queryList).getProcesses(processUUIDs);

    final Set<LightProcessDefinition> result = new HashSet<LightProcessDefinition>();
    for (final ProcessDefinition p : processes) {
      result.add(new LightProcessDefinitionImpl(p));
    }
    return result;
  }

  public List<LightProcessDefinition> getAllLightProcessesExcept(
      Set<ProcessDefinitionUUID> processUUIDs, int fromIndex, int pageSize, String queryList)
      throws RemoteException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    
    if(processUUIDs.isEmpty()){
      return getLightProcesses(fromIndex, pageSize, queryList);
    } else {
      List<InternalProcessDefinition> processes = new ArrayList<InternalProcessDefinition>();
      processes = EnvTool.getAllQueriers(queryList).getProcessesExcept(processUUIDs,fromIndex,pageSize);
  
      final List<LightProcessDefinition> result = new ArrayList<LightProcessDefinition>();
      for (final ProcessDefinition p : processes) {
        result.add(new LightProcessDefinitionImpl(p));
      }
      return result;
    }
  }

}
