/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.impl;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.connector.core.Connector;
import org.ow2.bonita.connector.core.Filter;
import org.ow2.bonita.connector.core.Mapper;
import org.ow2.bonita.connector.core.RoleResolver;
import org.ow2.bonita.definition.activity.ConnectorExecutor;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.BonitaInternalException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.UncancellableInstanceException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.internal.InternalRuntimeAPI;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.AttachmentInstance;
import org.ow2.bonita.facade.runtime.Comment;
import org.ow2.bonita.facade.runtime.InitialAttachment;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.impl.AttachmentInstanceImpl;
import org.ow2.bonita.facade.runtime.impl.CommentImpl;
import org.ow2.bonita.facade.runtime.impl.InternalActivityInstance;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.TaskManager;
import org.ow2.bonita.runtime.event.OutgoingEventInstance;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.EventService;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.GroovyException;
import org.ow2.bonita.util.GroovyUtil;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.ProcessUtil;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class RuntimeAPIImpl implements InternalRuntimeAPI {

  private static final Logger LOG = Logger.getLogger(RuntimeAPIImpl.class.getName());

  protected RuntimeAPIImpl() {
  }

  public void enableEventsInFailure(ActivityInstanceUUID activityUUID) {
    final EventService eventService = EnvTool.getEventService();
    eventService.enableEventsInFailureIncomingEvents(activityUUID);
  }

  public void enableEventsInFailure(ProcessInstanceUUID instanceUUID, String activityName) {
    Set<InternalActivityInstance> activities = EnvTool.getJournalQueriers().getActivityInstances(instanceUUID, activityName);
    for (InternalActivityInstance activity : activities) {
      enableEventsInFailure(activity.getUUID());
    }
  }

  /**
   * Create an instance of the specified process and return the processUUID
   */
  public ProcessInstanceUUID instantiateProcess(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    try {
      return instantiateProcess(processUUID, null, null);
    } catch (final VariableNotFoundException e) {
      //must never occur
      throw new BonitaRuntimeException(e);
    }
  }

  public ProcessInstanceUUID instantiateProcess(final ProcessDefinitionUUID processUUID, final Map<String, Object> variables)
  throws ProcessNotFoundException, VariableNotFoundException {
    return instantiateProcess(processUUID, variables, null);
  }

  public ProcessInstanceUUID instantiateProcess(final ProcessDefinitionUUID processUUID,
      final Map<String, Object> variables, Collection<InitialAttachment> attachments) throws ProcessNotFoundException, VariableNotFoundException {
    FacadeUtil.checkArgsNotNull(processUUID);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Starting a new instance of process : " + processUUID);
    }
    Execution rootExecution = ProcessUtil.createProcessInstance(processUUID, variables, attachments, null, null);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Started: " + rootExecution.getInstance());
    }
    rootExecution.getInstance().begin();
    return rootExecution.getInstance().getUUID();
  }

  public void executeTask(ActivityInstanceUUID taskUUID, boolean assignTask)
  throws TaskNotFoundException, IllegalTaskStateException, RemoteException {
    startTask(taskUUID, assignTask);
    finishTask(taskUUID, assignTask);
  }

  public void cancelProcessInstance(final ProcessInstanceUUID instanceUUID) throws InstanceNotFoundException,
  UncancellableInstanceException {
    //if this instance is a child execution, throw an exception
    FacadeUtil.checkArgsNotNull(instanceUUID);
    final InternalProcessInstance instance = FacadeUtil.getInstance(instanceUUID, null);
    if (instance == null) {
      throw new InstanceNotFoundException("bai_RAPII_1", instanceUUID);
    }
    //if this instance is a child execution, throw an exception
    if (instance == null
        || instance.getParentInstanceUUID() != null
        || !instance.getInstanceState().equals(InstanceState.STARTED)) {
      throw new UncancellableInstanceException("bai_RAPII_2", instanceUUID, instance.getParentInstanceUUID(), instance.getInstanceState());
    }
    instance.cancel();
  }

  public void cancelProcessInstances(Collection<ProcessInstanceUUID> instanceUUIDs) throws InstanceNotFoundException, UncancellableInstanceException {
    FacadeUtil.checkArgsNotNull(instanceUUIDs);
    for (ProcessInstanceUUID instanceUUID : instanceUUIDs) {
      cancelProcessInstance(instanceUUID);
    }
  }

  public void deleteProcessInstances(final Collection<ProcessInstanceUUID> instanceUUIDs) throws InstanceNotFoundException,
  UndeletableInstanceException {
    if (instanceUUIDs == null) {
      return;
    }
    for (ProcessInstanceUUID instanceUUID : instanceUUIDs) {
      deleteProcessInstance(instanceUUID);
    }
  }

  public void deleteProcessInstance(final ProcessInstanceUUID instanceUUID) throws InstanceNotFoundException,
  UndeletableInstanceException {
    //if this instance is a child execution, throw an exception
    //if this instance has children, delete them
    FacadeUtil.checkArgsNotNull(instanceUUID);

    final Querier allQueriers = EnvTool.getAllQueriers();
    final Querier journal = EnvTool.getJournalQueriers();
    final Querier history = EnvTool.getHistoryQueriers();

    InternalProcessInstance processInst = journal.getProcessInstance(instanceUUID);

    boolean inHistory = false;
    final boolean inJournal = processInst != null;
    if (!inJournal) {
      processInst = history.getProcessInstance(instanceUUID);
      inHistory = processInst != null;
    }

    if (processInst == null) {
      throw new InstanceNotFoundException("bai_RAPII_3", instanceUUID);
    }
    final ProcessInstanceUUID parentInstanceUUID = processInst.getParentInstanceUUID();
    //check that the parent instance does not exist anymore, else, throw an exception
    if (parentInstanceUUID != null && allQueriers.getProcessInstance(parentInstanceUUID) != null) {
      throw new UndeletableInstanceException("bai_RAPII_4", instanceUUID, parentInstanceUUID);
    }

    EnvTool.getLargeDataRepository().deleteData(Misc.getAttachmentCategories(instanceUUID));
    EnvTool.getUUIDService().disableProcess(processInst.getProcessDefinitionUUID());
    ProcessUtil.removeAllInstanceEvents(instanceUUID);

    if (inJournal) {
      final Recorder recorder = EnvTool.getRecorder();
      recorder.remove(processInst);
    } else if (inHistory) {
      final Archiver archiver = EnvTool.getArchiver();
      archiver.remove(processInst);
    }
    final Set<ProcessInstanceUUID> children = processInst.getChildrenInstanceUUID();
    for (final ProcessInstanceUUID child : children) {
      deleteProcessInstance(child);
    }
  }

  public void deleteAllProcessInstances(final Collection<ProcessDefinitionUUID> processUUIDs)
  throws ProcessNotFoundException, UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(processUUIDs);
    for (ProcessDefinitionUUID processUUID : processUUIDs) {
      deleteAllProcessInstances(processUUID);
    }
  }

  public void deleteAllProcessInstances(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException,
  UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(processUUID);
    final Querier querier = EnvTool.getAllQueriers();
    final ProcessDefinition process = querier.getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException("bai_RAPII_5", processUUID);
    }
    deleteAllProcessInstances(process);
  }

  public void deleteAllProcessInstances(final ProcessDefinition process) throws ProcessNotFoundException,
  UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(process);
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final Querier querier = EnvTool.getAllQueriers();
    Set<InternalProcessInstance> instances = querier.getProcessInstances(processUUID);
    for (final ProcessInstance instance : instances) {
      //deletes only parent instances
      if (instance.getParentInstanceUUID() == null) {
        try {
          deleteProcessInstance(instance.getUUID());
        } catch (final InstanceNotFoundException e) {
          String message = ExceptionManager.getInstance().getFullMessage("bai_RAPII_6");
          throw new BonitaInternalException(message, e);
        }
      }
    }
    instances = querier.getProcessInstances(processUUID);
    if (instances != null && !instances.isEmpty()) {
      final ProcessInstance first = instances.iterator().next();
      throw new UndeletableInstanceException("bai_RAPII_7", first.getUUID(), first.getParentInstanceUUID());
    }
  }

  public void startTask(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException  {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.start(taskUUID, assignTask);
  }

  public void startActivity(ActivityInstanceUUID activityUUID) throws ActivityNotFoundException, RemoteException {
    //nothing
  }

  public void finishTask(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.finish(taskUUID, assignTask);
  }

  public void suspendTask(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws TaskNotFoundException,
  IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.suspend(taskUUID, assignTask);
  }

  public void resumeTask(final ActivityInstanceUUID taskUUID, final boolean taskAssign) throws TaskNotFoundException, IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.resume(taskUUID, taskAssign);
  }

  public void assignTask(final ActivityInstanceUUID taskUUID) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.assign(taskUUID);
  }

  public void assignTask(final ActivityInstanceUUID taskUUID, final String userId) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID, userId);
    TaskManager.assign(taskUUID, userId);
  }

  public void assignTask(final ActivityInstanceUUID taskUUID, final java.util.Set<java.lang.String> candidates) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID, candidates);
    TaskManager.assign(taskUUID, candidates);
  }

  public void unassignTask(final ActivityInstanceUUID taskUUID) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    TaskManager.unAssign(taskUUID);
  }

  private void setProcessInstanceVariable(final ProcessInstance instance, final ProcessInstanceUUID instanceUUID,
      final String variableId, final Object variableValue) throws InstanceNotFoundException, VariableNotFoundException {

    if (instance == null) {
      throw new InstanceNotFoundException("bai_RAPII_20", instanceUUID);
    }

    if (!instance.getLastKnownVariableValues().containsKey(variableId)) {
      throw new VariableNotFoundException("bai_RAPII_21", instanceUUID, variableId);
    }
  }

  public void setProcessInstanceVariable(final ProcessInstanceUUID instanceUUID,
      final String variableId, final Object variableValue) throws InstanceNotFoundException, VariableNotFoundException {
    ProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(instanceUUID);
    setProcessInstanceVariable(instance, instanceUUID, variableId, variableValue);
    EnvTool.getRecorder().recordInstanceVariableUpdated(variableId, variableValue, instance.getUUID(), EnvTool.getUserId());
  }

  public void setActivityInstanceVariable(final ActivityInstanceUUID activityUUID,
      final String variableId, final Object variableValue) throws ActivityNotFoundException, VariableNotFoundException {
    ActivityInstance activity = EnvTool.getJournalQueriers().getActivityInstance(activityUUID);
    if (activity == null) {
      throw new ActivityNotFoundException("bai_RAPII_22", activityUUID);
    }
    final ProcessInstanceUUID instanceUUID = activity.getProcessInstanceUUID();
    final String activityId = activity.getActivityName();

    Execution execution = EnvTool.getJournalQueriers().getExecutionOnActivity(instanceUUID, activityUUID);
    if (execution == null) {
      throw new ActivityNotFoundException("bai_RAPII_23", activityUUID);
    }
    Execution ex = execution.getExecution(activityId);
    if (ex != null) {
      execution = ex;
    }
    final Recorder recorder = EnvTool.getRecorder();
    
    if (!activity.getLastKnownVariableValues().containsKey(variableId)) {
      throw new VariableNotFoundException("bai_RAPII_24", instanceUUID, activityId, variableId);
    }

    // local variable updated -> update only current activity
    recorder.recordActivityVariableUpdated(variableId, variableValue, execution.getActivityInstanceUUID(), EnvTool.getUserId());
  }

  public void setVariable(final ActivityInstanceUUID activityUUID,
      final String variableId, final Object variableValue) throws ActivityNotFoundException,
      VariableNotFoundException {
    try {
      setActivityInstanceVariable(activityUUID, variableId, variableValue);
    } catch (final VariableNotFoundException e) {
      final ActivityInstance activity = EnvTool.getAllQueriers().getActivityInstance(activityUUID);
      if (activity == null) {
        throw new ActivityNotFoundException("bai_RAPII_25", activityUUID);
      }
      try {
        setProcessInstanceVariable(activity.getProcessInstanceUUID(), variableId, variableValue);
      } catch (final InstanceNotFoundException e1) {
        // If activity exists, the process instance must exist too.
        Misc.unreachableStatement();
      }
    }
  }

  public void addComment(final ProcessInstanceUUID instanceUUID, final String message, final String userId)
  throws InstanceNotFoundException {
    CommentImpl comment = new CommentImpl(userId, message, instanceUUID);
    addComment(comment, instanceUUID);
  }
  
  public void addComment(final ActivityInstanceUUID activityUUID, final String message, final String userId)
  throws ActivityNotFoundException, InstanceNotFoundException {
    final ActivityInstance activity = EnvTool.getJournalQueriers().getActivityInstance(activityUUID);
    if (activity == null) {
      throw new ActivityNotFoundException("bai_RAPII_28", activityUUID);
    }
    CommentImpl comment = new CommentImpl(userId, message, activityUUID);
    addComment(comment, activityUUID.getProcessInstanceUUID());
  }

  private void addComment(Comment comment, ProcessInstanceUUID instanceUUID)
  throws InstanceNotFoundException {
    final InternalProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(instanceUUID);
    if (instance == null) {
      throw new InstanceNotFoundException("bai_RAPII_27", instanceUUID);
    }
    instance.addComment(comment);
  }

  @Deprecated
  public void addComment(final ProcessInstanceUUID instanceUUID, final ActivityInstanceUUID activityUUID,
      final String message, final String userId)
  throws InstanceNotFoundException, ActivityNotFoundException {
    final InternalProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(instanceUUID);
    if (instance == null) {
      throw new InstanceNotFoundException("bai_RAPII_27", instanceUUID);
    }
    if (activityUUID != null) {
      final ActivityInstance activity = EnvTool.getAllQueriers().getActivityInstance(activityUUID);
      if (activity == null) {
        throw new ActivityNotFoundException("bai_RAPII_28", activityUUID);
      }
    }
    CommentImpl comment;
    if (activityUUID != null) {
      comment = new CommentImpl(userId, message, activityUUID);
    } else {
      comment = new CommentImpl(userId, message, instanceUUID);
    }
    instance.addComment(comment);
  }

  public void addProcessMetaData(ProcessDefinitionUUID uuid, String key,
      String value) throws ProcessNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(uuid, key, value);
    InternalProcessDefinition process = EnvTool.getAllQueriers().getProcess(uuid);
    if (process == null) {
      throw new ProcessNotFoundException("bai_RAPII_29", uuid);
    }
    process.addAMetaData(key, value);
  }

  public void deleteProcessMetaData(ProcessDefinitionUUID uuid, String key)
  throws ProcessNotFoundException, RemoteException {
    FacadeUtil.checkArgsNotNull(uuid, key);
    InternalProcessDefinition process = EnvTool.getAllQueriers().getProcess(uuid);
    if (process == null) {
      throw new ProcessNotFoundException("bai_RAPII_29", uuid);
    }
    process.deleteAMetaData(key);
  }

  public Object evaluateGroovyExpression(String expression, ProcessInstanceUUID instanceUUID, boolean propagate)
  throws InstanceNotFoundException, GroovyException {
    return evaluateGroovyExpression(expression, instanceUUID, null, propagate);
  }

  public Object evaluateGroovyExpression(String expression, ProcessInstanceUUID instanceUUID, Map<String, Object> context, boolean propagate)
  throws InstanceNotFoundException, GroovyException {
    return GroovyUtil.evaluate(expression, context, instanceUUID, propagate);
  }

  public Object evaluateGroovyExpression(String expression, ActivityInstanceUUID activityUUID, boolean propagate)
  throws InstanceNotFoundException, ActivityNotFoundException, GroovyException {
    return evaluateGroovyExpression(expression, activityUUID, null, propagate);
  }

  public Object evaluateGroovyExpression(String expression, ActivityInstanceUUID activityUUID, Map<String, Object> context, boolean propagate)
  throws InstanceNotFoundException, ActivityNotFoundException, GroovyException {
    return GroovyUtil.evaluate(expression, context, activityUUID, propagate);
  }

  public Object evaluateGroovyExpression(String expression,
      ProcessDefinitionUUID processDefinitionUUID)
  throws ProcessNotFoundException, GroovyException, RemoteException {
    return evaluateGroovyExpression(expression, processDefinitionUUID, null);
  }

  public Object evaluateGroovyExpression(String expression,
      ProcessDefinitionUUID processDefinitionUUID, Map<String, Object> context) throws ProcessNotFoundException, GroovyException,
      RemoteException {
    return GroovyUtil.evaluate(expression, context, processDefinitionUUID, false);
  }

  public void addAttachment(ProcessInstanceUUID instanceUUID, String name, String fileName, byte[] value) throws RemoteException {
    AttachmentInstanceImpl attachmentInstance = new AttachmentInstanceImpl(name, instanceUUID, EnvTool.getUserId(), new Date());
    attachmentInstance.setFileName(fileName);
    final InternalProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(instanceUUID);
    ProcessUtil.addAttachment(instance, attachmentInstance, value);
  }

  public void addAttachment(ProcessInstanceUUID instanceUUID, String name, String label, String description, String fileName,
      Map<String, String> metadata, byte[] value) throws RemoteException {
    AttachmentInstanceImpl attachmentInstance = new AttachmentInstanceImpl(name, instanceUUID, EnvTool.getUserId(), new Date());
    attachmentInstance.setDescription(description);
    attachmentInstance.setFileName(fileName);
    attachmentInstance.setLabel(label);
    attachmentInstance.setMetaData(metadata);
    final InternalProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(instanceUUID);
    ProcessUtil.addAttachment(instance, attachmentInstance, value);
  }

  public void addAttachments(Map<AttachmentInstance, byte[]> attachments) throws RemoteException {
    if (attachments != null) {
      for (AttachmentInstance attachmentInstance : attachments.keySet()) {
        InternalProcessInstance instance = EnvTool.getJournalQueriers().getProcessInstance(attachmentInstance.getProcessInstanceUUID());
        ProcessUtil.addAttachment(instance, new AttachmentInstanceImpl(attachmentInstance), attachments.get(attachmentInstance)); 
      }
    }
  }

  public void setActivityInstancePriority(ActivityInstanceUUID activityInstanceUUID, int priority)
      throws ActivityNotFoundException {
    final ActivityInstance activity = EnvTool.getJournalQueriers().getActivityInstance(activityInstanceUUID);
    if (activity == null) {
      throw new ActivityNotFoundException("bai_RAPII_22", activityInstanceUUID);
    }
    final ProcessInstanceUUID instanceUUID = activity.getProcessInstanceUUID();
    Execution execution = EnvTool.getJournalQueriers().getExecutionOnActivity(instanceUUID, activityInstanceUUID);
    if (execution == null) {
      throw new ActivityNotFoundException("bai_RAPII_23", activityInstanceUUID);
    }
    final Recorder recorder = EnvTool.getRecorder();
    recorder.recordActivityPriorityUpdated(activityInstanceUUID, priority);
  }

  public void deleteEvents(String eventName, String toProcessName, String toActivityName, ActivityInstanceUUID actiivtyUUID) throws RemoteException {
    final EventService eventService = EnvTool.getEventService();
    Set<OutgoingEventInstance> events = eventService.getOutgoingEvents(eventName, toProcessName, toActivityName, actiivtyUUID);
    if (events != null) {
      for (OutgoingEventInstance event : events) {
        eventService.removeEvent(event);
      }
    }
  }

  public Map<String, Object> executeConnector(String connectorClassName, Map<String, Object[]> parameters, ProcessDefinitionUUID definitionUUID)
  throws RemoteException, Exception {
    return this.executeConnector(connectorClassName, parameters, definitionUUID, null);
  }

  public Map<String, Object> executeConnector(String connectorClassName, Map<String, Object[]> parameters) throws RemoteException, Exception {
    return this.executeConnector(connectorClassName, parameters, null, null);
  }

  public Map<String, Object> executeConnector(String connectorClassName, Map<String, Object[]> parameters, ClassLoader classLoader)
  throws RemoteException, Exception {
    return this.executeConnector(connectorClassName, parameters, null, classLoader);
  }

  private Map<String, Object> executeConnector(String connectorClassName, Map<String, Object[]> parameters, ProcessDefinitionUUID definitionUUID,
      ClassLoader classLoader) throws Exception {
    ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
    try {
      Connector connector = null;
      if (classLoader == null) {
        connector = (Connector) ClassDataLoader.getInstance(definitionUUID, connectorClassName);
      } else {
        Thread.currentThread().setContextClassLoader(classLoader);
        Class<?> objectClass = Class.forName(connectorClassName, true, classLoader);
        connector = (Connector) objectClass.newInstance();
      }
      if (connector instanceof Mapper) {
        throw new IllegalAccessException(connectorClassName + " is a instance of RoleResolver or Filter");
      }
      return ConnectorExecutor.executeConnector(connector, parameters);
    } finally {
      Thread.currentThread().setContextClassLoader(baseClassLoader);
    }
  }

  public Set<String> executeFilter(String connectorClassName, Map<String, Object[]> parameters, Set<String> members)
  throws RemoteException, Exception {
    return this.executeFilter(connectorClassName, parameters, members, null, null);
  }

  public Set<String> executeFilter(String connectorClassName, Map<String, Object[]> parameters, Set<String> members,
      ProcessDefinitionUUID definitionUUID) throws RemoteException, Exception {
    return this.executeFilter(connectorClassName, parameters, members, definitionUUID, null);
  }

  public Set<String> executeFilter(String connectorClassName, Map<String, Object[]> parameters, Set<String> members,
      ClassLoader classLoader) throws RemoteException, Exception {
    return this.executeFilter(connectorClassName, parameters, members, null, classLoader);
  }

  private Set<String> executeFilter(String connectorClassName, Map<String, Object[]> parameters, Set<String> members,
      ProcessDefinitionUUID definitionUUID, ClassLoader classLoader) throws Exception {
    FacadeUtil.checkArgsNotNull(members);
    ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
    try {
      Filter connector = null;
      if (classLoader == null) {
        connector = (Filter) ClassDataLoader.getInstance(definitionUUID, connectorClassName);
      } else {
        Thread.currentThread().setContextClassLoader(classLoader);
        Class<?> objectClass = Class.forName(connectorClassName, true, classLoader);
        connector = (Filter) objectClass.newInstance();
      }
      return ConnectorExecutor.executeFilter(connector, parameters, members);
    } finally {
      Thread.currentThread().setContextClassLoader(baseClassLoader);
    }
  }

  public Set<String> executeRoleResolver(String connectorClassName, Map<String, Object[]> parameters, ProcessDefinitionUUID definitionUUID)
  throws RemoteException, Exception {
    return this.executeRoleResolver(connectorClassName, parameters, definitionUUID, null);
  }

  public Set<String> executeRoleResolver(String connectorClassName, Map<String, Object[]> parameters) throws RemoteException, Exception {
    return this.executeRoleResolver(connectorClassName, parameters, null, null);
  }

  public Set<String> executeRoleResolver(String connectorClassName, Map<String, Object[]> parameters, ClassLoader classLoader)
  throws RemoteException, Exception {
    return this.executeRoleResolver(connectorClassName, parameters, null, classLoader);
  }

  private Set<String> executeRoleResolver(String connectorClassName, Map<String, Object[]> parameters, ProcessDefinitionUUID definitionUUID,
      ClassLoader classLoader) throws Exception {
    ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
    try {
      RoleResolver connector = null;
      if (classLoader == null) {
        connector = (RoleResolver) ClassDataLoader.getInstance(definitionUUID, connectorClassName);
      } else {
        Thread.currentThread().setContextClassLoader(classLoader);
        Class<?> objectClass = Class.forName(connectorClassName, true, classLoader);
        connector = (RoleResolver) objectClass.newInstance();
      }
      return ConnectorExecutor.executeRoleResolver(connector, parameters);
    } finally {
      Thread.currentThread().setContextClassLoader(baseClassLoader);
    }
  }

}
