/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.runtime.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.ow2.bonita.definition.activity.ConnectorExecutor;
import org.ow2.bonita.facade.def.InternalActivityDefinition;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.InstanceStateUpdate;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.services.handlers.FinishedInstanceHandler;
import org.ow2.bonita.type.Variable;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.ProcessUtil;
import org.ow2.bonita.util.VariableUtil;

public class InternalProcessInstance extends ProcessInstanceImpl {

  private static final long serialVersionUID = 370138202886825855L;
  protected Execution rootExecution;
  protected long userXPRead = 0;
  protected Map<String, String> transitionStates = new HashMap<String, String>();
  
  public enum TransitionState {
    /**
     * Transition is ready to be taken
     */
    READY,
    /**
     * An execution has taken this transition.
     * However, the destination node has not been executed
     * (e.g. node execution is asynchronous)
     */
    TAKEN,
    /**
     * This transition cannot be taken.
     * This has been aborted by the system (e.g. by a join XOR)
     */
    ABORTED
  }

  //mandatory for hibernate
  protected InternalProcessInstance() {
    super();
  }

  public InternalProcessInstance(final ProcessInstanceUUID instanceUUID, final InternalProcessDefinition process, final ProcessInstanceUUID rootInstanceUUID) {
    super(process.getUUID(), instanceUUID, rootInstanceUUID);
    Map<String, InternalActivityDefinition> initialActivities = process.getInternalInitialActivities();
    if (initialActivities.size() == 1) {
      InternalActivityDefinition initial = initialActivities.values().iterator().next();
      this.rootExecution = new Execution(getUUID().toString(), process, this, initial, Execution.STATE_CREATED, null);
    } else {
      // needs to create one root Execution (parent) and many other
      this.rootExecution = new Execution(getUUID().toString(), process, this, null, Execution.STATE_CREATED, null);
    }
  }

  public InternalProcessInstance(ProcessInstance src) {
    super(src);
    this.activities = new HashSet<ActivityInstance>();
    for (ActivityInstance activity : src.getActivities()) {
      this.activities.add(new InternalActivityInstance(activity));
    }
    
    for (VariableUpdate varUpdate : this.getVariableUpdates()) {
    	final Serializable value = varUpdate.getValue();
      if (value != null) {
        ((VariableUpdateImpl)varUpdate).setValue(ProcessUtil.createVariable(src.getProcessDefinitionUUID(), varUpdate.getName(), value));
      }
    }
    this.setVariables(src.getInitialVariableValues());
    //activities
    for (ActivityInstance activity : this.getActivities()) {
    	((InternalActivityInstance)activity).setVariables(VariableUtil.createVariableMap(src.getProcessDefinitionUUID(), activity.getVariablesBeforeStarted()));
    	for (VariableUpdate varUpdate : activity.getVariableUpdates()) {
      	final Serializable value = varUpdate.getValue();
        if (value != null) {
          ((VariableUpdateImpl)varUpdate).setValue(ProcessUtil.createVariable(src.getProcessDefinitionUUID(), varUpdate.getName(), value));
        }
      }
    }
    if (src.getClass().equals(InternalProcessInstance.class)) {
      final InternalProcessInstance other = (InternalProcessInstance) src;
      this.userXPRead = other.getUserXPRead();
    }
    
  }

  private void setVariables(Map<String, Object> variables) {
	  this.variables = VariableUtil.createVariableMap(getProcessDefinitionUUID(), variables);
	  this.clientVariables = null;
  }

	public void begin() {
    if (this.rootExecution.getNode() != null) {
      this.rootExecution.beginWithOneStartNode();
    } else {
      this.rootExecution.beginWithManyStartNodes();
    }
  }

  public void setParentInstanceUUID(ProcessInstanceUUID parentInstanceUUID) {
    this.parentInstanceUUID = parentInstanceUUID;
  }
  
  public void setInitialVaribales(Map<String, Variable> initialVariables) {
    this.variables = initialVariables;
  }

  public boolean isInstanceState(final InstanceState state) {
    return getInstanceState().equals(state);
  }

  public TransitionState getTransitionState(final String transitionName) {
    if (!transitionStates.containsKey(transitionName)) {
      transitionStates.put(transitionName, TransitionState.READY.toString());
    }
    return TransitionState.valueOf(transitionStates.get(transitionName));
  }

  public void setTransitionState(final String transitionName, final TransitionState state) {
    transitionStates.put(transitionName, state.toString());
  }

  public Execution getRootExecution() {
    return this.rootExecution;
  }

  @Override
  public String toString() {
    String value = "Instance " + getUUID() + "(state:" + getInstanceState();
    if (getParentInstanceUUID() != null) {
      value += ", child of " + getParentInstanceUUID();
    }
    value += ")";
    return value;
  }

  public List<Execution> getExecOnNode(final String nodeName) {
    return this.getExecOnNode(this.getRootExecution(), nodeName);
  }

  private List<Execution> getExecOnNode(
      final Execution exec,
      final String nodeName) {
    Misc.checkArgsNotNull(exec, nodeName);
    final List<Execution> res = new ArrayList<Execution>();
    if (exec.getExecutions() == null || exec.getExecutions().isEmpty()) {
      if (exec.getNode() != null && exec.getNode().getName().equals(nodeName)) {
        res.add(exec);
      }
    }
    if (exec.getExecutions() != null) {
      for (final Execution child : exec.getExecutions()) {
        res.addAll(this.getExecOnNode((Execution) child, nodeName));
      }
    }
    return res;
  }

  public void cancel() {
    ConnectorExecutor.executeConnectors(this.getRootExecution(), HookDefinition.Event.instanceOnCancel);
    
    // cancel execution
    this.getRootExecution().cancel();
    
    // record cancel
    EnvTool.getRecorder().recordInstanceCancelled(this.getUUID(), EnvTool.getUserId());
    
    ProcessUtil.removeInternalInstanceEvents(getUUID());

    // execute finished instance handler
    if (this.getParentInstanceUUID() == null) {
      this.finish();
    }
  }

  public void finish() {
    final FinishedInstanceHandler handler = EnvTool.getFinishedInstanceHandler();
    handler.handleFinishedInstance(this);
  }

  public void setInstanceState(final InstanceState newState, final String userId) {
    updateLastUpdateDate();
    InstanceState oldState = getInstanceState();
    this.state = newState;
    if (getInstanceStateUpdates() == null) {
      this.instanceStateUpdates = new ArrayList<InstanceStateUpdate>();
    }
    //add a state update
    Date newDate = new Date();
    this.getInstanceStateUpdates().add(new InstanceStateUpdateImpl(newDate, userId, oldState, newState));
    if (newState.equals(InstanceState.STARTED)) {
      startedDate = newDate.getTime();
      startedBy = userId;
    } else {
      endedDate = newDate.getTime();
      endedBy = userId;
      if (activities != null) {
        for (ActivityInstance activity : activities) {
          if (activity.getEndedDate() == null) {
            ((InternalActivityInstance)activity).setActivityState(ActivityState.ABORTED, BonitaConstants.SYSTEM_USER);
          }
        }
      }
    }
  }

  public void setUserXPRead(Date d) {
    this.userXPRead = d.getTime();
  }
  
  public long getUserXPRead() {
    return userXPRead;
  }
}
