/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Charles Souillard, Matthieu Chaffotte, Nicolas Chabanoles - BonitaSoft S.A.
 **/
package org.ow2.bonita.persistence.db;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.Hibernate;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.transform.Transformers;
import org.hibernate.type.Type;
import org.ow2.bonita.facade.def.InternalActivityDefinition;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.element.impl.MetaDataImpl;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.facade.privilege.Rule;
import org.ow2.bonita.facade.privilege.Rule.RuleType;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.Comment;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.WebTemporaryToken;
import org.ow2.bonita.facade.runtime.impl.CaseImpl;
import org.ow2.bonita.facade.runtime.impl.InternalActivityInstance;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.runtime.impl.LabelImpl;
import org.ow2.bonita.facade.runtime.impl.WebUserImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.persistence.EventDbSession;
import org.ow2.bonita.persistence.IdentityDbSession;
import org.ow2.bonita.persistence.JournalDbSession;
import org.ow2.bonita.persistence.PrivilegeDbSession;
import org.ow2.bonita.persistence.QuerierDbSession;
import org.ow2.bonita.persistence.WebDbSession;
import org.ow2.bonita.persistence.WebTokenManagementDbSession;
import org.ow2.bonita.runtime.event.EventCouple;
import org.ow2.bonita.runtime.event.EventInstance;
import org.ow2.bonita.runtime.event.IncomingEventInstance;
import org.ow2.bonita.runtime.event.OutgoingEventInstance;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.DateUtil;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.ProcessInstanceLastUpdateComparator;
import org.ow2.bonita.util.hibernate.GenericEnumUserType;

public class DbSessionImpl extends HibernateDbSession implements QuerierDbSession, JournalDbSession, WebDbSession, IdentityDbSession, EventDbSession, PrivilegeDbSession, WebTokenManagementDbSession {

  public DbSessionImpl(Session session) {
    setSession(session);
    //session.setFlushMode(FlushMode.COMMIT);
  }

  /*
   * QUERIER
   */
  private static Type definitionStateUserType =
    Hibernate.custom(GenericEnumUserType.class, new String[]{"enumClass"}, new String[]{ProcessState.class.getName()});

  private static Type activityStateUserType =
    Hibernate.custom(GenericEnumUserType.class, new String[]{"enumClass"}, new String[]{ActivityState.class.getName()});

  private static Type instanceStateUserType =
    Hibernate.custom(GenericEnumUserType.class, new String[]{"enumClass"}, new String[]{InstanceState.class.getName()});

  //private static Type instanceStateUserType =
  //  Hibernate.custom(GenericEnumUserType.class, new String[]{"enumClass"}, new String[]{InstanceState.class.getName()});

  public int getNumberOfParentProcessInstances() {
    final Query query = getSession().getNamedQuery("getNumberOfParentProcessInstances");
    query.setCacheable(true);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfProcessInstances() {
    final Query query = getSession().getNamedQuery("getNumberOfProcessInstances");
    query.setCacheable(true);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfProcesses() {
    final Query query = getSession().getNamedQuery("getNumberOfProcesses");
    query.setCacheable(true);
    return ((Long) query.uniqueResult()).intValue();
  }

  @SuppressWarnings("unchecked")
  public Set<TaskInstance> getUserInstanceTasks(String userId,
      ProcessInstanceUUID instanceUUID, ActivityState taskState) {
    Set<TaskInstance> result = new HashSet<TaskInstance>();
    Query query = getSession().getNamedQuery("getUserInstanceTasksWithState");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setString("instanceUUID", instanceUUID.toString());
    query.setParameter("state", taskState, activityStateUserType);
    result.addAll(query.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<TaskInstance> getUserTasks(String userId, Collection<ActivityState> taskStates) {
    Query query = getSession().getNamedQuery("getUserTasksWithStates");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setParameterList("states", taskStates, activityStateUserType);
    Set<TaskInstance> result = new HashSet<TaskInstance>();
    result.addAll(query.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<InternalActivityInstance> getActivityInstances(ProcessInstanceUUID instanceUUID, String activityName) {
    Query query = getSession().getNamedQuery("getActivityInstancesWithName");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    query.setString("name", activityName);
    Set<InternalActivityInstance> result = new HashSet<InternalActivityInstance>();
    result.addAll(query.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public List<InternalActivityInstance> getActivityInstancesFromRoot(ProcessInstanceUUID rootInstanceUUID) {
    if (rootInstanceUUID == null) {
      return Collections.emptyList();
    }
    Query query = getSession().getNamedQuery("getActivityInstancesFromRoot");
    query.setCacheable(true);
    query.setString("rootInstanceUUID", rootInstanceUUID.getValue());
    return (List<InternalActivityInstance>) query.list();
  }

  @SuppressWarnings("unchecked")
  public List<InternalActivityInstance> getActivityInstancesFromRoot(Set<ProcessInstanceUUID> rootInstanceUUIDs) {
    if (rootInstanceUUIDs == null || rootInstanceUUIDs.isEmpty()) {
      return Collections.emptyList();
    }
    Collection<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID processInstanceUUID : rootInstanceUUIDs) {
      uuids.add(processInstanceUUID.getValue());
    }

    Query query = getSession().getNamedQuery("getMatchingActivityInstancesFromRoot");
    query.setCacheable(true);
    query.setParameterList("rootInstanceUUIDs", uuids);
    return (List<InternalActivityInstance>) query.list();
  }

  @SuppressWarnings("unchecked")
public Map<ProcessInstanceUUID, InternalActivityInstance> getLastUpdatedActivityInstanceFromRoot(Set<ProcessInstanceUUID> rootInstanceUUIDs, boolean considerSystemTaks) {
    if (rootInstanceUUIDs == null || rootInstanceUUIDs.isEmpty()) {
      return Collections.emptyMap();
    }

    Map<ProcessInstanceUUID, InternalActivityInstance> result = new HashMap<ProcessInstanceUUID, InternalActivityInstance>();
    Query query;
    for (ProcessInstanceUUID processInstanceUUID : rootInstanceUUIDs) {
	    if(considerSystemTaks){
	    	query = getSession().getNamedQuery("getMatchingActivityInstancesFromRoot");
	    } else {
	    	query = getSession().getNamedQuery("getMatchingHumanTaskInstancesFromRoot");
	    }
  
	    query.setCacheable(true);
	    query.setParameterList("rootInstanceUUIDs", Arrays.asList(processInstanceUUID.getValue()));
	    query.setMaxResults(1);
    
	    List<InternalActivityInstance> tmp = (List<InternalActivityInstance>)query.list(); 
	    if(tmp!=null && tmp.size()>0){
	    	result.put(processInstanceUUID,(InternalActivityInstance)tmp.get(0));    
	    }
    }
    

    return result;
  }

  @SuppressWarnings("unchecked")
  public List<InternalActivityInstance> getActivityInstancesFromRoot(Set<ProcessInstanceUUID> rootInstanceUUIDs, ActivityState state) {
    if (rootInstanceUUIDs == null || rootInstanceUUIDs.isEmpty()) {
      return Collections.emptyList();
    }
    Collection<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID processInstanceUUID : rootInstanceUUIDs) {
      uuids.add(processInstanceUUID.getValue());
    }

    Query query = getSession().getNamedQuery("getMatchingActivityInstancesWithStateFromRoot");
    query.setCacheable(true);
    query.setParameterList("rootInstanceUUIDs", uuids);
    query.setParameter("state", state, activityStateUserType);
    return (List<InternalActivityInstance>) query.list();
  }

  public TaskInstance getOneTask(String userId, ActivityState taskState) {
    Query query = getSession().getNamedQuery("getOneTask");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setParameter("state", taskState, activityStateUserType);
    query.setMaxResults(1);
    return (TaskInstance) query.uniqueResult();
  }

  public TaskInstance getOneTask(String userId, ProcessDefinitionUUID processUUID, ActivityState taskState) {
    Query query = getSession().getNamedQuery("getOneTaskOfProcess");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setString("processUUID", processUUID.getValue());
    query.setParameter("state", taskState, activityStateUserType);
    query.setMaxResults(1);
    return (TaskInstance) query.uniqueResult();
  }

  public TaskInstance getOneTask(String userId, ProcessInstanceUUID instanceUUID, ActivityState taskState) {
    Query query = getSession().getNamedQuery("getOneTaskOfInstance");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setString("instanceUUID", instanceUUID.getValue());
    query.setParameter("state", taskState, activityStateUserType);
    query.setMaxResults(1);
    return (TaskInstance) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getUserInstances(String userId) {
    Query query = getSession().getNamedQuery("getUserInstances");
    query.setCacheable(true);
    query.setString("userId", userId);
    Set<InternalProcessInstance> result = new HashSet<InternalProcessInstance>();
    result.addAll(query.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getUserInstances(String userId, Date minStartDate) {
    Query query = getSession().getNamedQuery("getUserInstancesAfterDate");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setLong("minStartDate", minStartDate.getTime());
    Set<InternalProcessInstance> result = new HashSet<InternalProcessInstance>();
    result.addAll(query.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getUserParentInstances(String userId, Date minStartDate) {
    Query query = getSession().getNamedQuery("getUserParentInstancesAfterDate");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setLong("minStartDate", minStartDate.getTime());
    Set<InternalProcessInstance> result = new HashSet<InternalProcessInstance>();
    result.addAll(query.list());
    return result;
  }

  public Set<InternalProcessInstance> getUserInstancesExcept(String userId, Set<ProcessInstanceUUID> instances) {
    if (instances == null || instances.isEmpty()) {
      return getUserInstances(userId);
    }
    Collection<String> uuids = new HashSet<String>();
    if (instances != null) {
      for (ProcessInstanceUUID processInstanceUUID : instances) {
        uuids.add(processInstanceUUID.getValue());
      }
    }
    Query query = getSession().getNamedQuery("getUserInstancesExcept");
    query.setCacheable(true);
    query.setString("userId", userId);
    return executeSplittedQuery(InternalProcessInstance.class, query, "uuids", uuids);
  }

  @SuppressWarnings("unchecked")
  public Set<InternalActivityInstance> getActivityInstances(ProcessInstanceUUID instanceUUID) {
    Query query = getSession().getNamedQuery("findActivityInstances");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.toString());
    List<InternalActivityInstance> results = query.list();
    if (results != null) {
      return new HashSet<InternalActivityInstance>(results);
    }
    return null;
  }

  public long getLastProcessInstanceNb(ProcessDefinitionUUID processUUID) {
    Query query = getSession().getNamedQuery("getLastProcessInstanceNb");
    query.setCacheable(true);
    query.setString("processUUID", processUUID.getValue());
    query.setMaxResults(1);
    Long result = (Long) query.uniqueResult();
    if (result == null) {
      return -1;
    }
    return result;
  }

  public InternalProcessInstance getProcessInstance(ProcessInstanceUUID instanceUUID) {
    Query query = getSession().getNamedQuery("findProcessInstance");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.toString());
    query.setMaxResults(1);
    return (InternalProcessInstance) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstances() {
    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    final Query query = getSession().getNamedQuery("findAllProcessInstances");
    query.setCacheable(true);
    final List<InternalProcessInstance> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  @SuppressWarnings("unchecked")
  public List<InternalProcessInstance> getProcessInstances(int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("findAllProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    final List<InternalProcessInstance> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public List<InternalProcessInstance> getMostRecentProcessInstances(int maxResults, long time) {
    final Query query = getSession().getNamedQuery("getMostRecentProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(0);
    query.setLong("time", time);
    query.setMaxResults(maxResults);
    final List<InternalProcessInstance> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public List<InternalProcessInstance> getMostRecentParentProcessInstances(int maxResults, long time) {
    final Query query = getSession().getNamedQuery("getMostRecentParentProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(0);
    query.setLong("time", time);
    query.setMaxResults(maxResults);
    final List<InternalProcessInstance> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  public List<InternalProcessInstance> getMostRecentMatchingProcessInstances(Collection<ProcessInstanceUUID> instanceUUIDs, int maxResults, long time) {
    final Query query = getSession().getNamedQuery("getMostRecentMatchingProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(0);
    query.setLong("time", time);
    query.setMaxResults(maxResults);

    Set<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID uuid : instanceUUIDs) {
      uuids.add(uuid.toString());
    }

    List<InternalProcessInstance> allInstances = new ArrayList<InternalProcessInstance>();
    allInstances.addAll(executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids));
    Collections.sort(allInstances, new ProcessInstanceLastUpdateComparator());
    List<InternalProcessInstance> results = Misc.subList(InternalProcessInstance.class, allInstances, 0, maxResults);
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  public List<InternalProcessInstance> getMostRecentProcessesProcessInstances(Collection<ProcessDefinitionUUID> definitionUUIDs, int maxResults, long time) {
    final Query query = getSession().getNamedQuery("getMostRecentProcessesProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(0);
    query.setLong("time", time);
    query.setMaxResults(maxResults);

    Set<String> uuids = new HashSet<String>();
    if (definitionUUIDs != null) {
      for (ProcessDefinitionUUID uuid : definitionUUIDs) {
        uuids.add(uuid.toString());
      }
    }

    List<InternalProcessInstance> allInstances = new ArrayList<InternalProcessInstance>();
    allInstances.addAll(executeSplittedQuery(InternalProcessInstance.class, query, "processUUIDs", uuids));
    Collections.sort(allInstances, new ProcessInstanceLastUpdateComparator());
    List<InternalProcessInstance> results = Misc.subList(InternalProcessInstance.class, allInstances, 0, maxResults);
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public List<InternalProcessInstance> getParentProcessInstances(int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getParentInstances");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    final List<InternalProcessInstance> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public List<InternalProcessDefinition> getProcesses(int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getAllProcesses");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    final List<InternalProcessDefinition> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  public List<InternalProcessInstance> getProcessInstances(Collection<ProcessInstanceUUID> instanceUUIDs, int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    Set<String> uuids = new HashSet<String>();
    if (instanceUUIDs != null) {
      for (ProcessInstanceUUID uuid : instanceUUIDs) {
        uuids.add(uuid.toString());
      }
    }

    List<InternalProcessInstance> allInstances = new ArrayList<InternalProcessInstance>();
    allInstances.addAll(executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids));
    Collections.sort(allInstances, new ProcessInstanceLastUpdateComparator());
    List<InternalProcessInstance> results = Misc.subList(InternalProcessInstance.class, allInstances, fromIndex, pageSize);
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public List<ProcessInstanceUUID> getLabelsCaseUUIDs(String ownerName, Set<String> labelNames, int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getLabelsCaseUUIDs");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    query.setString("ownerName", ownerName);
    query.setParameterList("labelNames", labelNames);
    return query.list();
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getParentInstances() {
    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    final Query query = getSession().getNamedQuery("getParentInstances");
    query.setCacheable(true);
    final List<InternalProcessInstance> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessInstanceUUID> getParentInstancesUUIDs() {
    final Set<ProcessInstanceUUID> processInsts = new HashSet<ProcessInstanceUUID>();
    final Query query = getSession().getNamedQuery("getParentInstancesUUIDs");
    query.setCacheable(true);
    final List<ProcessInstanceUUID> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  public Set<InternalProcessInstance> getProcessInstances(Collection<ProcessInstanceUUID> instanceUUIDs) {
    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID uuid : instanceUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids);
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstancesWithTaskState(Collection<ActivityState> activityStates) {
    final Query getActivitiesQuery = getSession().getNamedQuery("findActivityInstancesWithTaskState");
    getActivitiesQuery.setCacheable(true);
    getActivitiesQuery.setParameterList("activityStates", activityStates, activityStateUserType);
    final List<String> uuids = getActivitiesQuery.list();

    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    if (uuids.isEmpty()) {
      return processInsts;
    }

    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    return executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids);
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstancesWithInstanceStates(Collection<InstanceState> instanceStates) {
    final Query getProcessInstancesQuery = getSession().getNamedQuery("findProcessInstancesWithInstanceStates");
    getProcessInstancesQuery.setCacheable(true);
    getProcessInstancesQuery.setParameterList("instanceStates", instanceStates, definitionStateUserType);
    final List<String> uuids = getProcessInstancesQuery.list();

    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    if (uuids.isEmpty()) {
      return processInsts;
    }

    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    return executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids);
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstances(final ProcessDefinitionUUID processUUID) {
    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    final Query query = getSession().getNamedQuery("findProcessInstances");
    query.setCacheable(true);
    query.setString("processUUID", processUUID.toString());
    final List<InternalProcessInstance> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstances(ProcessDefinitionUUID processUUID, InstanceState instanceState) {
    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    final Query query = getSession().getNamedQuery("findProcessInstancesWithState");
    query.setCacheable(true);
    query.setString("processUUID", processUUID.toString());
    query.setParameter("state", instanceState, instanceStateUserType);
    final List<InternalProcessInstance> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  public TaskInstance getTaskInstance(ActivityInstanceUUID taskUUID) {
    Query query = getSession().getNamedQuery("findTaskInstance");
    query.setCacheable(true);
    query.setString("taskUUID", taskUUID.toString());
    query.setMaxResults(1);
    return (TaskInstance) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<TaskInstance> getTaskInstances(final ProcessInstanceUUID instanceUUID) {
    final Query query = getSession().getNamedQuery("findTaskInstances");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.toString());
    final Set<TaskInstance> taskInstances = new HashSet<TaskInstance>();
    final List<TaskInstance> results = query.list();
    if (results != null) {
      taskInstances.addAll(results);
    }
    return taskInstances;
  }

  /*
   * DEFINITION
   */
  @SuppressWarnings("unchecked")
  private Set<InternalProcessDefinition> getProcessSet(final Query query) {
    final List<InternalProcessDefinition> results = query.list();
    if (results != null) {
      return new HashSet<InternalProcessDefinition>(results);
    }
    return null;
  }

  public Set<InternalProcessDefinition> getProcesses() {
    final Query query = getSession().getNamedQuery("getAllProcesses");
    query.setCacheable(true);
    return getProcessSet(query);
  }

  public Set<InternalProcessDefinition> getProcesses(String processId) {
    final Query query = getSession().getNamedQuery("getProcesses2");
    query.setCacheable(true);
    query.setString("processId", processId);
    return getProcessSet(query);
  }

  public InternalProcessDefinition getProcess(ProcessDefinitionUUID processUUID) {
    final Query query = getSession().getNamedQuery("getProcess");
    query.setCacheable(true);
    query.setString("processUUID", processUUID.toString());
    query.setMaxResults(1);
    return (InternalProcessDefinition) query.uniqueResult();
  }

  public InternalProcessDefinition getProcess(String processId, String version) {
    final Query query = getSession().getNamedQuery("getProcessFromIdAndVersion");
    query.setCacheable(true);
    query.setString("processId", processId);
    query.setString("version", version);
    return (InternalProcessDefinition) query.uniqueResult();
  }

  public Set<InternalProcessDefinition> getProcesses(ProcessState processState) {
    Query query = getSession().getNamedQuery("getProcessesFromState");
    query.setParameter("state", processState, definitionStateUserType);
    query.setCacheable(true);
    return getProcessSet(query);
  }

  public Set<InternalProcessDefinition> getProcesses(String processId, ProcessState processState) {
    Query query = getSession().getNamedQuery("getProcessesFromProcessIdAndState");
    query.setParameter("state", processState, definitionStateUserType);
    query.setCacheable(true);
    query.setString("processId", processId);
    return getProcessSet(query);
  }

  public String getLastProcessVersion(String processName) {
    Query query = getSession().getNamedQuery("getLastProcessVersion");
    query.setCacheable(true);
    query.setString("name", processName);
    query.setMaxResults(1);
    return (String) query.uniqueResult();
  }

  public InternalProcessDefinition getLastProcess(String processId, ProcessState processState) {
    Query query = getSession().getNamedQuery("getLastProcessFromProcessIdAndState");
    query.setParameter("state", processState, definitionStateUserType);
    query.setCacheable(true);
    query.setString("processId", processId);
    query.setMaxResults(1);
    return (InternalProcessDefinition) query.uniqueResult();
  }

  public Set<InternalProcessDefinition> getDependentProcesses(ProcessDefinitionUUID processUUID) {
    Query query = getSession().getNamedQuery("getDependentProcesses");
    query.setCacheable(true);
    query.setString("processName", processUUID.getProcessName());
    query.setString("processUUID", processUUID.getValue());
    return getProcessSet(query);
  }

  public InternalActivityInstance getActivityInstance(
      ProcessInstanceUUID instanceUUID, String activityId, String iterationId, String activityInstanceId) {
    Query query = getSession().getNamedQuery("getActivityInstance");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.toString());
    query.setString("activityId", activityId);
    query.setString("iterationId", iterationId);
    query.setString("activityInstanceId", activityInstanceId);
    return (InternalActivityInstance)query.uniqueResult();
  }

  public InternalActivityInstance getActivityInstance(
      ActivityInstanceUUID activityInstanceUUID) {
    Query query = getSession().getNamedQuery("getActivityInstanceFromUUID");
    query.setCacheable(true);
    query.setString("activityUUID", activityInstanceUUID.toString());
    return (InternalActivityInstance)query.uniqueResult();
  }

  public ActivityState getActivityInstanceState(ActivityInstanceUUID activityInstanceUUID) {
    Query query = getSession().getNamedQuery("getActivityInstanceStateFromUUID");
    query.setCacheable(true);
    query.setString("activityUUID", activityInstanceUUID.toString());
    return (ActivityState)query.uniqueResult();
  }

  public InternalActivityDefinition getActivityDefinition(ActivityDefinitionUUID activityDefinitionUUID) {
    final Query query = getSession().getNamedQuery("getActivityDefinition");
    query.setCacheable(true);
    query.setString("activityUUID", activityDefinitionUUID.toString());
    query.setMaxResults(1);
    return (InternalActivityDefinition) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getUpdatedWebProcessInstances(String user, int fromIndex, int pageSize) {
    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    final Query query = getSession().getNamedQuery("getUpdatedWebProcessInstances");
    query.setString("user", user);
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    final List<InternalProcessInstance> results = query.list();
    if (results != null) {
      processInsts.addAll(results);
    }
    return processInsts;
  }

  public Execution getExecutionWithEventUUID(final String eventUUID) {
    final Query query = getSession().getNamedQuery("getExecutionWithEventUUID");
    query.setCacheable(true);
    query.setString("eventUUID", eventUUID);
    return (Execution) query.uniqueResult();
  }

  public Execution getExecutionPointingOnNode(final ActivityInstanceUUID activityUUID) {
    final Query query = getSession().getNamedQuery("findInstanceExecutionPointingOnNode");
    query.setCacheable(true);
    query.setString("activityUUID", activityUUID.toString());
    return (Execution) query.uniqueResult();
  }

  /*
   * RUNTIME
   */
  public MetaDataImpl getMetaData(String key) {
    final Query query = getSession().getNamedQuery("getMetaData");
    query.setCacheable(true);
    query.setString("key", key);
    return (MetaDataImpl) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<ProcessInstanceUUID> getAllCases() {
    Set<ProcessInstanceUUID> result = new HashSet<ProcessInstanceUUID>();
    Query query = getSession().getNamedQuery("getAllCases");
    query.setCacheable(true);
    result.addAll(query.list());
    return result;
  }

  public Set<CaseImpl> getCases(Set<ProcessInstanceUUID> caseUUIDs) {
    Collection<String> uuids = new HashSet<String>();
    if (caseUUIDs != null) {
      for (ProcessInstanceUUID processInstanceUUID : caseUUIDs) {
        uuids.add(processInstanceUUID.getValue());
      }
    }
    Query query = getSession().getNamedQuery("getMatchingCases");
    query.setCacheable(true);
    return executeSplittedQuery(CaseImpl.class, query, "uuids", uuids);
  }

  @SuppressWarnings("unchecked")
  public List<LabelImpl> getUserCustomLabels(final String ownerName) {

    List<LabelImpl> result = new ArrayList<LabelImpl>();
    Query query = getSession().getNamedQuery("getUserCustomLabels");
    query.setCacheable(true);
    query.setString("ownerName", ownerName);
    result.addAll(query.list());

    return result;
  }

  @SuppressWarnings("unchecked")
  public List<LabelImpl> getSystemLabels(final String ownerName) {

    List<LabelImpl> result = new ArrayList<LabelImpl>();
    Query query = getSession().getNamedQuery("getSystemLabels");
    query.setString("ownerName", ownerName);
    query.setCacheable(true);
    List<LabelImpl> theTempResult = query.list();
    result.addAll(theTempResult);

    return result;
  }

  public LabelImpl getLabel(final String ownerName, final String labelName) {
    final Query query = getSession().getNamedQuery("getLabelByID");
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setCacheable(true);
    query.setMaxResults(1);
    LabelImpl theResult = (LabelImpl) query.uniqueResult();

    return theResult;
  }

  public WebUserImpl getUser(String userName) {
    final Query query = getSession().getNamedQuery("getUser");
    query.setString("name", userName);
    query.setCacheable(true);
    query.setMaxResults(1);
    WebUserImpl theResult = (WebUserImpl) query.uniqueResult();

    return theResult;
  }

  @SuppressWarnings("unchecked")
  public Set<LabelImpl> getLabels(final String ownerName) {
    Set<LabelImpl> result = new HashSet<LabelImpl>();
    Query query = getSession().getNamedQuery("getAllLabels");
    query.setString("ownerName", ownerName);
    query.setCacheable(true);
    List<LabelImpl> theTempResult = query.list();
    result.addAll(theTempResult);

    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<LabelImpl> getCaseLabels(String ownerName, ProcessInstanceUUID case_) {
    //select case.labelName from case where case.ownerName = :ownerName and case.uuid = :caseUUID
    //select label from label where label.ownerName = :ownerName and label.labelName in (:labelNames)

    Query getUserCases = getSession().getNamedQuery("getUserCases");
    getUserCases.setString("ownerName", ownerName);
    getUserCases.setString("caseId", case_.getValue());
    getUserCases.setCacheable(true);
    List<String> caseLabelsNames = getUserCases.list();

    Set<LabelImpl> result = new HashSet<LabelImpl>();
    if (!caseLabelsNames.isEmpty()) {
      Query query = getSession().getNamedQuery("getLabels");
      query.setString("ownerName", ownerName);
      query.setCacheable(true);
      result = executeSplittedQuery(LabelImpl.class, query, "labelNames", caseLabelsNames);
    }
    return result;
  }

  @SuppressWarnings("unchecked")
  private <T> Set<T> executeSplittedQuery(Class<T> clazz, final Query query, final String parameterName, final Collection<? extends Object> values) {
    if (values == null || values.isEmpty()) {
      return Collections.emptySet();
    }
    final Set<T> result = new HashSet<T>();
    if (values.size() <= BonitaConstants.MAX_QUERY_SIZE) {
      query.setParameterList(parameterName, values);
      List<T> l = query.list();
      result.addAll(l);
      return result;
    }

    List<Collection<Object>> newValues = Misc.splitCollection(values, BonitaConstants.MAX_QUERY_SIZE);
    for (Collection<Object> set : newValues) {
      query.setParameterList(parameterName, set);
      List<T> l = query.list();
      result.addAll(l);
    }
    return result;
  }

  private <T> List<T> executeSplittedQueryList(Class<T> clazz, final Query query, final String parameterName, final Collection<? extends Object> values) {
    return executeSplittedQueryList(clazz, query, parameterName, values, BonitaConstants.MAX_QUERY_SIZE);
  }

  @SuppressWarnings("unchecked")
  private <T> List<T> executeSplittedQueryList(Class<T> clazz, final Query query, final String parameterName, final Collection<? extends Object> values, final int size) {
    if (values == null || values.isEmpty()) {
      return Collections.emptyList();
    }
    final List<T> result = new ArrayList<T>();
    if (values.size() <= size) {
      query.setParameterList(parameterName, values);
      return query.list();
    }

    List<Collection<Object>> newValues = Misc.splitCollection(values, size);
    for (Collection<Object> set : newValues) {
      query.setParameterList(parameterName, set);
      List<T> l = query.list();
      result.addAll(l);
    }
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<CaseImpl> getCases(String ownerName, String labelName) {
    Set<CaseImpl> result = new HashSet<CaseImpl>();
    Query query = getSession().getNamedQuery("getLabelCases");
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setCacheable(true);
    List<CaseImpl> theTempResult = (List<CaseImpl>) query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<CaseImpl> getCases(ProcessInstanceUUID case_) {
    Set<CaseImpl> result = new HashSet<CaseImpl>();
    Query query = getSession().getNamedQuery("getCases");
    query.setString("caseId", case_.getValue());
    query.setCacheable(true);
    List<CaseImpl> theTempResult = (List<CaseImpl>) query.list();
    result.addAll(theTempResult);
    return result;
  }

  public CaseImpl getCase(ProcessInstanceUUID caseUUID, String ownerName, String labelName) {
    Query query = getSession().getNamedQuery("getCase");
    query.setString("caseUUID", caseUUID.getValue());
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setCacheable(true);
    return (CaseImpl) query.uniqueResult();
  }

  public Set<LabelImpl> getLabels(String ownerName, Set<String> labelsName) {
    Query query = getSession().getNamedQuery("getLabels");
    query.setString("ownerName", ownerName);
    query.setCacheable(true);
    Set<LabelImpl> result = executeSplittedQuery(LabelImpl.class, query, "labelNames", labelsName);
    return result;
  }

  public Set<LabelImpl> getLabels(Set<String> labelsName) {
    Query query = getSession().getNamedQuery("getLabelsWithName");
    query.setCacheable(true);
    Set<LabelImpl> result = executeSplittedQuery(LabelImpl.class, query, "labelNames", labelsName);
    return result;
  }

  
  
  public int getCasesNumberWithTwoLabels(String ownerName, String label1Name, String label2Name) {
    Query query = getSession().getNamedQuery("getCasesNumberWithTwoLabels");
    query.setString("label1", CaseImpl.buildLabel(ownerName, label1Name));
    query.setString("label2", CaseImpl.buildLabel(ownerName, label2Name));
    query.setReadOnly(true);

    int result = ((Long) query.uniqueResult()).intValue();
    return result;
  }
  
  public int getCasesNumber(String ownerName, String labelName) {
    Query query = getSession().getNamedQuery("getCasesNumber");
    query.setString("ownerName", ownerName);
    query.setString("label", labelName);
    query.setReadOnly(true);
    int result = ((Long) query.uniqueResult()).intValue();
    return result;
  }
  
  @SuppressWarnings("unchecked")
  public Set<CaseImpl> getCasesWithTwoLabels(String ownerName, String label1Name, String label2Name, int limit) {
    Query query = getSession().getNamedQuery("getCasesWithTwoLabelsWithLimit");
    query.setString("label1", CaseImpl.buildLabel(ownerName, label1Name));
    query.setString("label2", CaseImpl.buildLabel(ownerName, label2Name));
    query.setReadOnly(true);
    query.setMaxResults(limit);

    final List<CaseImpl> results = query.list();
    if (results != null) {
      return new HashSet<CaseImpl>(results);
    } else {
      return new HashSet<CaseImpl>();
    }
  }
  
  @SuppressWarnings("unchecked")
  public Set<CaseImpl> getCases(String ownerName, String labelName, int limit) {
    Query query = getSession().getNamedQuery("getLabelCases");
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setMaxResults(limit);
    final List<CaseImpl> results = query.list();
    if (results != null) {
      return new HashSet<CaseImpl>(results);
    } else {
      return new HashSet<CaseImpl>();
    }
  }

  public Set<ProcessInstanceUUID> getCases(String ownerName, Set<String> theLabelsName) {
    Query query = getSession().getNamedQuery("getLabelsCases");
    query.setString("ownerName", ownerName);
    query.setCacheable(true);
    return executeSplittedQuery(ProcessInstanceUUID.class, query, "labelNames", theLabelsName);
  }

  //getAllWebCases
  //getMatchingCases

  public void deleteAllCases() {
    final Session session = getSession();
    Query query = session.getNamedQuery("getAllWebCases");
    query.setCacheable(true);

    for (Object webCase : query.list()) {
      session.delete(webCase);
    }
  }

  public void deleteCases(Set<ProcessInstanceUUID> webCases) {
    final Session session = getSession();
    Query query = session.getNamedQuery("getMatchingCases");
    query.setCacheable(true);

    Set<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID webCase : webCases) {
      uuids.add(webCase.getValue());
    }
    for (CaseImpl webCase : executeSplittedQuery(CaseImpl.class, query, "uuids", uuids)) {
      session.delete(webCase);
    }
  }

  public Set<ProcessInstanceUUID> getCasesUUIDs(String ownerName, String labelName, Set<ProcessInstanceUUID> cases) {
    Collection<String> uuids = new HashSet<String>();
    if (cases != null) {
      for (ProcessInstanceUUID processInstanceUUID : cases) {
        uuids.add(processInstanceUUID.getValue());
      }
    }

    Query query = getSession().getNamedQuery("getLabelCasesUUIDsSublist");
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setCacheable(true);
    return executeSplittedQuery(ProcessInstanceUUID.class, query, "caseUUIDs", uuids);
  }

  public Set<ProcessInstanceUUID> getLabelCases(String labelName, Set<ProcessInstanceUUID> caseUUIDs) {
    if (caseUUIDs == null || caseUUIDs.isEmpty()) {
      return Collections.emptySet();
    }
    Collection<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID processInstanceUUID : caseUUIDs) {
      uuids.add(processInstanceUUID.getValue());
    }

    Query query = getSession().getNamedQuery("getLabelNameCases");
    query.setString("labelName", labelName);
    query.setCacheable(true);
    return executeSplittedQuery(ProcessInstanceUUID.class, query, "uuids", uuids);
  }

  public Set<CaseImpl> getLabelCases(String ownerName, Set<String> labelsNames, Set<ProcessInstanceUUID> caseUUIDs) {
    if (caseUUIDs == null || caseUUIDs.isEmpty()) {
      return Collections.emptySet();
    }
    Collection<String> uuids = new HashSet<String>();
    for (ProcessInstanceUUID processInstanceUUID : caseUUIDs) {
      uuids.add(processInstanceUUID.getValue());
    }

    Query query = getSession().getNamedQuery("getLabelsNameCases");
    query.setParameterList("labelsNames", labelsNames);
    query.setString("ownerName", ownerName);
    query.setCacheable(true);
    return executeSplittedQuery(CaseImpl.class, query, "uuids", uuids);
  }

  public Set<CaseImpl> getCases(String ownerName, String labelName, Set<ProcessInstanceUUID> caseList) {
    Collection<String> uuids = new HashSet<String>();
    if (caseList != null) {
      for (ProcessInstanceUUID processInstanceUUID : caseList) {
        uuids.add(processInstanceUUID.getValue());
      }
    }

    Query query = getSession().getNamedQuery("getLabelCasesSublist");
    query.setString("ownerName", ownerName);
    query.setString("labelName", labelName);
    query.setCacheable(true);
    return executeSplittedQuery(CaseImpl.class, query, "caseUUIDs", uuids);
  }

  /*
   * IDENTITY
   */
  public RoleImpl getRole(String name) {
    final Query query = getSession().getNamedQuery("findRole");
    query.setCacheable(true);
    query.setString("name", name);
    return (RoleImpl) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<RoleImpl> getRoles() {
    final Query query = getSession().getNamedQuery("getRoles");
    query.setCacheable(true);
    final List<RoleImpl> results = query.list();
    if (results != null) {
      return new HashSet<RoleImpl>(results);
    } else {
      return new HashSet<RoleImpl>();
    }
  }

  public UserImpl getIdentityUser(String username) {
    final Query query = getSession().getNamedQuery("findIdentityUser");
    query.setCacheable(true);
    query.setString("username", username);
    return (UserImpl) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<UserImpl> getIdentityUsers() {
    final Query query = getSession().getNamedQuery("getIdentityUsers");
    query.setCacheable(true);
    final List<UserImpl> results = query.list();
    if (results != null) {
      return new HashSet<UserImpl>(results);
    } else {
      return new HashSet<UserImpl>();
    }
  }

  public List<Integer> getNumberOfExecutingCasesPerDay(Date since, Date to) {
    List<Integer> executingCases = new ArrayList<Integer>();
    while (since.before(to) || since.equals(to)) {
      Date nextDayBegining = DateUtil.getBeginningOfTheDay(DateUtil.getNextDay(since));
      final Query query = getSession().getNamedQuery("getNumberOfExecutingCases");
      query.setCacheable(true);
      query.setLong("date", nextDayBegining.getTime());
      executingCases.add(((Long) query.uniqueResult()).intValue());
      since = DateUtil.getNextDay(since);
    }
    return executingCases;
  }

  public List<Integer> getNumberOfFinishedCasesPerDay(Date since, Date to) {
    List<Integer> finishedCases = new ArrayList<Integer>();
    while (since.before(to) || since.equals(to)) {
      Date beginningOfTheDay = DateUtil.getBeginningOfTheDay(since);
      Date nextBeginningOfTheDay = DateUtil.getBeginningOfTheDay(DateUtil.getNextDay(since));
      final Query query = getSession().getNamedQuery("getNumberOfFinishedCases");
      query.setCacheable(true);
      query.setLong("beginningOfTheDay", beginningOfTheDay.getTime());
      query.setLong("nextBeginningOfTheDay", nextBeginningOfTheDay.getTime());
      finishedCases.add(((Long) query.uniqueResult()).intValue());
      since = DateUtil.getNextDay(since);
    }
    return finishedCases;
  }

  public List<Integer> getNumberOfOpenStepsPerDay(Date since, Date to) {
    List<Integer> finishedCases = new ArrayList<Integer>();
    while (since.before(to) || since.equals(to)) {
      Date nextDayBegining = DateUtil.getBeginningOfTheDay(DateUtil.getNextDay(since));
      final Query query = getSession().getNamedQuery("getNumberOfOpenSteps2");
      query.setCacheable(true);
      query.setLong("date", nextDayBegining.getTime());
      finishedCases.add(((Long) query.uniqueResult()).intValue());
      since = DateUtil.getNextDay(since);
    }
    return finishedCases;
  }

  public int getNumberOfOpenSteps() {
    final Query query = getSession().getNamedQuery("getNumberOfOpenSteps");
    query.setCacheable(true);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfOverdueSteps(Date currentDate) {
    final Query query = getSession().getNamedQuery("getNumberOfOverdueSteps");
    query.setCacheable(true);
    query.setLong("currentDate", currentDate.getTime());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfStepsAtRisk(Date currentDate, Date atRisk) {
    final Query query = getSession().getNamedQuery("getNumberOfStepsAtRisk");
    query.setCacheable(true);
    query.setLong("atRisk", atRisk.getTime());
    query.setLong("currentDate", currentDate.getTime());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfUserOpenSteps(String userId) {
    final Query query = getSession().getNamedQuery("getNumberOfUserOpenSteps");
    query.setCacheable(true);
    query.setString("userId", userId);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfUserOverdueSteps(String userId, Date currentDate) {
    final Query query = getSession().getNamedQuery("getNumberOfUserOverdueSteps");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setLong("currentDate", currentDate.getTime());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfUserStepsAtRisk(String userId, Date currentDate, Date atRisk) {
    final Query query = getSession().getNamedQuery("getNumberOfUserStepsAtRisk");
    query.setCacheable(true);
    query.setString("userId", userId);
    query.setLong("atRisk", atRisk.getTime());
    query.setLong("currentDate", currentDate.getTime());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfFinishedSteps(int priority, Date since) {
    final Query query = getSession().getNamedQuery("getNumberOfFinishedSteps");
    query.setCacheable(true);
    query.setInteger("priority", priority);
    query.setLong("since", since.getTime());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfOpenSteps(int priority) {
    final Query query = getSession().getNamedQuery("getNumberOfPriorityOpenSteps");
    query.setCacheable(true);
    query.setInteger("priority", priority);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfUserFinishedSteps(String userId, int priority, Date since) {
    final Query query = getSession().getNamedQuery("getNumberOfUserFinishedSteps");
    query.setCacheable(true);
    query.setInteger("priority", priority);
    query.setLong("since", since.getTime());
    query.setString("userId", userId);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfUserOpenSteps(String userId, int priority) {
    final Query query = getSession().getNamedQuery("getNumberOfPriorityUserOpenSteps");
    query.setCacheable(true);
    query.setInteger("priority", priority);
    query.setString("userId", userId);
    return ((Long) query.uniqueResult()).intValue();
  }

  @SuppressWarnings("unchecked")
  public Set<IncomingEventInstance> getIncomingEvents(ProcessInstanceUUID instanceUUID) {
    Set<IncomingEventInstance> result = new HashSet<IncomingEventInstance>();
    final Query query = getSession().getNamedQuery("getInstanceIncomingEvents");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    List<IncomingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<IncomingEventInstance> getIncomingEvents(ActivityDefinitionUUID activityUUID) {
    Set<IncomingEventInstance> result = new HashSet<IncomingEventInstance>();
    final Query query = getSession().getNamedQuery("getActivityDefinitionIncomingEvents");
    query.setCacheable(true);
    query.setString("activityUUID", activityUUID.getValue());
    List<IncomingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<OutgoingEventInstance> getOutgoingEvents(ProcessInstanceUUID instanceUUID) {
    Set<OutgoingEventInstance> result = new HashSet<OutgoingEventInstance>();
    final Query query = getSession().getNamedQuery("getInstanceOutgoingEvents");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    List<OutgoingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<OutgoingEventInstance> getOutgoingEvents() {
    Set<OutgoingEventInstance> result = new HashSet<OutgoingEventInstance>();
    final Query query = getSession().getNamedQuery("getOutgoingEvents");
    query.setCacheable(true);
    List<OutgoingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<OutgoingEventInstance> getOutgoingEvents(String eventName, String toProcessName, String toActivityName, ActivityInstanceUUID actiivtyUUID) {
    Set<OutgoingEventInstance> result = new HashSet<OutgoingEventInstance>();
    final Query query = getSession().getNamedQuery("getOutgoingEventsWithUUID");
    query.setString("name", eventName);
    query.setString("processName", toProcessName);
    query.setString("activityName", toActivityName);
    query.setString("activityUUID", actiivtyUUID.getValue());
    query.setCacheable(true);
    List<OutgoingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<IncomingEventInstance> getIncomingEvents(String eventName, String toProcessName, String toActivityName, ActivityInstanceUUID actiivtyUUID) {
    Set<IncomingEventInstance> result = new HashSet<IncomingEventInstance>();
    final Query query = getSession().getNamedQuery("getIncomingEventsWithUUID");
    query.setString("name", eventName);
    query.setString("processName", toProcessName);
    query.setString("activityName", toActivityName);
    query.setString("activityUUID", actiivtyUUID.getValue());
    query.setCacheable(true);
    List<IncomingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<EventInstance> getConsumedEvents() {
    final Set<EventInstance> result = new HashSet<EventInstance>();

    final Query oeiQuery = getSession().getNamedQuery("getOutgoingConsumedEvents");
    oeiQuery.setCacheable(true);
    result.addAll(oeiQuery.list());

    final Query ieiQuery = getSession().getNamedQuery("getIncomingConsumedEvents");
    ieiQuery.setCacheable(true);
    result.addAll(ieiQuery.list());
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<OutgoingEventInstance> getOverdueEvents() {
    final long currentTime = System.currentTimeMillis();

    final Set<OutgoingEventInstance> result = new HashSet<OutgoingEventInstance>();

    final Query oeiQuery = getSession().getNamedQuery("getOverdueEvents");
    oeiQuery.setLong("current", currentTime);
    oeiQuery.setCacheable(true);
    result.addAll(oeiQuery.list());

    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<IncomingEventInstance> getIncomingEvents() {
    Set<IncomingEventInstance> result = new HashSet<IncomingEventInstance>();
    final Query query = getSession().getNamedQuery("getIncomingEvents");
    query.setCacheable(true);
    List<IncomingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<IncomingEventInstance> getActivityIncomingEvents(ActivityInstanceUUID activityUUID) {
    Set<IncomingEventInstance> result = new HashSet<IncomingEventInstance>();
    final Query query = getSession().getNamedQuery("getActivityIncomingEvents");
    query.setString("activityUUID", activityUUID.getValue());
    query.setCacheable(true);
    List<IncomingEventInstance> theTempResult = query.list();
    result.addAll(theTempResult);
    return result;
  }

  @SuppressWarnings("unchecked")
  public Set<EventCouple> getEventsCouples() {
    final Query lockedCouplesQuery = getSession().getNamedQuery("getLockedEventsCouples");
    lockedCouplesQuery.setCacheable(true);
    lockedCouplesQuery.setLong("current", System.currentTimeMillis());
    lockedCouplesQuery.setResultTransformer(Transformers.aliasToBean(EventCouple.class));
    List<EventCouple> lockedCouples = lockedCouplesQuery.list();

    final Query freeCouplesQuery = getSession().getNamedQuery("getEventsCouples");
    freeCouplesQuery.setCacheable(true);
    freeCouplesQuery.setLong("current", System.currentTimeMillis());
    freeCouplesQuery.setResultTransformer(Transformers.aliasToBean(EventCouple.class));
    List<EventCouple> freeCouples = freeCouplesQuery.list();

    Set<EventCouple> result = new HashSet<EventCouple>();
    result.addAll(freeCouples);
    result.addAll(lockedCouples);
    return result;
  }

  public IncomingEventInstance getIncomingEvent(long incomingId) {
    final Query query = getSession().getNamedQuery("getIncomingEvent");
    query.setCacheable(true);
    query.setLong("id", incomingId);
    return (IncomingEventInstance) query.uniqueResult();
  }

  public OutgoingEventInstance getOutgoingEvent(long outgoingId) {
    final Query query = getSession().getNamedQuery("getOutgoingEvent");
    query.setCacheable(true);
    query.setLong("id", outgoingId);
    return (OutgoingEventInstance) query.uniqueResult();
  }

  public Long getNextDueDate() {
    final Query query = getSession().getNamedQuery("getNextDueDate");
    query.setCacheable(true);
    query.setMaxResults(1);
    Object result = query.uniqueResult();
    if (result != null) {
      return (Long) result;
    }
    return null;
  }

  public Set<InternalProcessDefinition> getProcesses(Set<ProcessDefinitionUUID> definitionUUIDs) {
    Query query = getSession().getNamedQuery("getProcessesList");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(InternalProcessDefinition.class, query, "definitionsUUIDs", uuids);
  }

  public Set<InternalProcessDefinition> getProcesses(Set<ProcessDefinitionUUID> definitionUUIDs, ProcessState processState) {
    Query query = getSession().getNamedQuery("getProcessesListByState");
    query.setParameter("state", processState);
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(InternalProcessDefinition.class, query, "definitionsUUIDs", uuids);
  }

  public List<InternalProcessDefinition> getProcesses(Set<ProcessDefinitionUUID> definitionUUIDs, int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getAllProcessesList");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQueryList(InternalProcessDefinition.class, query, "definitionsUUIDs", uuids);
  }

  public InternalProcessDefinition getLastProcess(Set<ProcessDefinitionUUID> definitionUUIDs, ProcessState processState) {
    Query query = getSession().getNamedQuery("getLastProcessFromProcessSetAndState");
    query.setParameter("state", processState, definitionStateUserType);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    query.setCacheable(true);
    query.setMaxResults(1);
    Set<InternalProcessDefinition> result = executeSplittedQuery(InternalProcessDefinition.class, query, "definitionsUUIDs", uuids);
    return result.iterator().next();
  }

  public Set<InternalProcessInstance> getProcessInstances(Set<ProcessDefinitionUUID> definitionUUIDs) {
    final Query query = getSession().getNamedQuery("getProcessInstancesFromDefinitionUUIDs");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(InternalProcessInstance.class, query, "definitionsUUIDs", uuids);
  }

  public Set<InternalProcessInstance> getUserInstances(String userId, Set<ProcessDefinitionUUID> definitionUUIDs) {
    Query query = getSession().getNamedQuery("getUserInstancesFromDefinitionUUIDs");
    query.setCacheable(true);
    query.setString("userId", userId);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(InternalProcessInstance.class, query, "definitionsUUIDs", uuids);
  }

  public int getNumberOfProcessInstances(Set<ProcessDefinitionUUID> definitionUUIDs) {
    final Query query = getSession().getNamedQuery("getNumberOfProcessInstancesFromDefinitionUUIDs");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    query.setParameterList("definitionsUUIDs", uuids);
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfParentProcessInstances(Set<ProcessDefinitionUUID> definitionUUIDs) {
    final Query query = getSession().getNamedQuery("getNumberOfParentProcessInstancesFromDefinitionUUIDs");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    query.setParameterList("definitionsUUIDs", uuids);
    return ((Long) query.uniqueResult()).intValue();
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstancesWithTaskState(Collection<ActivityState> activityStates,
      Set<ProcessDefinitionUUID> definitionUUIDs) {
    final Query getActivitiesQuery = getSession().getNamedQuery("findActivityInstancesWithTaskStateAndDefinitionsUUIDs");
    getActivitiesQuery.setCacheable(true);
    getActivitiesQuery.setParameterList("activityStates", activityStates, activityStateUserType);
    List<String> uuids = new ArrayList<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    getActivitiesQuery.setParameterList("definitionsUUIDs", uuids);
    uuids = getActivitiesQuery.list();

    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    if (uuids.isEmpty()) {
      return processInsts;
    }

    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    return executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids);
  }

  @SuppressWarnings("unchecked")
  public Set<InternalProcessInstance> getProcessInstancesWithInstanceStates(Collection<InstanceState> instanceStates,
      Set<ProcessDefinitionUUID> visibleProcessUUIDs) {
    final Query getProcessInstancesQuery = getSession().getNamedQuery("ProcessInstancesWithInstanceStatesAndDefinitionsUUIDs");
    getProcessInstancesQuery.setCacheable(true);
    getProcessInstancesQuery.setParameterList("instanceStates", instanceStates, instanceStateUserType);
    List<String> uuids = new ArrayList<String>();
    for (ProcessDefinitionUUID uuid : visibleProcessUUIDs) {
      uuids.add(uuid.toString());
    }
    getProcessInstancesQuery.setParameterList("definitionsUUIDs", uuids);
    uuids = getProcessInstancesQuery.list();

    final Set<InternalProcessInstance> processInsts = new HashSet<InternalProcessInstance>();
    if (uuids.isEmpty()) {
      return processInsts;
    }

    final Query query = getSession().getNamedQuery("findMatchingProcessInstances");
    query.setCacheable(true);
    return executeSplittedQuery(InternalProcessInstance.class, query, "instanceUUIDs", uuids);
  }

  public TaskInstance getOneTask(String userId, ActivityState taskState, Set<ProcessDefinitionUUID> definitionUUIDs) {
    Query query = getSession().getNamedQuery("getOneTaskOfProcessFromDefinitionUUIDs");
    query.setCacheable(true);
    query.setString("userId", userId);
    List<String> uuids = new ArrayList<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    query.setParameterList("definitionsUUIDs", uuids);
    query.setParameter("state", taskState, activityStateUserType);
    query.setMaxResults(1);
    return (TaskInstance) query.uniqueResult();
  }

  public List<InternalProcessInstance> getProcessInstances(Set<ProcessDefinitionUUID> definitionUUIDs,
      int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getProcessInstancesWithDefinitionUUIDs");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : definitionUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQueryList(InternalProcessInstance.class, query, "definitionsUUIDs", uuids);
  }

  public Rule getRule(String name) {
    Query query = getSession().getNamedQuery("getRuleByName");
    query.setCacheable(true);
    query.setString("name", name);
    query.setMaxResults(1);
    return (Rule) query.uniqueResult();
  }

  @SuppressWarnings("unchecked")
  public Set<Rule> getRules() {
    Query query = getSession().getNamedQuery("getRules");
    query.setCacheable(true);
    List<Rule> results = query.list();
    if (results != null) {
      return new HashSet<Rule>(results);
    } else {
      return Collections.emptySet();
    }
  }

  @SuppressWarnings("unchecked")
  public Set<Rule> getRules(Set<String> rulesName) {
    Query query = getSession().getNamedQuery("getRuleListByName");
    query.setCacheable(true);
    query.setParameterList("names", rulesName);
    List<Rule> results = query.list();
    if (results != null) {
      return new HashSet<Rule>(results);
    } else {
      return Collections.emptySet();
    }
  }

  @SuppressWarnings("unchecked")
  public Set<Rule> getRulesByType(Set<String> ruleTypes) {
    Query query = getSession().getNamedQuery("getRuleListByType");
    query.setCacheable(true);
    query.setParameterList("types", ruleTypes);
    List<Rule> results = query.list();
    if (results != null) {
      return new HashSet<Rule>(results);
    } else {
      return Collections.emptySet();
    }
  }

  @SuppressWarnings("unchecked")
  public Set<Rule> getAllApplicableRules(String entityID) {
    Query query = getSession().getNamedQuery("getRuleListByEntity");
    query.setCacheable(true);
    query.setString("entity", entityID);
    List<Rule> results = query.list();
    if (results != null) {
      return new HashSet<Rule>(results);
    } else {
      return Collections.emptySet();
    }
  }

  @SuppressWarnings("unchecked")
  public Set<Rule> getAllApplicableRules(String entityID, RuleType ruleType) {
    Query query = getSession().getNamedQuery("getRuleListByEntityAndType");
    query.setCacheable(true);
    query.setString("entity", entityID);
    String name = ruleType.name();
    query.setString("ruletype", name);
    List<Rule> results = query.list();
    if (results != null) {
      return new HashSet<Rule>(results);
    } else {
      return Collections.emptySet();
    }
  }

  public List<InternalProcessDefinition> getProcessesExcept(
      Set<ProcessDefinitionUUID> processUUIDs, int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getAllProcessesListExcept");
    query.setCacheable(true);
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : processUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQueryList(InternalProcessDefinition.class, query, "definitionsUUIDs", uuids);
  }

  public int getNumberOfActivityInstanceComments(ActivityInstanceUUID activityUUID) {
    final Query query = getSession().getNamedQuery("getNumberOfActivityComments");
    query.setCacheable(true);
    query.setString("activityUUID", activityUUID.getValue());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfComments(ProcessInstanceUUID instanceUUID) {
    final Query query = getSession().getNamedQuery("getNumberOfAllComments");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    return ((Long) query.uniqueResult()).intValue();
  }

  public int getNumberOfProcessInstanceComments(ProcessInstanceUUID instanceUUID) {
    final Query query = getSession().getNamedQuery("getNumberOfInstanceComments");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    return ((Long) query.uniqueResult()).intValue();
  }

  @SuppressWarnings("unchecked")
  public Map<ActivityInstanceUUID, Integer> getNumberOfActivityInstanceComments(Set<ActivityInstanceUUID> activityUUIDs) {
    final Query query = getSession().getNamedQuery("getActivitiesComments");
    query.setCacheable(true);
    List<String> uuids = new ArrayList<String>();
    for (ActivityInstanceUUID uuid : activityUUIDs) {
      uuids.add(uuid.toString());
    }
    query.setParameterList("uuids", uuids);
    List<Comment> comments = (List<Comment>) query.list();

    Map<ActivityInstanceUUID, Integer> result = new HashMap<ActivityInstanceUUID, Integer>();
    for (Comment comment : comments) {
      final ActivityInstanceUUID activityUUID = comment.getActivityUUID();
      if (result.containsKey(activityUUID)) {
        result.put(activityUUID, result.get(activityUUID) + 1);
      } else {
        result.put(activityUUID, 1);
      }

    }
    return  result;
  }

  @SuppressWarnings("unchecked")
  public List<Comment> getActivityInstanceCommentFeed(ActivityInstanceUUID activityUUID) {
    final Query query = getSession().getNamedQuery("getActivityComments");
    query.setCacheable(true);
    query.setString("activityUUID", activityUUID.getValue());
    return (List<Comment>) query.list();
  }

  @SuppressWarnings("unchecked")
  public List<Comment> getProcessInstanceCommentFeed(ProcessInstanceUUID instanceUUID) {
    final Query query = getSession().getNamedQuery("getInstanceComments");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    return (List<Comment>) query.list();
  }

  @SuppressWarnings("unchecked")
  public List<Comment> getCommentFeed(ProcessInstanceUUID instanceUUID) {
    final Query query = getSession().getNamedQuery("getAllComments");
    query.setCacheable(true);
    query.setString("instanceUUID", instanceUUID.getValue());
    return (List<Comment>) query.list();
  }


  @SuppressWarnings("unchecked")
  public List<Rule> getRules(RuleType ruleType, int fromIndex, int pageSize) {
    final Query query = getSession().getNamedQuery("getRuleListByExactType");
    query.setCacheable(true);
    query.setString("ruletype", ruleType.name());
    query.setFirstResult(fromIndex);
    query.setMaxResults(pageSize);
    final List<Rule> results = query.list();
    if (results == null) {
      return Collections.emptyList();
    }
    return results;
  }

  @SuppressWarnings("unchecked")
  public Set<String> getAllExceptions(String entityID, RuleType ruleType) {
    final Query query = getSession().getNamedQuery("getExceptionListByEntityAndRuleType");
    query.setString("entity", entityID);
    query.setString("ruletype", ruleType.name());
    query.setCacheable(true);
    List<String> results = query.list();
    if (results != null) {
      return new HashSet<String>(results);
    } else {
      return Collections.emptySet();
    }

  }

  @SuppressWarnings("unchecked")
  public Set<ProcessDefinitionUUID> getAllProcessDefinitionUUIDs() {
    final Query query = getSession().getNamedQuery("getAllProcessDefinitionUUIDs");
    query.setCacheable(true);
    List<ProcessDefinitionUUID> results = query.list();
    if (results != null) {
      return new HashSet<ProcessDefinitionUUID>(results);
    } else {
      return Collections.emptySet();
    }

  }

  public Set<ProcessDefinitionUUID> getAllProcessDefinitionUUIDsExcept(
      Set<ProcessDefinitionUUID> processUUIDs) {
    final Query query = getSession().getNamedQuery("getAllProcessDefinitionUUIDsExcept");
    query.setCacheable(true);
    Set<String> uuids = new HashSet<String>();
    for (ProcessDefinitionUUID uuid : processUUIDs) {
      uuids.add(uuid.toString());
    }
    return executeSplittedQuery(ProcessDefinitionUUID.class, query, "definitionsUUIDs", uuids);

  }

  public long getNumberOfRules(RuleType ruleType) {
    final Query query = getSession().getNamedQuery("getNumberOfRulesForType");
    query.setString("ruletype", ruleType.name());
    query.setCacheable(true);
    return ((Long) query.uniqueResult()).intValue();
  }

  //-- Web token management.
  public WebTemporaryToken getToken(String token) {
    final Query query = getSession().getNamedQuery("getTemporaryTokenFromKey");
    query.setString("tokenKey", token);
    query.setCacheable(true);
    query.setMaxResults(1);
    return ((WebTemporaryToken) query.uniqueResult());
  }

  @SuppressWarnings("unchecked")
  public Set<WebTemporaryToken> getExpiredTokens() {
    final Query query = getSession().getNamedQuery("getExpiredTemporaryTokens");
    query.setLong("currentDate", new Date().getTime()); 
    query.setCacheable(true);
    List<WebTemporaryToken> results = query.list();
    if (results != null) {
      return new HashSet<WebTemporaryToken>(results);
    } else {
      return Collections.emptySet();
    }
  }

}
