package org.ow2.bonita.runtime.event;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Synchronization;

import org.ow2.bonita.env.Environment;
import org.ow2.bonita.env.Transaction;
import org.ow2.bonita.runtime.model.Execution;
import org.ow2.bonita.services.CommandService;
import org.ow2.bonita.services.EventService;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.Command;
import org.ow2.bonita.util.EnvTool;

public class EventCoupleExceptionHandler implements Synchronization, Command<Object> {

  private static final Logger LOG = Logger.getLogger(EventCoupleExceptionHandler.class.getName());

  private static final long serialVersionUID = 1L;

  protected CommandService commandService;
  protected EventCoupleId eventCouple;
  protected Throwable exception;

  public EventCoupleExceptionHandler(EventCoupleId eventCouple, Throwable exception, CommandService commandService) {
    this.commandService = commandService;
    this.eventCouple = eventCouple;
    this.exception = exception;
  }


  public void beforeCompletion() {
  }

  public void afterCompletion(int status) {
    // after the transaction rolled back, 
    // execute this job exception handler object as a command with 
    // the command service so that this gets done in a separate 
    // transaction
    LOG.severe("starting new transaction for handling event couple exception");
    commandService.execute(this);
    LOG.severe("completed transaction for handling event couple exception");
  }

  public Object execute(Environment environment) throws Exception {
    final EventService eventService = EnvTool.getEventService();
    final Querier journal = EnvTool.getJournalQueriers();
    final EventExecutor eventExecutor = EnvTool.getEventExecutor();
    final OutgoingEventInstance outgoing = eventService.getOutgoingEvent(eventCouple.getOutgoing());
    final IncomingEventInstance incoming = eventService.getIncomingEvent(eventCouple.getIncoming());
    final String executionEventUUID = incoming.getExecutionUUID();
    final Execution execution = journal.getExecutionWithEventUUID(executionEventUUID);
    
    // serialize the stack trace
    final StringWriter sw = new StringWriter();
    exception.printStackTrace(new PrintWriter(sw));

    
    if (execution != null && incoming.isExecutionLocked()) {
      if (!Execution.STATE_ACTIVE.equals(execution.getState())) {
        execution.unlock();
      }
      execution.lock(execution.getState());
    }
    incoming.setLockOwner(null);
    outgoing.setLockOwner(null);

    int decrementedRetries = incoming.getRetries() - 1;
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Decrementing retries on incoming: " + incoming + ": " + decrementedRetries);
    }
    incoming.setRetries(decrementedRetries);
    incoming.setException(sw.toString());

    // notify the event executor after the transaction is completed
    Transaction transaction = environment.get(Transaction.class);
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("registering event executor notifier with " + transaction);
    }
    transaction.registerSynchronization(new EventAddedNotification(eventExecutor));
    return null;
  }

}
