package org.ow2.bonita.runtime.event;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.env.Environment;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.services.EventService;
import org.ow2.bonita.util.Command;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.GroovyExpression;
import org.ow2.bonita.util.GroovyUtil;

public class GetEventsCouplesCommand implements Command<Set<EventCoupleId>> {
  private static final long serialVersionUID = -479276850307735480L;
  static final Logger LOG = Logger.getLogger(GetEventsCouplesCommand.class.getName());
  
  public Set<EventCoupleId> execute(Environment environment) throws Exception {
    final EventService eventService = EnvTool.getEventService();
    final Set<EventCouple> eventCouples = eventService.getEventsCouples();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Found eventCouples in DB " + eventCouples);
    }
    if (eventCouples != null && !eventCouples.isEmpty()) {
      Set<EventCouple> validCouples = new HashSet<EventCouple>();
      final Set<Long> usedIncomings = new HashSet<Long>();
      final Set<Long> usedOutgoings = new HashSet<Long>();
      for (EventCouple eventCouple : eventCouples) {
        if (!usedIncomings.contains(eventCouple.getIncoming().getId()) && !usedOutgoings.contains(eventCouple.getOutgoing().getId())) {
          //matcher
          boolean match = true;
          final IncomingEventInstance incoming = eventCouple.getIncoming();
          final OutgoingEventInstance outgoing = eventCouple.getOutgoing();
          final String expression = incoming.getExpression();
          final Map<String, Object> parameters = outgoing.getParameters(); 
          if (expression != null) {
            final String groovyExpression = GroovyExpression.START_DELIMITER + expression + GroovyExpression.END_DELIMITER;
            final ActivityInstanceUUID activityUUID = incoming.getActivityInstanceUUID();
            if (activityUUID != null) {
              match = (Boolean) GroovyUtil.evaluate(groovyExpression, parameters, activityUUID, false);  
            } else {
              match = (Boolean) GroovyUtil.evaluate(groovyExpression, parameters);
            }
          }
          if (match) {
            usedIncomings.add(incoming.getId());
            usedOutgoings.add(outgoing.getId());
            validCouples.add(eventCouple);
            if (LOG.isLoggable(Level.INFO)) {
              LOG.info("Adding eventCouple:[incoming=" + incoming.getId() + ", outgoing=" + outgoing.getId() + "] to the queue");
            }
          }
        }
      }
      if (validCouples.isEmpty()) {
        return Collections.emptySet();
      }
      final Set<EventCoupleId> result = new HashSet<EventCoupleId>();
      for (EventCouple eventCouple : validCouples) {
        final IncomingEventInstance incoming = eventCouple.getIncoming();
        final OutgoingEventInstance outgoing = eventCouple.getOutgoing();
        //ieiUUID and oeiUUID must be different from each other else they are going to be consumed many times
        final String ieiUUID = UUID.randomUUID().toString();
        incoming.setLockOwner(ieiUUID);
        String oeiUUID = null;
        do {
          oeiUUID = UUID.randomUUID().toString();  
        } while (oeiUUID.equals(ieiUUID));
        
        outgoing.setLockOwner(oeiUUID);
        result.add(new EventCoupleId(incoming.getId(), outgoing.getId()));
      }
      if (LOG.isLoggable(Level.INFO)) {
        LOG.info("New event couples to execute: " + result.toString());
      }
      return result;
    }
    return null;
  }

}
