/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.facade.runtime.Comment;
import org.ow2.bonita.facade.runtime.impl.CaseImpl;
import org.ow2.bonita.facade.runtime.impl.CommentImpl;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;

/**
 * @author Anthony Birembaut
 *
 */
public final class DbMigration {

  private static final Logger LOG = Logger.getLogger(DbMigration.class.getName());

  private DbMigration() { }

  /**
   * Migrate DB schema of the database defined in the environment.
   */
  public static void migrateDb(String configurationName, String db, String currentVersion, String targetVersion) throws Exception {
    SessionFactory sessionFactory = DbTool.getSessionFactory(configurationName.replaceAll("-configuration", "-session-factory"));
    LOG.info("Running " + db + " " + configurationName + " DB migration...");
    InputStream inputStream = findMigrationScript(db, currentVersion, targetVersion);
    executeScript(sessionFactory, inputStream, db);
    executePostMigration(sessionFactory, targetVersion);
    LOG.info("---------------Schema migrated---------------");
  }

  @SuppressWarnings({ "unchecked", "deprecation" })
  private static void executePostMigration(SessionFactory sessionFactory, String targetVersion) throws Exception {
    if ("5.2".equals(targetVersion)) {
      final Session session = sessionFactory.openSession();
      session.getTransaction().begin();
      
      //Process instance comments migration
      final Query getInstancesQuery = session.createQuery("SELECT instance FROM org.ow2.bonita.facade.runtime.impl.InternalProcessInstance AS instance");
      final List<Object> processInstances = getInstancesQuery.list();
      for (Object processInstance : processInstances) {
        final InternalProcessInstance instance = (InternalProcessInstance)processInstance;
        final List<Comment> comments = instance.getCommentFeed(); 
        if (comments != null && !comments.isEmpty()) {
          for (Comment comment : comments) {
            final CommentImpl impl = (CommentImpl) comment;
            impl.setInstanceUUID(instance.getUUID());
          }
        }
      }
      
      //User id -> UUID migration
      final Query getUsersQuery = session.createQuery("SELECT DISTINCT user FROM org.ow2.bonita.facade.identity.impl.UserImpl AS user");
      final List<Object> users = getUsersQuery.list();
      for (Object user : users) {
        final UserImpl myUser = (UserImpl)user;
        myUser.setUUID(Long.toString(myUser.getId()));
      }
      
      //Role id -> UUID migration
      final Query getRolesQuery = session.createQuery("SELECT DISTINCT role FROM org.ow2.bonita.facade.identity.impl.RoleImpl AS role");
      final List<Object> roles = getRolesQuery.list();
      for (Object role : roles) {
        final RoleImpl myrole = (RoleImpl)role;
        myrole.setUUID(Long.toString(myrole.getId()));
      }
      
      //CaseImpl label
      final Query getCasesQuery = session.createQuery("SELECT webcase FROM org.ow2.bonita.facade.runtime.impl.CaseImpl AS webcase");
      final List<Object> cases = getCasesQuery.list();
      for (Object case_ : cases) {
        final CaseImpl mycase = (CaseImpl) case_;
        mycase.setLabel(CaseImpl.buildLabel(mycase.getOwnerName(), mycase.getLabelName()));
      }
      
      session.getTransaction().commit();
      session.close();
    }
  }

  public static void executeScript(SessionFactory sessionFactory, InputStream inputStream, String db) {
    byte[] bytes = IoUtil.readBytes(inputStream);
    String scriptContent = new String(bytes);
    List<String> commands = getCommands(scriptContent, db);

    Session session = sessionFactory.openSession();
    session.getTransaction().begin();

    LOG.severe("DB Commands Execution");
    for (String command: commands) {
      LOG.severe("Executing command : " + command);
      try {
        session.createSQLQuery(command).executeUpdate();
      } catch (Exception e) {
        System.err.println("Error while executing command: " + command);
        e.printStackTrace();
        LOG.severe(e.getMessage());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  public static List<String> getCommands(String scriptContent, String db) {
    String commandsSeparator = ";";
    if (db.equals("sqlserver") || db.equals("sybase")) {
      commandsSeparator = "go";
    }
    List<String> commands = new ArrayList<String>();
    int index = 0;
    while (index < scriptContent.length()) {
      int separatorIndex = scriptContent.indexOf(commandsSeparator, index);
      int separatorUpperCaseIndex = scriptContent.indexOf(commandsSeparator.toUpperCase(), index);
      if (separatorUpperCaseIndex != -1 && separatorUpperCaseIndex < separatorIndex) {
        separatorIndex = separatorUpperCaseIndex;
      }
      if (separatorIndex == -1) {
        separatorIndex = scriptContent.length();
      }
      String command = scriptContent.substring(index, separatorIndex).trim();
      if (command.length() > 0) {
        commands.add(command);
      }
      index = separatorIndex + commandsSeparator.length();
    }
    return commands;
  }

  public static InputStream findMigrationScript(String db, String currentVersion, String targetVersion) {

    StringBuilder migrationScript = new StringBuilder();
    migrationScript.append("/migration/")
    .append(db)
    .append("-")
    .append(currentVersion)
    .append("-")
    .append(targetVersion)
    .append(".sql");
    LOG.info("Loading Script " + migrationScript.toString());
    InputStream inputStream = DbMigration.class.getResourceAsStream(migrationScript.toString());
    if (inputStream == null) {
      String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_2");
      throw new IllegalArgumentException(message);
    }
    return inputStream;
  }

  public static void main(String[] args) throws Exception {
    if (args == null || args.length != 3) {
      String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_1");
      throw new IllegalArgumentException(message);
    }
    String db = args[0].toLowerCase();
    String currentVersion = args[1];
    String targetVersion = args[2];
    LOG.info("Starting History DB migration");
    migrateDb(EnvConstants.HB_CONFIG_HISTORY, db, currentVersion, targetVersion);
    LOG.info("Starting Core DB migration");
    migrateDb(EnvConstants.HB_CONFIG_CORE, db, currentVersion, targetVersion);
  }

}
