/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.cache.Cache;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Table;
import org.hibernate.tool.hbm2ddl.SchemaExport;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.env.EnvironmentFactory;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.def.element.impl.MetaDataImpl;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.type.lob.Lob;

/**
 * @author Guillaume Porcher
 * Helper class to initialize DB tables from the environment.
 */
public final class DbTool {

  static final Logger LOG = Logger.getLogger(DbTool.class.getName());

  private DbTool() { }

  public static Configuration getConfiguration(final String configurationName) {
    final EnvironmentFactory envFactory = GlobalEnvironmentFactory.getEnvironmentFactory();
    if (envFactory == null) {
      String message = ExceptionManager.getInstance().getFullMessage("bh_DBT_1");
      throw new BonitaRuntimeException(message);
    }
    final Configuration config = (Configuration) envFactory.get(configurationName);
    if (config == null) {
      String message = ExceptionManager.getInstance().getFullMessage(
          "bh_DBT_2", configurationName);
      throw new BonitaRuntimeException(message);
    }
    return config;
  }

  private static SchemaExport getSchemaExport(final Configuration config) {
    return new SchemaExport(config, config.buildSettings());
  }

  /**
   * Export DB schema to the database defined in the environment.
   */
  @SuppressWarnings("unchecked")
  public static void recreateDb(final String configurationName) {
    final Configuration config = getConfiguration(configurationName);

    if (LOG.isLoggable(Level.FINE)) {
      final SessionFactoryImplementor sessionFactory = getSessionFactory(configurationName.replaceAll("-configuration", "-session-factory"));
      if (sessionFactory != null) {
        final Dialect dialect = sessionFactory.getDialect();
        String[] script = config.generateSchemaCreationScript(dialect);
        for (String s : script) {
          LOG.fine(s); 
        }
      }
    }
    if (isOnDb("mysql", config)) {
      LOG.severe("Running on MySQL DB, updating schema...");
      final PersistentClass pc = config.getClassMapping(Lob.class.getName());
      final Table table = pc.getTable();
      final Iterator<Column> columns = (Iterator<Column>) table.getColumnIterator();
      while (columns.hasNext()) {
        final Column column = columns.next();
        final String columnName = "BLOB_VALUE_";
        if (column.getName().equals(columnName)) {
          LOG.severe("Updating " + columnName + " column...");
          column.setSqlType("LONGBLOB");
          column.setLength(518576);
        }
      }
    }
    /*
    if (isOnDb("sybase", config)) {
      LOG.severe("Running on Sybase DB, updating schema...");
      //iterate over all tables and all columns to replace type=text
      final Iterator<Table> tables = (Iterator<Table>) config.getTableMappings();
      while (tables.hasNext()) {
        final Table table = tables.next();
        final Iterator<Column> columns = table.getColumnIterator();
        while (columns.hasNext()) {
          final Column column = columns.next();
          System.err.println("Column.name=" + column.getName() + ", column=" + column.getDefaultValue());
          if (!column.getName().equals("BLOB_VALUE_") && column.getSqlType() != null && column.getSqlType().equals("CLOB")) {
            LOG.severe("Updating " + column.getName() + " column...");
            column.setSqlType("LONGVARCHAR");
          }
        }
        
      }      
    }
    */
    
    final SchemaExport schemaExport = getSchemaExport(config);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Creating schema...");
    }
    /*
    try {
      System.err.println("\n\n\n\n\n\n\n\n\n******************************");
      String[] script = config.generateSchemaCreationScript((Dialect) Class.forName(config.getProperty("hibernate.dialect")).newInstance());
      for (String s : script) {
        System.err.println(s);
      }
      System.err.println("\n\n\n\n\n\n\n\n\n******************************");
    } catch (Exception e) {
      e.printStackTrace();
    }
    */
    schemaExport.create(false, true);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Schema created.");
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Adding default users...");
    }
    addDefaultUsers(configurationName);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Default users added.");
    }
  }

  public static void addDefaultUsers(final String configurationName) {
    final SessionFactoryImplementor sessionFactory = getSessionFactory(configurationName.replaceAll("-configuration", "-session-factory"));
    if (sessionFactory != null) {
      Session session = null;
      try {
        session = sessionFactory.openSession();
        Transaction tx = session.beginTransaction();
        RoleImpl userRole = createDefaultRole("user", "User", "The user Role");
        RoleImpl adminRole = createDefaultRole("admin", "Admin", "The admin Role");
        Set<RoleImpl> userRoles = new HashSet<RoleImpl>();
        userRoles.add(userRole);
        Set<RoleImpl> adminRoles = new HashSet<RoleImpl>();
        adminRoles.add(adminRole);
        addDefaultUser(session, "admin", "", "", "bpm", adminRoles);
        addDefaultUser(session, "john", "John", "Doe", "bpm", userRoles);
        addDefaultUser(session, "jack", "Jack", "Doe", "bpm", userRoles);
        addDefaultUser(session, "james", "James", "Doe", "bpm", userRoles);
        MetaDataImpl defaultUsersAddedMetadata = new MetaDataImpl("DEFAULT_USERS_CREATED", "true");
        session.save(defaultUsersAddedMetadata);
        tx.commit();
      } catch (HibernateException e) {
        LOG.log(Level.WARNING, "Unable to add the default users in the DB", e);
      } finally {
        if (session != null) {
          session.close();
        }
      }
    }
  }

  private static RoleImpl createDefaultRole(String name, String label, String description) throws HibernateException {
    RoleImpl userRole = new RoleImpl(name);
    userRole.setLabel(label);
    userRole.setDescription(description);
    return userRole;
  }

  private static void addDefaultUser(Session session, String username, String firstName, String lastName, String password, Set<RoleImpl> roles) throws HibernateException {
    UserImpl user = new UserImpl(username, password);
    user.setFirstName(firstName);
    user.setLastName(lastName);
    for (RoleImpl role : roles) {
      user.addRole(role);
    }
    session.save(user);
  }

  public static SessionFactoryImplementor getSessionFactory(final String sessionFactoryName) {
    final SessionFactory sessionFactory = (SessionFactory) GlobalEnvironmentFactory.getEnvironmentFactory().get(sessionFactoryName);
    if (sessionFactory != null && sessionFactory instanceof SessionFactoryImplementor) {
      return (SessionFactoryImplementor) sessionFactory;
    }
    return null;
  }

  public static void dropDb(final String configurationName) {
    final SchemaExport schemaExport = getSchemaExport(getConfiguration(configurationName));
    schemaExport.drop(false, true);
  }

  public static void main(final String[] args) {
    if (args == null || args.length < 1) {
      String message = ExceptionManager.getInstance().getFullMessage(
          "bh_DBT_3", EnvConstants.HB_CONFIG_CORE, EnvConstants.HB_CONFIG_HISTORY);
      throw new IllegalArgumentException(message);
    }
    for (String conf : args) {
      recreateDb(conf);
    }
  }

  public static boolean isOnDb(final String dbName, final Configuration config) {
    final String lowerDbName = dbName.toLowerCase();
    final String dialect = getDbDialect(config).toLowerCase();
    final String url = getDbUrl(config).split(":")[1].toLowerCase();
    return dialect.contains(lowerDbName) || url.contains(lowerDbName);
  }
  
  
  public static String getDbDialect(final Configuration config) {
    return config.getProperty("hibernate.dialect");
  }
  public static String getDbUrl(final Configuration config) {
    return config.getProperty("hibernate.connection.url");
  }
  public static String getDbUseQueryCache(final Configuration config) {
    return config.getProperty("hibernate.cache.use_query_cache");
  }
  public static String getDbUseSecondLeveleCache(final Configuration config) {
    return config.getProperty("hibernate.cache.use_second_level_cache");
  }
  public static String getDbDriverClass(final Configuration config) {
    return config.getProperty("hibernate.connection.driver_class");
  }

  @SuppressWarnings("unchecked")
  public static void cleanCache(final String sessionFactoryName) {
    // clean 2nd level cache:
    final SessionFactoryImplementor sessionFactory = getSessionFactory(sessionFactoryName);
    if (sessionFactory != null) {
      // clear query results
      sessionFactory.evictQueries();
      // clear 2nd level cache
      final Map<String, Cache> caches = ((SessionFactoryImplementor)sessionFactory).getAllSecondLevelCacheRegions();
      for (final Cache cache : caches.values()) {
        cache.clear();
      }
    }
  }

}
