/**
 * Copyright (C) 2009  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import static org.ow2.bonita.util.GroovyExpression.END_DELIMITER;
import static org.ow2.bonita.util.GroovyExpression.START_DELIMITER;
import groovy.lang.Binding;
import groovy.lang.Script;

import java.io.IOException;
import java.io.NotSerializableException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.QueryDefinitionAPI;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.InternalProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.exception.ActivityDefNotFoundException;
import org.ow2.bonita.facade.exception.DataFieldNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.impl.StandardAPIAccessorImpl;
import org.ow2.bonita.facade.impl.StandardQueryAPIAccessorImpl;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.AttachmentInstance;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.runtime.impl.InternalActivityInstance;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.runtime.impl.ObjectVariable;
import org.ow2.bonita.facade.runtime.impl.ProcessInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;

/**
 * 
 * @author Matthieu Chaffotte
 *
 */
public class GroovyUtil {

  private static class SimpleBinding extends Binding {
    Map<String, Object> allVariables = new HashMap<String, Object>();

    private InternalProcessDefinition processDefinition;
    private ProcessDefinition clientProcessDefinition;
    private InternalProcessInstance processInstance;
    private ProcessInstance clientProcessInstance;
    private ActivityInstance activityInstance;
    private ActivityInstance clientActivityInstance;

    public SimpleBinding(final Map<String, Object> variables,
        final InternalProcessDefinition processDefinition, final InternalProcessInstance processInstance, final ActivityInstance activityInstance
    ) throws IOException, ClassNotFoundException {
      super();
      if (variables != null) {
        for (Map.Entry<String, Object> variable : variables.entrySet()) {
          Object value = variable.getValue();
          if (value != null && value instanceof ObjectVariable) {
            value = ((ObjectVariable)value).getValue();
          }
          allVariables.put(variable.getKey(), value);
        }
      }
      this.processDefinition = processDefinition;
      this.processInstance = processInstance;
      this.activityInstance = activityInstance;
    }

    @Override
    public Object getVariable(String name) {
      if ("processDefinition".equals(name)) {
        if (clientProcessDefinition == null && processDefinition != null) {
          clientProcessDefinition = new ProcessDefinitionImpl(processDefinition);
        }
        return clientProcessDefinition;
      }
      if ("processInstance".equals(name)) {
        if (clientProcessInstance == null && processInstance != null) {
          clientProcessInstance = new ProcessInstanceImpl(processInstance);
        }
        return clientProcessInstance;
      }
      if ("activityInstance".equals(name)) {
        if (clientActivityInstance == null && activityInstance != null) {
          clientActivityInstance = new InternalActivityInstance(activityInstance);
        }
        return clientActivityInstance;
      }
      if ("loggedUser".equals(name)) {
        try {
          return EnvTool.getUserId();
        } catch (Throwable e) {
          return null;
        }
      }
      if (allVariables.containsKey(name)) {
        return allVariables.get(name);
      }
      return super.getVariable(name);
    }

    @Override
    public void setVariable(String name, Object value) {
      allVariables.put(name, value);
    }

    @Override
    public Map<String, Object> getVariables() {
      return allVariables;
    }
  }

  private static class PropagateBinding extends SimpleBinding {
    private Collection<String> varToPropagate = new HashSet<String>();
    private Collection<String> initialProcessVariableKeys = new HashSet<String>();
    private Collection<String> initialObjectVariables = new HashSet<String>();
    private Collection<String> initialNullVariables = new HashSet<String>();
    private ActivityDefinitionUUID activityUUID;
    private ProcessDefinitionUUID processUUID;

    public PropagateBinding(final Map<String, Object> variables, final Set<String> initialProcessVariableKeys, 
        final ActivityDefinitionUUID activityUUID, final ProcessDefinitionUUID processUUID,
        final InternalProcessDefinition processDefinition, final InternalProcessInstance processInstance, final ActivityInstance activityInstance
    ) throws IOException, ClassNotFoundException {
      super(variables, processDefinition, processInstance, activityInstance);
      this.initialProcessVariableKeys.addAll(initialProcessVariableKeys);
      if (variables != null) {
        for (Map.Entry<String, Object> variable : variables.entrySet()) {
          Object value = variable.getValue();
          String varName = variable.getKey();
          if (value != null && value instanceof ObjectVariable && initialProcessVariableKeys.contains(varName)) {
            value = ((ObjectVariable)value).getValue();
            initialObjectVariables.add(varName);
          }
          if (value == null) {
            initialNullVariables.add(varName);
          }
        }
      }
      this.activityUUID = activityUUID;
      this.processUUID = processUUID;
    }

    @Override
    public Object getVariable(String name) {
      varToPropagate.add(name);
      return super.getVariable(name);
    }

    @Override
    public void setVariable(String name, Object value) {
      super.setVariable(name, value);
      if (initialProcessVariableKeys.contains(name)) {
        varToPropagate.add(name);
      }
    }

    public Map<String, Object> getVariablesToPropagate() throws NotSerializableException, IOException, ClassNotFoundException, ActivityDefNotFoundException, DataFieldNotFoundException, ProcessNotFoundException {     
      Map<String, Object> variablesToPropagate = new HashMap<String, Object>();
      for (String varName : varToPropagate) {
        if (initialProcessVariableKeys.contains(varName)) {
          Object value = allVariables.get(varName);
          if (initialObjectVariables.contains(varName)) {
            value = new ObjectVariable(value);
          } else if (initialNullVariables.contains(varName)) {
            //try to get the type from definition
            StandardQueryAPIAccessorImpl accessor = new StandardQueryAPIAccessorImpl();
            QueryDefinitionAPI queryDefinitionAPI = accessor.getQueryDefinitionAPI(AccessorUtil.QUERYLIST_JOURNAL_KEY);
            DataFieldDefinition datafield = null;
            if (activityUUID != null) {
              try {
                datafield = queryDefinitionAPI.getActivityDataField(activityUUID, varName);
              } catch (DataFieldNotFoundException e) {
                datafield = queryDefinitionAPI.getProcessDataField(activityUUID.getProcessUUID(), varName);
              }
            }
            if (datafield == null && processUUID != null) {
              datafield = queryDefinitionAPI.getProcessDataField(processUUID, varName);  
            }
            if (datafield != null && datafield.getDataTypeClassName().equals(ObjectVariable.class.getName())) {
              value = new ObjectVariable(value);
            }
          }
          variablesToPropagate.put(varName, value);
        }
      }
      return variablesToPropagate;
    }
  }

  public static Object evaluate(String expression, Map<String, Object> variables, ClassLoader classLoader) throws GroovyException {
    return evaluate(expression, variables, null, null, null, false, classLoader);
  }

  public static Object evaluate(String expression, Map<String, Object> variables) throws GroovyException {
    return evaluate(expression, variables, null, null, null, false, null);
  }

  public static Object evaluate(String expression, Map<String, Object> context, ActivityInstanceUUID activityUUID, boolean propagate) throws GroovyException {
    ProcessInstanceUUID processInstanceUUID = null;
    ProcessDefinitionUUID processDefinitionUUID = null;
    processInstanceUUID = activityUUID.getProcessInstanceUUID();
    processDefinitionUUID = processInstanceUUID.getProcessDefinitionUUID();
    return evaluate(expression, context, processDefinitionUUID, activityUUID, processInstanceUUID, propagate, null);
  }

  public static Object evaluate(String expression, Map<String, Object> context, ProcessInstanceUUID instanceUUID, boolean propagate) throws GroovyException {
    ProcessDefinitionUUID processDefinitionUUID = null;
    if (instanceUUID != null) {
      processDefinitionUUID = instanceUUID.getProcessDefinitionUUID();
    }
    return evaluate(expression, context, processDefinitionUUID, null, instanceUUID, propagate, null);
  }

  public static Object evaluate(String expression, Map<String, Object> context, ProcessDefinitionUUID processUUID, boolean propagate) throws GroovyException {
    return evaluate(expression, context, processUUID, null, null, propagate, null);
  }

  private static Object evaluate(String expression, Map<String, Object> context, ProcessDefinitionUUID processDefinitionUUID, ActivityInstanceUUID activityUUID, 
      ProcessInstanceUUID instanceUUID, boolean propagate, ClassLoader classLoader) throws GroovyException {
    if (expression == null || "".equals(expression.trim())) {
      throw new GroovyException("The expression is null or empty");
    } else {
      int begin = expression.indexOf(START_DELIMITER);
      int end = expression.indexOf(END_DELIMITER);
      if (begin >= end) {
        throw new GroovyException("The expression is not a Groovy one: " + expression);
      }
    }
    ClassLoader ori = Thread.currentThread().getContextClassLoader();
    try {
      ProcessDefinitionUUID pUUID = processDefinitionUUID;
      if (pUUID == null && instanceUUID != null) {
        pUUID = instanceUUID.getProcessDefinitionUUID();
      }
      if (pUUID == null && activityUUID != null) {
        pUUID = activityUUID.getProcessInstanceUUID().getProcessDefinitionUUID();
      }
      if (pUUID != null && classLoader == null) {
        ClassLoader processClassLoader = ClassDataLoader.getProcessClassLoader(pUUID);
        Thread.currentThread().setContextClassLoader(processClassLoader);
      } else if (classLoader != null) {
        Thread.currentThread().setContextClassLoader(classLoader);
      }
      Map<String, Object> allVariables = new HashMap<String, Object>();
      ActivityDefinitionUUID activityDefinitionUUID = activityUUID == null ? null : activityUUID.getActivityDefinitionUUID();
      Map<String, Object> processRuntimeVariables = new HashMap<String, Object>();
      InternalProcessInstance instance = null; 
      if (instanceUUID != null) {
        StandardQueryAPIAccessorImpl accessor = new StandardQueryAPIAccessorImpl();
        QueryRuntimeAPI api = accessor.getQueryRuntimeAPI();
        processRuntimeVariables.putAll(api.getProcessInstanceVariables(instanceUUID));
        //The history queriers are also needed because we want to be able to evaluate expressions on terminated instances
        instance = EnvTool.getAllQueriers().getProcessInstance(instanceUUID);
        List<AttachmentInstance> attachments = instance.getAttachments();
        if (attachments != null) {
          for (AttachmentInstance attachment : attachments) {
            allVariables.put(attachment.getName(), attachment);
          }
        }
      }

      ActivityInstance activityInstance = null;
      if (activityUUID != null) {
        instance = EnvTool.getAllQueriers().getProcessInstance(activityUUID.getProcessInstanceUUID());
        activityInstance = EnvTool.getAllQueriers().getActivityInstance(activityUUID);
        long maxDate = -1;
        if (!activityInstance.getState().equals(ActivityState.READY) && !activityInstance.getState().equals(ActivityState.SUSPENDED)
            && !activityInstance.getState().equals(ActivityState.EXECUTING)) {
          maxDate = activityInstance.getLastStateUpdate().getUpdatedDate().getTime();  
        } else {
          maxDate = new Date().getTime();
        }
        List<VariableUpdate> instanceVarUpdates = instance.getVariableUpdates();
        Map<String, VariableUpdate> tmp = new HashMap<String, VariableUpdate>();

        for (VariableUpdate varUpdate : instanceVarUpdates) {
          if (varUpdate.getDate().getTime() <= maxDate) {
            tmp.put(varUpdate.getName(), varUpdate);
          }
        }
        processRuntimeVariables.putAll(instance.getInitialVariableValues()); 
        for (VariableUpdate varUpdate : tmp.values()) {
          processRuntimeVariables.put(varUpdate.getName(), varUpdate.getValue()); 
        }
        StandardQueryAPIAccessorImpl accessor = new StandardQueryAPIAccessorImpl();
        QueryRuntimeAPI api = accessor.getQueryRuntimeAPI();
        processRuntimeVariables.putAll(api.getActivityInstanceVariables(activityUUID));
      }
      InternalProcessDefinition process = null;
      if (processDefinitionUUID != null) {
        process = EnvTool.getAllQueriers().getProcess(processDefinitionUUID);
        Set<DataFieldDefinition> datafields = process.getDataFields();
        if (datafields != null) {
          for (DataFieldDefinition datafield : datafields) {
            if (!processRuntimeVariables.containsKey(datafield.getName())) {
              allVariables.put(datafield.getName(), datafield.getInitialValue());
            }
          }
        }
      }

      allVariables.putAll(processRuntimeVariables);
      if (context != null) {
        allVariables.putAll(context);
      }

      Binding binding = null;
      if (propagate) {
        binding = new PropagateBinding(allVariables, processRuntimeVariables.keySet(), activityDefinitionUUID, processDefinitionUUID, process, instance, activityInstance);
      } else {
        binding = new SimpleBinding(allVariables, process, instance, activityInstance);
      }
      Object result = null;
      if (Misc.isJustAGroovyExpression(expression)) {
        expression = expression.substring(START_DELIMITER.length());
        expression = expression.substring(0, expression.lastIndexOf(END_DELIMITER));
        result = evaluateGroovyExpression(expression, binding);
      } else {
        result = evaluate(getExpressions(expression), binding);
      }
      if (propagate && instanceUUID != null) {
        propagateVariables(((PropagateBinding)binding).getVariablesToPropagate(), activityUUID, instanceUUID);
      }
      return result;
    } catch (Exception e) {
      throw new GroovyException(e.getMessage(), e);
    } finally {
      if (ori != null && ori != Thread.currentThread().getContextClassLoader()) {
        Thread.currentThread().setContextClassLoader(ori);
      }
    }
  }

  private static void propagateVariables(Map<String, Object> variables, ActivityInstanceUUID activityUUID, ProcessInstanceUUID instanceUUID) throws GroovyException {
    StandardAPIAccessorImpl accessor = new StandardAPIAccessorImpl();
    RuntimeAPI runtime = accessor.getRuntimeAPI();
    if (variables != null) {
      for (String variable : variables.keySet()) {
        try {
          if (activityUUID != null) {
            runtime.setVariable(activityUUID, variable, variables.get(variable));
          } else {
            runtime.setProcessInstanceVariable(instanceUUID, variable, variables.get(variable));
          }
        } catch (BonitaException e) {
          throw new GroovyException(e.getMessage());
        }
      }
    }
  }

  private static Object evaluateGroovyExpression(String expression, Binding binding) {
    ClassLoader scriptClassLoader = Thread.currentThread().getContextClassLoader();
    Script script = GroovyScriptBuilder.getScript(expression, scriptClassLoader);
    script.setBinding(binding);
    return script.run();
  }

  private static String evaluate(List<String> expressions, Binding binding) {
    StringBuilder builder = new StringBuilder();
    int i = 0;
    while (i < expressions.size()) {
      String expression = expressions.get(i);
      if (expression.equals(START_DELIMITER)) {
        expression = expressions.get(++i);
        builder.append(evaluateGroovyExpression(expression, binding));
      } else {
        builder.append(expression);
      }
      i++;
    }
    return builder.toString();
  }

  private static List<String> getExpressions(String expression) {
    List<String> expressions = new ArrayList<String>();
    String concat = new String(expression);
    while (concat.contains(START_DELIMITER)) {
      int index = concat.indexOf(START_DELIMITER);
      expressions.add(concat.substring(0, index));
      concat = concat.substring(index);
      expressions.add(START_DELIMITER);
      int endGroovy = Misc.getGroovyExpressionEndIndex(concat);
      expressions.add(concat.substring(2, endGroovy - 1));
      concat = concat.substring(endGroovy);
    }
    expressions.add(concat);
    return expressions;
  }

}
