/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.facade.def.element.impl.IterationDescriptor;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition.JoinType;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition.SplitType;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;

/**
 * @author Guillaume Porcher
 *
 */
public final class IterationDetection {
  private static final Logger LOG = Logger.getLogger(IterationDetection.class.getName());

  /**
   * Helper object to store data about cycle structure
   *
   * @author Guillaume Porcher
   */
  private static class CycleObject {
    // List of entry nodes
    protected Set<String> entryNodes = new HashSet<String>();
    // List of exit nodes
    protected Set<String> exitNodes = new HashSet<String>();
    // List of nodes in the cycle
    protected Set<String> nodesInPath = new HashSet<String>();
    // List of transitions that are in the cycle
    protected Set<MyTransition> transitions = new HashSet<MyTransition>();

    @Override
    public String toString() {
      return "entryNodes: " + this.entryNodes + ", exitNodes: " + this.exitNodes + " path: " + this.nodesInPath + ", transitions: " + this.transitions;
    }
  }

  /**
   * Simple object that describes a transition.
   *
   * @author Guillaume Porcher
   */
  private static class MyTransition {
    // source node
    protected String fromNode;
    // destination node
    protected String toNode;
    @Override
    public String toString() {
      return this.fromNode + "->" + this.toNode;
    }
  }

  private static class InnerProcess {
    Map<String, Node> nodes = new HashMap<String, Node>();
    public void addNode(Node node) {
      this.nodes.put(node.getName(), node);
    }
    public Node getNode(String nodeName) {
      return nodes.get(nodeName);
    }
    public boolean hasNode(String name) {
      return nodes.containsKey(name);
    }
  }

  private static class Node {
    private String name;
    private Set<Transition> incomingTransitions = new HashSet<Transition>();
    private Set<Transition> outgoingTransitions = new HashSet<Transition>();
    public Node(String name) {
      super();
      this.name = name;
    }
    public String getName() {
      return name;
    }
    public Set<Transition> getOutgoingTransitions() {
      return outgoingTransitions;
    }
    public void addIncomingTransition(Transition t) {
      this.incomingTransitions.add(t);
    }
    public void addOutgoingTransition(Transition t) {
      this.outgoingTransitions.add(t);
    }
    public void removeOutgoingTransition(Transition transition) {
      this.outgoingTransitions.remove(transition);
    }
    public void removeIncomingTransition(Transition transition) {
      this.incomingTransitions.remove(transition);
    }
    public boolean hasOutgoingTransitions() {
      return !outgoingTransitions.isEmpty();
    }
  }

  private static class Transition {
    Node source;
    Node destination;
    public Transition(Node source, Node destination) {
      super();
      this.source = source;
      this.destination = destination;
    }
    public Node getDestination() {
      return destination;
    }
    public Node getSource() {
      return source;
    }
  }
  private IterationDetection() { }

  /**
   * Creates a spanning tree in spanningTree corresponding to the input process
   * and lists transitions that are not in the spanning tree.
   * Use a Deep First Search Algorithm for that.
   * Transitions that are not in the spanning tree can be:
   * - transitions that creates a second path between two nodes (e.g. split/join)
   * - transitions that creates loops
   * @param inProcess input process
   * @param spanningTree process where the corresponding spanning tree will be created
   * @return list of transitions of input process that are not in the spanning tree.
   */
  private static List<TransitionDefinition> createSpanningTree(final ProcessDefinition inProcess, final InnerProcess spanningTree) {
    List<TransitionDefinition> transitions = new ArrayList<TransitionDefinition>();
    for (ActivityDefinition inNode : inProcess.getInitialActivities().values()) {
      final Node outNode = new Node(inNode.getName());
      spanningTree.addNode(outNode);

      // Create a spanning tree in outProcess and list transitions that are not in the spanning tree.
      // Use a Deep First Search Algorithm for that.
      transitions.addAll(IterationDetection.processNode(inProcess, inNode, outNode, spanningTree));
    }
    return transitions;
  }

  /**
   * Deep first search algorithm that creates spanning tree
   * @param inNode node of the input process currently processed
   * @param outNode corresponding node in the spanning tree
   * @param outProcess spanning tree
   * @return list of unprocessed transitions during the DFS
   */
  private static List<TransitionDefinition> processNode(final ProcessDefinition inProcess, final ActivityDefinition inNode, final Node outNode, final InnerProcess outProcess) {
    final List<TransitionDefinition> nonProcessedTransitions = new ArrayList<TransitionDefinition>();
    if (inNode.hasOutgoingTransitions()) {
      for (final TransitionDefinition t : inNode.getOutgoingTransitions()) {
        final ActivityDefinition destNode = inProcess.getActivity(t.getTo());
        final String destName = destNode.getName();
        if (!outProcess.hasNode(destName)) {
          // create node
          final Node newNode = new Node(destName);
          outProcess.addNode(newNode);

          final Transition transition = new Transition(outNode, newNode);
          outNode.addOutgoingTransition(transition);
          newNode.addIncomingTransition(transition);

          nonProcessedTransitions.addAll(IterationDetection.processNode(inProcess, inProcess.getActivity(t.getTo()), newNode, outProcess));
        } else {
          // node already in the spanning tree, the transition is not in the spanning tree.
          nonProcessedTransitions.add(t);
        }
      }
    }
    return nonProcessedTransitions;
  }

  /**
   * Checks if there is a path from source to destination
   * @param sourceNode source
   * @param destNode destination
   * @return true if there is a path from source to destination
   */
  private static boolean existsPath(final Node sourceNode, final Node destNode) {
    if (sourceNode.equals(destNode)) {
      return true;
    }
    if (sourceNode.hasOutgoingTransitions()) {
      for (final Transition t : sourceNode.getOutgoingTransitions()) {
        if (IterationDetection.existsPath(t.getDestination(), destNode)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Find transitions in the given list of transitions that would form a cycle if added to the given process.
   * If a transition can be added to the process without creating a cycle, then the transition is added.
   * The process returned is a DAG (directed acyclic graph)
   * @param process input process (this process will be modified by calling this method)
   * @param transitions transitions to check
   * @return list of transitions that cannot be added to the DAG (transitions that would create a cycle).
   */
  private static List<TransitionDefinition> findCycleTransitions(final ProcessDefinition inProcess, final InnerProcess process, final List<TransitionDefinition> transitions) {
    final List<TransitionDefinition> cycleTransitions = new ArrayList<TransitionDefinition>();
    for (final TransitionDefinition t : transitions) {
      final String sourceNodeName = t.getFrom();
      final String destNodeName = t.getTo();

      // Check if when a transition is added to the spanning tree, it creates a cycle
      // check if there is already a path from transition dest to transition source.
      // adding the transition would then create a cycle.
      final Node sourceNode = process.getNode(sourceNodeName);
      final Node destNode = process.getNode(destNodeName);
      if (IterationDetection.existsPath(destNode, sourceNode)) {
        cycleTransitions.add(t);
      } else {
        // add transition to the DAG
        final Transition transition = new Transition(sourceNode, destNode);
        sourceNode.addOutgoingTransition(transition);
        destNode.addIncomingTransition(transition);
      }
    }
    return cycleTransitions;
  }


  /**
   * Fill the path and transitions fields of the CycleObject c.
   * @param c cycle object to fill
   * @param sourceNode source node of the path
   * @param destNode destination of the path
   * @return
   */
  private static boolean fillCycleObject(final CycleObject c, final Node sourceNode, final Node destNode) {
    if (sourceNode.equals(destNode)) {
      c.nodesInPath.add(sourceNode.getName());
      return true;
    }
    if (sourceNode.hasOutgoingTransitions()) {
      boolean res = false;
      for (final Transition t : sourceNode.getOutgoingTransitions()) {
        if (IterationDetection.fillCycleObject(c, t.getDestination(), destNode)) {
          c.nodesInPath.add(sourceNode.getName());
          final MyTransition myT = new MyTransition();
          myT.fromNode = t.getSource().getName();
          myT.toNode = t.getDestination().getName();
          c.transitions.add(myT);
          res = true;
        }
      }
      return res;
    }
    return false;
  }

  /**
   * Find entry nodes in the cycle. Checks that entry nodes does not use Xor Join.
   * @param c cycle object to fill
   * @param inProcess input process
   */
  private static void fillCycleObjectEntryNodes(final CycleObject c, final ProcessDefinition inProcess) {
    boolean hasEntryPointXor = false;
    for (final String nodeName : c.nodesInPath) {
      final ActivityDefinition sourceNode = inProcess.getActivity(nodeName);
      if (sourceNode.hasIncomingTransitions()) {
        // check for entryNode:
        // n is an entry node if:
        // - a transition from a node that is not in the cycle goes to n.
        // - a transition which is not in the cycle goes to n.
        for (final TransitionDefinition t : sourceNode.getIncomingTransitions()) {
          JoinType joinType;
          boolean isEntryNode = true;
          if (c.nodesInPath.contains(t.getFrom())) {
            for (final MyTransition myTr : c.transitions) {
              if (myTr.fromNode.equals(t.getFrom())
                  && myTr.toNode.equals(t.getTo())) {
                isEntryNode = false;
              }
            }
          }
          if (isEntryNode) {
            c.entryNodes.add(sourceNode.getName());
            joinType = sourceNode.getJoinType();
            if (JoinType.XOR.equals(joinType)) {
              hasEntryPointXor = true;
            } else if (JoinType.AND.equals(joinType)) {
              String message = ExceptionManager.getInstance().getFullMessage(
                  "bd_ID_1", c, sourceNode.getName());
              throw new BonitaRuntimeException(message);
            }
          }
        }
      }
    }
    if (c.entryNodes.size() == 0) {
      String message = ExceptionManager.getInstance().getFullMessage("bd_ID_2", c);
      throw new BonitaRuntimeException(message);
    }
    if (!hasEntryPointXor) {
      String message = ExceptionManager.getInstance().getFullMessage("bd_ID_3", c);
      throw new BonitaRuntimeException(message);
    }
  }

  /**
   * Find exit nodes in the cycle. Checks if exit nodes only use Xor split type.
   * @param c cycle object to fill
   * @param inProcess input process
   */
  private static void checkExitNodes(final CycleObject c, final ProcessDefinition inProcess) {
    for (final String nodeName : c.nodesInPath) {
      final ActivityDefinition sourceNode = inProcess.getActivity(nodeName);
      if (sourceNode.hasOutgoingTransitions()) {
        //check for exit
        for (final TransitionDefinition t : sourceNode.getOutgoingTransitions()) {
          if (!c.nodesInPath.contains(t.getTo())) {
            c.exitNodes.add(sourceNode.getName());
            // sourceNode is an exit node
            final SplitType splitType = sourceNode.getSplitType();
            // Only allow XOR split for exit nodes
            if (!SplitType.XOR.equals(splitType)) {
              // TODO: change this to an exception
              //              throw new BonitaRuntimeException(
              //                  "Error in iteration : " + nodeName + " is an exit node for cycle "
              //                  + c.path + "." + Misc.LINE_SEPARATOR
              //                  + "Split type of this node is " + splitType + " but only XOR is allowed.");
              IterationDetection.LOG.severe("Potential issue in iteration : " + nodeName + " is an exit node for cycle "
                  + c.nodesInPath + "." + Misc.LINE_SEPARATOR
                  + "Split type of this node is " + splitType + " but only XOR is supported." + Misc.LINE_SEPARATOR
                  + "An exception will be thrown at runtime if more than one transition is enabled at the same time.");
            }
          }
        }
      }
    }
  }

  /**
   * Create structure to add to the InternalProcess to handle cycles.
   * For that, the cycle structure is checked, then entry nodes are resolved.
   * @param inProcess InternalProcess
   * @param dag acyclic graph on which cycle structure will be found
   * @param transitions list of transitions that creates cycles
   * @return a map where the key is the name of a node in the input process,
   * and the value a list of IterationDescriptor that needs to be added to the node definition.
   */
  private static Set<IterationDescriptor> findCycleInits(final ProcessDefinition inProcess,
      final InnerProcess dag, final List<TransitionDefinition> transitions) {

    final Set<IterationDescriptor> iterationDescriptors = new HashSet<IterationDescriptor>();

    for (final TransitionDefinition t : transitions) {
      final String sourceNodeName = t.getFrom();
      final String destNodeName = t.getTo();

      final Node sourceNode = dag.getNode(sourceNodeName);
      final Node destNode = dag.getNode(destNodeName);


      // Add current transition to the DAG to create a cycle
      final Transition transition = new Transition(sourceNode, destNode);
      sourceNode.addOutgoingTransition(transition);
      destNode.addIncomingTransition(transition);

      // Create object that represents the cycle structure
      final CycleObject c = new CycleObject();
      // compute the cycle path and transitions
      if (IterationDetection.fillCycleObject(c, destNode, sourceNode)) {
        final MyTransition myT = new MyTransition();
        myT.fromNode = sourceNodeName;
        myT.toNode = destNodeName;
        c.transitions.add(myT);
      }
      // find entry nodes
      IterationDetection.fillCycleObjectEntryNodes(c, inProcess);

      // check exit nodes
      IterationDetection.checkExitNodes(c, inProcess);

      if (IterationDetection.LOG.isLoggable(Level.FINE)) {
        IterationDetection.LOG.fine("CycleObject :"  + c);
      }

      // Remove added transition to have a DAG for other transitions.
      sourceNode.removeOutgoingTransition(transition);
      destNode.removeIncomingTransition(transition);

      // Create IterationDescriptors for the process
      final Set<String> entryNodes = new HashSet<String>();
      final Set<String> exitNodes = new HashSet<String>();
      final Set<String> otherNodes = new HashSet<String>();
      for (final String nodeName : c.nodesInPath) {
        if (c.entryNodes.contains(nodeName)) {
          entryNodes.add(nodeName);
        }
        if (c.exitNodes.contains(nodeName)) {
          exitNodes.add(nodeName);
        }
        if (!c.entryNodes.contains(nodeName) && !c.exitNodes.contains(nodeName)) {
          otherNodes.add(nodeName);
        }
      }
      final IterationDescriptor itDescr = new IterationDescriptor(otherNodes, entryNodes, exitNodes);
      iterationDescriptors.add(itDescr);
    }
    return iterationDescriptors;
  }

  /**
   * Find cycles in the InternalProcess and add the required structures to handle iterations
   * @param inProcess process to check and modify
   * @return inProcess
   */
  public static ProcessDefinition findIterations(final ProcessDefinitionImpl inProcess) {
    final InnerProcess outProcess = new InnerProcess();
    // Create a spanning tree in outProcess and list transitions that are not in the spanning tree.
    // Use a Deep First Search Algorithm for that.
    final List<TransitionDefinition> unprocessedTransitions = IterationDetection.createSpanningTree(inProcess, outProcess);
    if (IterationDetection.LOG.isLoggable(Level.FINE)) {
      IterationDetection.LOG.info("Unprocessed transitions: " + unprocessedTransitions);
    }
    // Transitions can be out of the spanning tree because:
    // it's in a split join
    // or it's a cycle
    // The second pass determines if the transition creates a cycle and creates a DAG (Directed acyclic graph)
    final List<TransitionDefinition> cycleTransitions = IterationDetection.findCycleTransitions(inProcess, outProcess, unprocessedTransitions);
    if (IterationDetection.LOG.isLoggable(Level.FINE)) {
      IterationDetection.LOG.info("Cycle transitions: " + cycleTransitions);
    }

    // create Iteration descriptors
    final Set<IterationDescriptor> iterationDescriptors = IterationDetection.findCycleInits(inProcess, outProcess, cycleTransitions);
    if (IterationDetection.LOG.isLoggable(Level.FINE)) {
      IterationDetection.LOG.info("iteration descriptors: " + iterationDescriptors);
    }

    // update process
    for (final IterationDescriptor iterationDescriptor : iterationDescriptors) {
      inProcess.addIterationDescriptors(iterationDescriptor);
      //System.out.println("Adding iterationDescriptor to process: " + iterationDescriptor);
      for (String activityName : iterationDescriptor.getCycleNodes()) {
        final ActivityDefinitionImpl activity = (ActivityDefinitionImpl) inProcess.getActivity(activityName);
        activity.setInCycle(true);
        //System.out.println("Activity " + activityName + " is in a cycle");
      }
    }

    return inProcess;
  }
}
